/*                Bistro development code
* 
* This code may be freely distributed and modified under the
* terms of the GNU Lesser General Public Licence.  This should
* be distributed with the code.  If you do not have a copy,
* see:
*
*     http://www.gnu.org/copyleft/lesser.html
* 
* Copyright for this code is held jointly by the individual
* authors.  These should be listed in @author doc comments.
* 
*/

package org.starbeans.modules.bistro.services;

import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileUtil;
import org.openide.loaders.DataFolder;
import org.openide.util.Lookup;

/** The place to find all services.
 *
 * @author George Waldon
 */
public class BistroServiceRegistry implements BioServiceRegistry {

    private static final Logger LOGGER = Logger.getLogger(BistroServiceRegistry.class.getName());
    private  volatile static BioServiceRegistry registry;
    private static BioService rootservice;
    private ServiceLookup bistroLookup;
    private Map<String, String> serviceIDs = new HashMap<String, String>();
        
    public static BioServiceRegistry getDefault() {
        if(registry==null) {
            registry = new BistroServiceRegistry();
            ((BistroServiceRegistry)registry).initRootFolder(); //will instantiate all services
        }
        return registry;
    }
    
    private BistroServiceRegistry() {
        bistroLookup = new ServiceLookup();
        
    }

    public BioService findService(String className) {
        Object o = null;
        try {
            o = getBistroLookup().lookup(Class.forName(className));
        } catch (ClassNotFoundException ex) {
            LOGGER.log(Level.WARNING, "", ex);
        }
        if(o instanceof BioService)
            return (BioService)o;
        // Always instantiate the root if necessary
        if(className==null || className.equals(BioServiceRoot.class.getName()))
            return BioServiceRoot.getDefault();
        return null;
    }
    
    public BioService findServicebyID(String id) {
        if(!serviceIDs.containsKey(id)) {
            LOGGER.warning("ID: " + id + "  not found in registry");
            return null;
        }
        String className = serviceIDs.get(id);
        return findService(className);
    }

    /** Get the bioservice registry lookup.
     * 
     */
    public Lookup getBistroLookup() {
        return bistroLookup;
    }

    /**
     * register an object to the registry servidec lookup.
     */
    public void register(BioService service) {
        bistroLookup.register(service);
        if(serviceIDs.containsKey(service.getID()) || serviceIDs.containsValue(service.getClass().getName()))
            LOGGER.severe("Duplicate key or class in registry");
        serviceIDs.put(service.getID(), service.getClass().getName());
    }

    /**
     * unregister an object to the registry servidec lookup.
     */
    public void unregister(BioService service) {
        bistroLookup.unregister(service);
        serviceIDs.remove(service.getID());
    }

    
    //========================================================================
    
    private ServiceFolder rootfolder;
    private Object lock = new Object();
    private volatile boolean inited;

    
    public void initRootFolder() {
        if(inited==true)
            return;
        synchronized(lock) {
            if(inited==true)
            return;
        rootfolder = new ServiceFolder(findBioServiceRootFolder());
        rootfolder.waitFinished();
        inited = true;
        }
    }
    
    /**
     * Returns "Taverna/BioServices" folder.
     */
    private static DataFolder findBioServiceRootFolder() {
        FileObject fo = FileUtil.getConfigFile("Taverna/BioServices");
        if (fo == null) {
            IllegalStateException ex = new IllegalStateException();
            LOGGER.log(Level.SEVERE,"Taverna/BioServices folder not found in layer",ex);
            throw ex;
        }
        DataFolder df = DataFolder.findFolder(fo);
        return df;
    }

}
