/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2009 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or sbbuild/licenses/cddl-gpl-2-cp.txt. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * sbbuild/licenses/cddl-gpl-2-cp.txt.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2009 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 *
 * Portions Copyrighted 2008-2010 George Waldon
 */


package org.starbeans.modules.corelite.contextmenus;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.Action;
import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JComponent;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.JSeparator;
import javax.swing.event.PopupMenuEvent;
import javax.swing.event.PopupMenuListener;
import org.openide.awt.DynamicMenuContent;
import org.openide.cookies.InstanceCookie;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileUtil;
import org.openide.loaders.DataFolder;
import org.openide.loaders.DataObject;
import org.openide.loaders.DataShadow;
import org.openide.loaders.FolderInstance;
import org.openide.loaders.InstanceSupport;
import org.openide.nodes.Node;
import org.openide.util.ContextAwareAction;
import org.openide.util.HelpCtx;
import org.openide.util.Lookup;
import org.openide.util.NbBundle;
import org.openide.util.actions.Presenter;
import org.openide.util.actions.SystemAction;

/** An action allowing the implementation of dynamic context menus in the IDE
 * using the XML layer. The set of actions and submenus associated with a Context2Action
 * is described in the XML layer as followed. Once described in the layer,
 * simply add Context2Action instances in the menu or use 
 * SystemAction.get(mycontext2action.class) to implement them in the GUI. Menus
 * will be updated dynamically and only enabled items will be shown. 
 *
 * <PRE> 
 * &lt;filesystem>  
 *  &lt;folder name="Context2Actions"&gt;
 *      &lt;folder name = "myfolder"&gt;
 *          &lt;!-- one or more Context2Action instances --&gt;
 *          &lt;file name="a-context2-action.instance"/&gt;
 *          &lt;list of instances of actions, separators, and comparators&gt;
 *          &lt;folder name = "mysubfolder"&gt;
 *              &lt;list of instances of actions, separators, and comparators&gt;              
 *          &lt;/folder&gt;
 *      &lt;/folder&gt;
 *      &lt;folder name = "mysecondfolder"&gt;
 *          etc.
 *      &lt;/folder&gt;
 *  &lt;/folder&gt;
 * &lt;/folder&gt;
 * </PRE>
 *
 * Top subfolders must contain one or more instances of Context2Action (excluding
 * DataShadow of Context2Action); 
 * Inlining ({@link #setInline()}) indicates that first-order submenus are inline 
 * with the current menu; mix inline and non-inline variants of a Context2Action
 * to generate varied effects in the IDE. Max inlining ({@link #setInlineCountMax()})
 * set a threeshold between inlining and popup.
 * @author  George Waldon
 */

//PENDING Have Context2Action to contain other Context2Action. This is currently
// allowed (see use of DataShadow) with the risk of infinite recurrence when
// users modify menus.

public abstract class Context2Action extends SystemAction implements ContextActionFolder,
ContextAwareAction, Presenter.Menu, Presenter.Popup {
    
    static final long serialVersionUID = 1L;
    
    private static final Logger LOGGER = Logger.getLogger(Context2Action.class.getName());
    private ContextActionFolder folder;
    private boolean isInitialized = false;
    private boolean isInline = false;
    private int inlineCount= 0;
    
    public Context2Action() {
    }
    
    /** Call this method before using this action. Normally done during warming-up.
     *
     * @param df associated topfolder in the Context2Action folder. 
     */
    public synchronized void initialize(DataFolder df) {
        if(!isInitialized) {
            folder = new LazyFolder(df, this);
            if( isIconDrawn() ) {
                Icon icon = folder.getFolderIcon();
                if(icon!=null)
                    setIcon(icon);
            }
            isInitialized = true;
        }
    }
    
    /** Indicate whether the first submenu is inlined in the current menu or is
     * shown as a popup menu.
     */
    public boolean isInlineMenu() {
        return isInline;
    }
    
    /** Set the the first submenu inlined in the current menu; false by default.
     *
     * @param isinline set to true for inlining.
     */
    public void setInline( boolean isinline ) {
        this.isInline = isinline;
    }
    
    /** Indicate the maximum number of menu items to be inlined; If the menu contains
     * less items than the maximum allowed, items are inlined, otherwise they
     * are shown in a popup. 
     * 
     * For popup menus, require {@link #isInlineMenu()} to return 
     * <CODE>true</CODE> and {@link #getInlineCountMax()} to return a value
     * different than zero.
     *
     * In the main menus, balance between inlining and popups requires
     * two Context2Action to work in tandem; one of them must be inlined 
     * and both must have {@link #setInlineCountMax()} set to the same value. 
     */
    public void setInlineCountMax(int number) {
        if(number>0)
            inlineCount = number;
    }
    
    public int getInlineCountMax() {
        return inlineCount;
    }

    
    private DataFolder getParentFolder() {
        FileObject fo = FileUtil.getConfigFile ("Context2Actions"); //NOI18N
        if (fo == null) {
            LOGGER.warning("Context2Action folder not found in layer"); //NOI18N
            return null;
        }
        
        DataFolder C2folder = DataFolder.findFolder(fo);
        DataObject[] arr = C2folder.getChildren();
        for (int i = 0; i < arr.length; i++) {
            DataObject obj = arr[i];
            if (!obj.isValid())
                continue;
            FileObject fc = obj.getPrimaryFile();
            if(fc.isFolder()) {
                DataFolder cfolder = DataFolder.findFolder(fc);               
                DataObject[] darray = cfolder.getChildren();
                for(int j = 0; j<darray.length;j++) {
                    //hack, instances of DataShadow are excluded from the pool of top Context2Action
                    if(darray[j] instanceof DataShadow)
                        continue;
                    InstanceCookie cookie = darray[j].getCookie(InstanceCookie.class);
                    if( cookie instanceof InstanceCookie.Of) {
                        if( ((InstanceCookie.Of)cookie).instanceOf(this.getClass()) ) {
                            return cfolder;
                        }      
                    }
                }
            }
        }
        return null;
    }
    
    private ContextActionFolder getContextFolder() {
        if( !isInitialized ) {
            // sometimes called by MenuWarmUp, but otherwise 
            // already initialized during warm-up.
            DataFolder df = getParentFolder();
            if(df==null) {
                LOGGER.info("Context2action " + this.toString() + " not initialized"); //NOI18N
                return null;
            }
            initialize(df);
        }
        return folder;
    }
        
    /** Get the localized name of the folder.
     *
     * @return the localized name of the folder node delegate.
     */
    public String getDisplayName() {
            return getContextFolder().getDisplayName();
    }
    
    /** Get the non-localized name of the folder.
     *
     * @return layer name of the folder.
     */
    public String getFolderName() {
        return getContextFolder().getFolderName();
    }
    
    public String getName() {
        return getDisplayName() ;
    }
    
    public Icon getFolderIcon() {
        return getContextFolder().getFolderIcon();
    }
    
    abstract protected boolean isIconDrawn();
    
    /** Action only performed by delegate.
     */
    public void actionPerformed(java.awt.event.ActionEvent e) {
        assert false;
    }
        
    public Object[] getItems() {
       return getContextFolder().getItems();
    }
    
    public void prepareFolder() {
        getContextFolder().prepareFolder();
    }
    
    public Action getAction() {
        return this;
    }
    
    public Action createContextAwareInstance(Lookup actionContext) {
        return new DelegateAction(this, actionContext);
    }

    //Called by MenuBar during menu warming-task
    public JMenuItem getMenuPresenter() {
        if(!isInitialized) { //Early click on menu.
            DataFolder df = getParentFolder();
            if(df==null) {
                LOGGER.info("Context2action " + this.toString() + " not initialized"); //NOI18N
                return null;
            }
            initialize(df);
        }
        
        if(isInlineMenu())
            return new Context2Action.Inline(this,true,null,getInlineCountMax());
        else
            return new Context2Action.Popup(this);
    }
   
    public JMenuItem getPopupPresenter() {
        return new Context2Action.Popup(this);
    }
    
    /** Used in place of Context2Action which cannot be instantiated for
     * second-level menus.
     */
    private class SubstituteContext2Action implements Action, ContextActionFolder,
    ContextAwareAction, Presenter.Menu, Presenter.Popup {
        
        static final long serialVersionUID = 200408260712000L;
        
        private ContextActionFolder folder;
        
        private SubstituteContext2Action(LazyFolder folder) {
            this.folder = folder;
        }
        
        /** Get the localized name of the folder.
         * @return the localized name of the folder node delegate.
         */
        public String getDisplayName() {
            return folder.getDisplayName();
        }
        
        /** Get the non-localized name of the folder.
         * @return layer name of the folder.
         */
        public String getFolderName() {
            return folder.getFolderName();
        }
        
        public String getName() {
            return getDisplayName() ;
        }
        
        public Icon getFolderIcon() {
            return folder.getFolderIcon();
        }
        
        /** Action only performed by delegate.
         */
        public void actionPerformed(java.awt.event.ActionEvent e) {
            assert false;
        }
        
        public org.openide.util.HelpCtx getHelpCtx() {
            return HelpCtx.DEFAULT_HELP ; //HELP
        }
        
        public Object[] getItems() {
            return folder.getItems();
        }
        
        public void prepareFolder() {
            folder.prepareFolder();
        }
        
        public Action getAction() {
            return this;
        }
        
        public Action createContextAwareInstance(Lookup actionContext) {
            return new DelegateAction(this, actionContext);
        }
        
        public Object getValue(String key) {
            if (NAME.equals (key))
                return getName();
            else if(Action.SMALL_ICON.equals(key))
                return null;
                /* No Icon for sub-menus
                return getIcon(); */
            else
                return null;
        }
        
        public Icon getIcon() {
            return getFolderIcon();
        }

        /** Second generation menu folders have a presenter */
        public JMenuItem getMenuPresenter() { 
            return new Context2Action.Popup(this);
        }
        
        public JMenuItem getPopupPresenter() { 
            return new Context2Action.Popup(this);
        }
        
        public void addPropertyChangeListener(PropertyChangeListener listener) {}
        public boolean isEnabled() { return true; }
        public void putValue(String key, Object value) { }
        public void removePropertyChangeListener(PropertyChangeListener listener) { }
        public void setEnabled(boolean b) {  }
    }
    
    
    /** Folder whith lazy action creation.
     */
    private class LazyFolder implements ContextActionFolder {
        private DataFolder master;
        private ActionFolder slave;
        
        /** For creation of <code>associate</code> action.
         */
        private final Object lock = new Object();
        
        /** Context2Action or pseudo associated with this folder.
         */
        private Action associate;
        
        /** List of Actions in the folder. */
        private ArrayList<Object> alist = new ArrayList<Object>();
        
        public LazyFolder(final DataFolder df, Action associate) {
            master = df;
            try {
                master.setSortMode(DataFolder.SortMode.NONE);
            } catch (Exception e) {
                //do nothing
            }
            this.associate = associate;
        }
        
        public void prepareFolder() {
            if(slave == null) {
                slave = new ActionFolder();
                slave.waitFinished();
            }
        }
        
        public Object[] getItems() {
            prepareFolder();
            return alist.toArray();
        }
        
        public Action getAction() {
            synchronized (lock) {
                if(associate != null)
                    return associate;
                else
                    return associate = new SubstituteContext2Action(this);
            }
        }
        
        public String getFolderName() {
            return master.getName();
        }
        
        public String getDisplayName() {
            return master.getNodeDelegate().getDisplayName();
        }
        
        public Icon getFolderIcon() {
            Node n = master.getNodeDelegate();
            return new ImageIcon(
            n.getIcon(java.beans.BeanInfo.ICON_COLOR_16x16));
        }
        
        /**
         */
        private class ActionFolder extends FolderInstance {
            
            /** Start loading the content of the master folder.
             */
            public ActionFolder() {
                super(master);
                recreate();
            }
            
            /** The name of this folder
             *
             * @return the folder name of the master folder
             */
            @Override
            public String instanceName() {
                return LazyFolder.this.getFolderName();
            }
            
            /**
             * @return a <code>List</code>
             */
            @Override
            public Class instanceClass() {
                return LazyFolder.this.getAction().getClass();
            }
            
            @Override
            protected InstanceCookie acceptDataObject(DataObject dob) {
                return super.acceptDataObject(dob);
            }
            
            /** Accepts only cookies that can provide <code>Menu</code>.
             * @param cookie an <code>InstanceCookie</code> to test
             * @return cookie if <code>Menu</code> available.
             */
            @Override
            protected InstanceCookie acceptCookie(InstanceCookie cookie)
            throws IOException, ClassNotFoundException {
                Class c = cookie.instanceClass();
                
                // Prevent all nested Context2Action
                //if( Context2Action.class.isAssignableFrom(c) )
                //    return null;

                //PENDING Just remove contained Context2Action based on folder name
                // but do not prevent cyclic nesting that could generate a stack overflow;
                if (cookie != null) {
                    try {
                        final Object o = cookie.instanceCreate();
                        if (o instanceof Context2Action) {
                            String name = ((Context2Action)o).getFolderName();
                            String localName = instanceName();
                            if(name.equals(localName))
                                return null;
                        }
                    } catch (IOException ex) {
                        LOGGER.log(Level.INFO,"",ex);
                    } catch (ClassNotFoundException ex) {
                        LOGGER.log(Level.INFO,"",ex);
                    }
                }
                boolean is =
                Presenter.Menu.class.isAssignableFrom(c) ||
                JMenuItem.class.isAssignableFrom(c) ||
                JSeparator.class.isAssignableFrom(c) ||
                Action.class.isAssignableFrom(c);
                return is ? cookie : null;
            }
            
            /**
             * Returns a <code>LazyFolder</code> cookie for the specified
             * <code>DataFolder</code>.
             * @param df a <code>DataFolder</code> to create the cookie for
             * @return a <code>LazyFolder</code> for the specified folder
             */
            @Override
            protected InstanceCookie acceptFolder(DataFolder df) {
                return new InstanceSupport.Instance(new LazyFolder(df, null));
            }
            
            /** Create the objects represented by this folder.
             * @param cookies array of instance cookies for the folder
             * @return the updated <code>List</code> of Object contained in the folder.
             */
            protected Object createInstance(InstanceCookie[] cookies)
            throws IOException, ClassNotFoundException {
                ArrayList<Object> list = LazyFolder.this.alist;
                list.clear();
                LinkedList<Object> cll = new LinkedList<Object>();
                for (int i = 0; i < cookies.length; i++) {
                    try {
                        Object o = cookies[i].instanceCreate();
                        cll.add(o);
                    } catch (ClassNotFoundException ex) {
                        LOGGER.log(Level.INFO,"",ex);
                    } catch (IOException ex) {
                        LOGGER.log(Level.INFO,"",ex);
                    }
                }
                
                Iterator it = cll.iterator();
                while (it.hasNext()) {
                    Object obj = it.next();
                    list.add(obj);
                }
                return list;
            }


            // Block in ATW since NB6.7
            //@Override
            //protected Task postCreationTask(Runnable run) {
            //    return new AWT_Task(run);
            //}

            /** Pass in AWT thread.
             */
            private final class AWT_Task extends org.openide.util.Task {
                private boolean done;
                
                public AWT_Task(Runnable rn) {
                    super(rn);
                    org.openide.util.Mutex.EVENT.readAccess(this);
                }
                
                @Override
                public void run() {
                    if (!done) {
                        super.run();
                        done = true;
                    }
                }
                
                @Override
                public void waitFinished() {
                    if (javax.swing.SwingUtilities.isEventDispatchThread()) {
                        run();
                    } else {
                        super.waitFinished();
                    }
                }
            }
        }
    }
    
    
    /** Delegate an action representing a folder (ContextActionFolder)
     * which acts according to current context.
     */
    private static final class DelegateAction extends Object
    implements Action, Presenter.Menu, Presenter.Popup, Lookup.Provider {
        
        private Action delegator;
        private Lookup lookup;
        
        private PropertyChangeSupport support = new PropertyChangeSupport(this);
        
        public DelegateAction(Action delegator, Lookup context) {
            this.delegator = delegator;
            this.lookup = context;
        }
        
        @Override
        public String toString() {
            return super.toString() + "[delegate folder=" + delegator + "]"; //NOI18N
        }
        
        public Lookup getLookup() {
            return lookup;
        }
        
        public void actionPerformed(java.awt.event.ActionEvent e) {
            //There is nothing to perform
        }
        
        public void putValue(String key, Object obj) {}
                
        public Object getValue(String key) {
            /* Need to hack since getValue is final in SystemAction */
            if(Action.SMALL_ICON.equals(key)) {
                if (delegator instanceof Context2Action) {
                    Context2Action c2a = (Context2Action) delegator;
                    return c2a.isIconDrawn()? c2a.getValue(key) : null;
                }                    
            }                
            return delegator.getValue(key);
        }
        
        public Action getDelegator() {
             return delegator;
        }
        
        public boolean isEnabled() {
            return true;
        }
        
        public void setEnabled(boolean boo) {
        }
        
        public void addPropertyChangeListener(PropertyChangeListener listener) {
            support.addPropertyChangeListener(listener);
        }
        
        public void removePropertyChangeListener(PropertyChangeListener listener) {
            support.removePropertyChangeListener(listener);
        }
        
        public javax.swing.JMenuItem getMenuPresenter() {
            return null;
        }
        
        public javax.swing.JMenuItem getPopupPresenter() {
            if( delegator instanceof Context2Action ) {
                if(((Context2Action)delegator).isInlineMenu() ) {
                    List<Object> l = generate(this,false);
                    int len = l.size();
                    if( len == 0 )
                        return new EmptyMenuItem();
                    int max = ((Context2Action)delegator).getInlineCountMax();
                    if( max != 0 ) {
                        if( len> max )
                            return new Context2Action.Popup(this);
                        else
                            return new Context2Action.Inline(this,false,l,max);
                    } else
                        return new Context2Action.Inline(this,false,l,max);
                }
                else
                    return new Context2Action.Popup(this);
            } else
                 return new Context2Action.Popup(this);
        }
    }
    
    /** Inline menu.
     */
    private static final class Inline extends JMenuItem implements DynamicMenuContent {
        
        static final long serialVersionUID = 1L;
        
        /** Associated context action */
        private Action c2action;
        private boolean ismenu;
        private List<Object> lItems;
        private int inlineCountMax = 0;
        
        Inline(Action c2Action, boolean isMenu, List<Object> l, int maxCount) {
            super();
            this.c2action = c2Action;
            this.ismenu = isMenu;
            this.lItems = l;
            this.inlineCountMax = maxCount;
        }

        public JComponent[] getMenuPresenters() {
            return synchMenuPresenters(new JComponent[0]);
        }  
        
        public JComponent[] synchMenuPresenters(JComponent[] items) {
            if( ismenu ) {
                lItems = generate(c2action,ismenu);
                if( inlineCountMax != 0 && lItems.size() > inlineCountMax )
                    lItems.clear();
            }
            if( lItems.isEmpty() && !ismenu ) {
                JMenuItem emptyitem = new JMenuItem(NbBundle.getMessage(
                        ContextActionFolder.class, "LBL_noSubMenu"));
                emptyitem.setEnabled(false);
                lItems.add(emptyitem);
            }
            //No unnecessary separator created
            JMenuItem[] arraymi = lItems.toArray(new JMenuItem[lItems.size()]);
            return (arraymi.length==1 && arraymi[0]==null)? null:arraymi;
        }
    }
    

    
    /** Menu which is either empty or contains one submenu to open 
     * the popup.
     *
     * Called for first level popups and menus.
     */
    private static final class Popup extends JMenuItem implements DynamicMenuContent {
        
        static final long serialVersionUID =200408221053000L;
        
        /** Associated delegate action */
        private Action action;
         
        public Popup(Action action) {
            /* Dummy name for testing purpose */
            super("popupname"); //NOI18N
            this.action = action;
        }
        
        private ContextActionFolder getContextActionFolder() {
            ContextActionFolder caf = null;
            if( action instanceof DelegateAction )
                caf = (ContextActionFolder)((DelegateAction) action).getDelegator();
            else if( action instanceof ContextActionFolder )
                caf = (ContextActionFolder)action;
            return caf;
        }

        public JComponent[] getMenuPresenters() {
            return synchMenuPresenters(new JComponent[0]);
        } 

        public JComponent[] synchMenuPresenters(JComponent[] items) {
            ContextActionFolder caf = getContextActionFolder();
            if( caf != null ) {
                if( isFolderEnabled(caf) ) {
                    //menu
                    if( action instanceof Context2Action ) {
                        if( !((Context2Action)action).isInlineMenu() ) {
                            int maxCount = ((Context2Action)action).getInlineCountMax();
                            if( maxCount != 0 ) {
                                List lItems = generate((Context2Action)action,true);
                                if( maxCount >= lItems.size() ) {
                                    return new JMenuItem[0];
                                }
                            }
                        }
                    } 
                    return new JMenuItem[] { new SubMenu(action) };
                } else
                    return new JMenuItem[0];
            } else
                return new JMenuItem[0];
        }
  
        /** A menu that updates its submenu before showing */
        private final class SubMenu extends JMenu implements PopupMenuListener {
            
            /* current popmenu we are listening to */
            private JPopupMenu curpop = null;
            
            /** Context action that originated this menu */
            private Action c2delegate;
            
            SubMenu(Action c2action) {
                super((String) c2action.getValue(Action.NAME));
                this.c2delegate = c2action;
                
                // Icon drawing
                if( c2action instanceof DelegateAction ) {
                    Action ac = null;
                    Icon icon = null;
                    ac = ((DelegateAction) c2action).getDelegator();
                    if( ac instanceof Context2Action ) {
                        if( ((Context2Action) ac).isIconDrawn()) {
                            icon = ((ContextActionFolder) ac).getFolderIcon();
                            setIcon(icon);
                        } 
                    }               
                //same in the menu
                } else if( c2action instanceof Context2Action ) {
                    Action ac = null;
                    Icon icon = null;
                    if( ((Context2Action) c2action).isIconDrawn()) {
                        icon = ((ContextActionFolder) c2action).getFolderIcon();
                        setIcon(icon);
                    }
                }
            }
                        
            @Override
            public JPopupMenu getPopupMenu() {
                JPopupMenu popup = super.getPopupMenu();
                fillPopmenu(popup);
                return popup;
            }
            
            private void fillPopmenu(JPopupMenu popmenu) {
                if (curpop == null) {
                    popmenu.addPopupMenuListener(this);
                    curpop = popmenu;
                    removeAll();
                    Iterator it = generate(c2delegate, false).iterator();
                    while( it.hasNext() ) {
                        java.awt.Component item = (java.awt.Component)it.next();
                        if( item == null ) {
                            addSeparator();
                        } else {
                            add( item );
                        }
                    }
                    
                    // Check if is empty
                    if(getMenuComponentCount() == 0) {
                        JMenuItem emptyitem = new JMenuItem(NbBundle.getMessage(
                        ContextActionFolder.class, "LBL_noSubMenu"));
                        emptyitem.setEnabled(false);
                        add(emptyitem);
                    }
                }
            }
            
            /* Implementation of PopupMenuListener */
            public void popupMenuCanceled(PopupMenuEvent e) {}
            public void popupMenuWillBecomeVisible(PopupMenuEvent e) {}
            public void popupMenuWillBecomeInvisible(PopupMenuEvent e) {
                curpop.removePopupMenuListener(this);
                curpop = null;
            }
        }
    }
    
    /** Implementation method that regenerates the items in the menu or
     * in the array.
     *
     * @param forMenu true if will be presented in menu or false if presented in popup
     * @param list (can be null)
     */
    private static List<Object> generate(Action c2action, boolean forMenu) {
        ContextActionFolder caf;
        if( c2action instanceof DelegateAction ) {
            DelegateAction delegate = (DelegateAction)c2action; //get the delegated action
            caf = (ContextActionFolder) delegate.getDelegator(); //get the delegator
        } else {
            caf = (ContextActionFolder) c2action;
        }
        Object[] elms = caf.getItems();
        List<Object> list = new ArrayList<Object>( elms.length );      
        boolean firstItemAdded = false;
        
        // Get context.
        Lookup lookup;
        if(c2action instanceof Lookup.Provider) {
            lookup = ((Lookup.Provider)c2action).getLookup();
        } else {
            lookup = null;
        }
        
        for (int i = 0; i < elms.length; i++) {
            Object o;
            
            if(elms[i] instanceof javax.swing.JSeparator) {
                if(!firstItemAdded) 
                    continue;
                int size = list.size();
                if( !(size!=0 && list.get(size-1)==null)) //do not repeat separators
                    list.add(null);
                continue;
            }

            //ok, here is a folder, get the context action in place, but
            //check if it contains something before!
            if(elms[i] instanceof ContextActionFolder) {
                if( isFolderEnabled((ContextActionFolder)elms[i]))
                    elms[i] = ((ContextActionFolder) elms[i]).getAction();
                else
                    continue;
            }
            
            // Retrieve context sensitive action instance if possible.
            if(lookup != null && elms[i] instanceof ContextAwareAction) {
                o = ((ContextAwareAction)elms[i]).createContextAwareInstance(lookup);
            } else {
                o = elms[i];
            }
            
            if (o == null) {
                int size = list.size();
                if( !(size!=0 && list.get(size-1)==null)) //do not repeat separators
                    list.add(null);
            } else {
                boolean isPopup = (o instanceof Presenter.Popup);
                boolean isMenu = (o instanceof Presenter.Menu);
                if (!((forMenu && isMenu) || (!forMenu && isPopup))
                && (isMenu || isPopup)) {
                    continue;
                }
                if(o instanceof Action) {
                    if (((Action)o).isEnabled()) {
                        JMenuItem mi;
                        if (forMenu && isMenu) {
                            mi = ((Presenter.Menu)o).getMenuPresenter();
                        } else if (!forMenu && isPopup) {
                            mi = ((Presenter.Popup)o).getPopupPresenter();
                        } else if (!isMenu && !isPopup) {
                            mi = new JMenuItem();
                            org.openide.awt.Actions.connect(mi, (Action)o, !forMenu);                    
                        } else {
                            // Should not be here.
                            continue;
                        }
                        
                        if( mi instanceof DynamicMenuContent ) {
                            JComponent[] comps = ((DynamicMenuContent)mi).getMenuPresenters();
                            if(comps.length==0) { //separator
                                mi = null;
                                int size = list.size();
                                if( !(size!=0 && list.get(size-1)==null) )
                                    list.add(mi); //do not repeat separators
                            } else if(comps.length==1) {
                                mi = (JMenuItem)comps[0];
                                Icon icon = (Icon) ((Action)o).getValue(SMALL_ICON);
                                if( icon!=null ) {
                                    // Folder icons!
                                    // hack: indirectly test if is SystemAction with a defined resource for icon
                                    String m = (String)((Action)o).getValue("iconBase");  //NOI18N
                                    if( m != null )
                                        mi.setIcon(icon);
                                }
                                list.add(mi);
                            } else {
                                for(int m=0; m<comps.length; m++) {
                                    mi = (JMenuItem)comps[m];
                                    list.add(mi);
                                }
                            }
                        } else {
                            Icon icon = (Icon) ((Action)o).getValue(SMALL_ICON);
                            if( icon!=null ) {
                                // Folder icons!
                                // hack: indirectly test if is SystemAction with a defined resource for icon
                                String m = (String)((Action)o).getValue("iconBase");  //NOI18N
                                if( m != null )
                                    mi.setIcon(icon);
                            }
                            list.add(mi);
                        }
                        firstItemAdded = true;
                    }
                }
            }
        }
        //remove end separators at end of poppup
        if(!forMenu) {
            while(true) {
                int lastIndex = list.size();
                if(lastIndex!=0 && list.get(lastIndex-1)==null)
                    list.remove(lastIndex-1);
                else break;
            }
        }
        return list;
    }
    
    /** Tells if folder contains enabled actions, and thus must be
     * presented.
     */
    private static boolean isFolderEnabled(ContextActionFolder folder) {
        boolean en = false;
        Object[] items = folder.getItems();
        
        // Get action context.
        Lookup lookup;
        if(folder instanceof Lookup.Provider) {
            lookup = ((Lookup.Provider)folder).getLookup();
        }  else {
            lookup = null;
        }
        
        for (int i=0; i<items.length; i++) {
            if( items[i] instanceof ContextActionFolder ) {
                boolean ennext = isFolderEnabled((ContextActionFolder) items[i]);
                if( ennext == true ) {
                    en = ennext;
                    break;
                }
                continue;
            }
            // Get context aware action instance if needed.
            Action act;
            // Retrieve context aware action instance if possible.
            if(lookup != null && items[i] instanceof ContextAwareAction) {
                act = ((ContextAwareAction)items[i]).createContextAwareInstance(lookup);
            } else {
                if( items[i] instanceof Action )
                    act = (Action) items[i];
                else
                    continue; //separator?
            }
            
            if( act.isEnabled() ) {
                en = true;
                break;
            }
        }
        
        return en;
    }

    /** A convenient dynamic empty JMenuItem.
     *
     */
    private static class EmptyMenuItem extends JMenuItem implements DynamicMenuContent {

        public JComponent[] getMenuPresenters() {
            return synchMenuPresenters(null);
        }

        public JComponent[] synchMenuPresenters(JComponent[] arg0) {
            return new JComponent[0];
        }

    }
}
