/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2009 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or sbbuild/licenses/cddl-gpl-2-cp.txt. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * sbbuild/licenses/cddl-gpl-2-cp.txt.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2009 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 *
 * Portions Copyrighted 2008-2010 George Waldon
 */

package org.starbeans.modules.corelite.contextmenus;

import java.awt.EventQueue;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.openide.cookies.InstanceCookie;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileUtil;
import org.openide.loaders.DataFolder;
import org.openide.loaders.DataObject;
import org.openide.loaders.DataShadow;
import org.openide.loaders.FolderInstance;
import org.openide.loaders.InstanceSupport;
import org.openide.util.actions.SystemAction;


/**
 *
 * @author  George Waldon
 */
public class Context2Manager {

    private static final Logger LOGGER = Logger.getLogger(Context2Manager.class.getName());
    
    /* The default Context2Manager */
    private static Context2Manager defManager;
    
    /* List of Context2Action in the main folder. */
    private ArrayList<Object> flist = new ArrayList<Object>();
    
    /** "Context2Actions" Folder */
    private MainContextFolder topfolder;
    
    public Object[] getTopContexts() {
        return flist.toArray();
    }
    
    private static GlobalListener gl;
    
    /** Lazy initialization of global listener.
     */
    private static synchronized GlobalListener gl () {
        if (gl == null) {
            gl = new GlobalListener ();
        }
        return gl;
    }
    
    private Context2Manager() {};
    
    /** Get the default Context2Manager.
     *
     * @return the instance created from Service Folder
     */
    public static Context2Manager getDefault() {
        synchronized (Context2Manager.class) {
            if(defManager == null ) {
                defManager = new Context2Manager();
            }
            return defManager;
        }
    }
    
    public void initMainFolder() {
        topfolder = new MainContextFolder(findContext2Folder());
        topfolder.waitFinished();
    }
    
    /**
     * Returns "Context2Action" folder.
     */
    public static DataFolder findContext2Folder() {
        FileObject fo = FileUtil.getConfigFile("Context2Actions"); //NOI18N
        if (fo == null) {
            IllegalStateException ex = new IllegalStateException();
            LOGGER.log(Level.WARNING,"Context2Action folder not found in layer",ex); //NOI18N
            throw ex;
        }
        DataFolder df = DataFolder.findFolder(fo);
        return df;
    }

    public void updateFolders() {
        Object[] obs = getTopContexts();
        if (obs != null) {
            for (int i=0; i<obs.length; i++) {
                createActionFolder(obs);
            }
        }
    }
    
    private static void createActionFolder(Object[] obs) {
        for (int i=0; i<obs.length; i++) {
            if (obs[i] instanceof ContextActionFolder) {
                try {
                    ContextActionFolder caf = (ContextActionFolder) obs[i];
                    caf.prepareFolder();
                    createActionFolder(caf.getItems());
                } catch (Exception e) {
                }
            } else if(obs[i] instanceof SystemAction) {
                // Just to have all SystemAction correctly initialized.
                SystemAction act = (SystemAction) obs[i];
                act.removePropertyChangeListener ( gl );
                act.addPropertyChangeListener ( gl );
            }
        }
    }

    
    private final class MainContextFolder extends FolderInstance {
        
        private boolean inCreation = false;
        
        public MainContextFolder(final DataFolder folder) {
            super(folder);
            inCreation = true;
            recreate();
            inCreation = false;
        }
        
        public synchronized void recreate() {
            super.recreate();
            if( !inCreation ) {
                final Context2Manager c2m = Context2Manager.getDefault();
                try {
                    EventQueue.invokeLater(new Runnable() {
                        public void run() {
                            c2m.updateFolders();
                        }
                    });
                } catch (Exception ex) {
                    ex.printStackTrace();
                }
            }
        }
        
        /** Class returned by createInstance()
         *
         * @return List.class - List of MainContextFolder;
         */
        @Override
        public Class instanceClass() {
            return List.class;
        }
        
        /** Removes all top folders from the list
         */
        private void cleanUp() {
            Context2Manager.this.flist.clear();
        }
        
        /** Adds top folder to the list of folders
         */
        private void add(Object o) {
            Context2Manager.this.flist.add(o);
        }
        
        protected Object createInstance(InstanceCookie[] cookies) throws java.io.IOException, ClassNotFoundException {
            cleanUp();
            for (int i = 0; i < cookies.length; i++) {
                Object o = cookies[i].instanceCreate();
                add(o);
            }
            return  Context2Manager.this.flist;
        }
        
        /* Main folder only contains top folders */
        @Override
        protected InstanceCookie acceptCookie(InstanceCookie cookie)
        throws IOException, ClassNotFoundException {
            IllegalStateException ex = new IllegalStateException(); //NOI18N
            LOGGER.log(Level.SEVERE,"cookies found in top level context2actions folder",ex); //NOI18N
            throw ex;
        }
        
        /* Returns an InstanceCookie of a Context2Action
         * for the specified DataFolder; only one is necessary to accept the 
         * folder.
         */
        @Override
        protected InstanceCookie acceptFolder(DataFolder df) {
            ArrayList<InstanceCookie> l = new ArrayList<InstanceCookie>(2);
            DataObject[] arr = df.getChildren();
            ArrayList<DataObject> list = new ArrayList<DataObject>(arr.length);
            for (int i = 0; i < arr.length; i++) {
                list.add(arr[i]);
            }
            Iterator<DataObject> dataObjects = list.iterator();
            while (dataObjects.hasNext()) {
                DataObject obj = dataObjects.next();
                if (!obj.isValid())
                    continue;
                //hack, instances of DataShadow are excluded from the pool of top Context2Action
                if(obj instanceof DataShadow)
                    continue;
                InstanceCookie cookie = obj.getCookie(InstanceCookie.class);
                if( cookie instanceof InstanceCookie.Of) {
                    if( ((InstanceCookie.Of)cookie).instanceOf(Context2Action.class) ) {
                        l.add(cookie);
                    }
                }
            }
            
            if(l.size() == 0) {
                LOGGER.fine("no Context2Action in folder: " + df.getName()); //NOI18N
                return null;
            }
            InstanceCookie iccurrent = null;
            Iterator<InstanceCookie> c2it = l.iterator();
            while(c2it.hasNext()) {
                try {
                    InstanceCookie ic = c2it.next();
                    Class cl = ic.instanceClass();
                    Class<? extends SystemAction> sc = cl.asSubclass(SystemAction.class);
                    SystemAction si = SystemAction.get(sc);
                    ((Context2Action) si).initialize(df);
                    if(iccurrent==null)
                        iccurrent =  new InstanceSupport.Instance(si);
                    //break; //no, allow initialization of all Context2Action
                }
                catch (Exception e) {
                    //Should not happend
                    LOGGER.log(Level.INFO,"Error in Context2Action initialization",e); //NOI18N
                }
            }
            return iccurrent;
        }
    }
    
    private static class GlobalListener implements PropertyChangeListener {        
        
        public GlobalListener() {
        }
        
        public void propertyChange (PropertyChangeEvent ev) {
        }
        
        private void actionsListChanged() {
        }

    }
}
