/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or sbbuild/licenses/cddl-gpl-2-cp.txt. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * sbbuild/licenses/cddl-gpl-2-cp.txt.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 2004-2007 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */


package org.starbeans.modules.seqformat;

import java.io.*;
import java.util.ArrayList;


/** This class is able to filter '\r' characters which is useful
 * when the parser requests a source not opened in the Editor.
 */
public class FilterReader extends Reader {
    
    /** New lines in this file was delimited by '\n' */
    static final byte NEW_LINE_N = 0;
    
    /** New lines in this file was delimited by '\r' */
    static final byte NEW_LINE_R = 1;
    
    /** New lines in this file was delimited by '\r\n' */
    static final byte NEW_LINE_RN = 2;
    
    /** Encapsulated reader */
    Reader reader;
    
    /** Character buffer */
    char[] charBuff;
    char[] readBuff;
    int howmany;
    
    /** The position at the current line. */
    int position;
    
    /** The count of types new line delimiters used in the file */
    final int[] newLineTypes;
    
    /** Creates new stream.
     * @param is encapsulated input stream.
     */
    public FilterReader(InputStream is) throws IOException {
        this(is, null);
    }
    
    public FilterReader(InputStream is, String encoding) throws IOException {
        if (encoding == null)
            reader = new InputStreamReader(is);
        else
            reader = new InputStreamReader(is, encoding);
        position = 0;
        newLineTypes = new int[] { 0, 0, 0 };
    }
    
    /** Read the array of chars */
    public int read(char[] cbuf, int off, int len) throws IOException {
        
        if (charBuff == null) {
            readCharBuff();
            translateToCharBuff();
        }
        
        if (howmany <= 0) {
            return -1;
        } else {
            int min = Math.min(len, howmany);
            System.arraycopy(charBuff, position, cbuf, off, min);
            howmany -= min;
            position += min;
            return min;
        }
    }
    
    /** Reads readBuff */
    final void readCharBuff() throws IOException {
        
        char[] tmp = new char[2048];
        int read;
        ArrayList<char[]> buffs = new ArrayList<char[]>(20);
        
        for (;;) {
            read = readFully(tmp);
            buffs.add(tmp);
            if (read < 2048) {
                break;
            } else {
                tmp = new char[2048];
            }
        }
        
        int listsize = buffs.size() - 1;
        int size = listsize * 2048 + read;
        readBuff = new char[size];
        charBuff = new char[size];
        int copy = 0;
        
        for (int i = 0; i < listsize; i++) {
            char[] tmp2 = buffs.get(i);
            System.arraycopy(tmp2, 0, readBuff, copy, 2048);
            copy += 2048;
        }
        System.arraycopy(tmp, 0, readBuff, copy, read);
    }
    
    /** reads fully given buffer */
    final int readFully(final char[] buff) throws IOException {
        int read = 0;
        int sum = 0;
        
        do {
            read = reader.read(buff, sum, buff.length - sum);
            sum += read;
        } while ((sum < buff.length) && (read > 0));
        
        return sum + 1;
    }
    
    /** Called after raw filling from an underlying reader */
    final void translateToCharBuff() {
        position = 0;
        
        // points to first unused cell in charBuff
        int charBuffPtr = 0;
        int stop = readBuff.length - 1;
        
        // read char
        int c;
        // ptr to first not processed char in readBuff
        int i = 0;
        // points to a character right after a newline
        int lastNewLine = 0;
        
        //process newlines so only '\n' appears in the charBuff
        //count all kinds of newlines - most used will be used on save
        while (i < stop) {
            c = readBuff[i];
            switch (c) {
                case (int)'\n':
                    newLineTypes[NEW_LINE_N]++;
                    charBuff[charBuffPtr++] = '\n';
                    lastNewLine = charBuffPtr;
                    i++;
                    break;
                case (int)'\r':
                    int c2 = readBuff[i + 1];
                    if (c2 != (int)'\n') {
                        newLineTypes[NEW_LINE_R]++;
                        i++;
                    } else {
                        i +=2;
                        newLineTypes[NEW_LINE_RN]++;
                    }
                    charBuff[charBuffPtr++] = '\n';
                    lastNewLine = charBuffPtr;
                    break;
                    
                default:
                    charBuff[charBuffPtr++] = readBuff[i++];
            }
        }
        
        if (i == stop) {
            c = readBuff[i];
            switch (c) {
                case (int)'\n':
                    newLineTypes[NEW_LINE_N]++;
                    charBuff[charBuffPtr++] = '\n';
                    break;
                case (int)'\r':
                    newLineTypes[NEW_LINE_R]++;
                    charBuff[charBuffPtr++] = '\n';
                    break;
                    
                default:
                    charBuff[charBuffPtr++] = readBuff[i++];
            }
        }
        howmany = charBuffPtr;
        readBuff = null;
    }
    
    /** Close underlayed writer. */
    public void close() throws IOException {
        reader.close();
    }
}