/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or sbbuild/licenses/cddl-gpl-2-cp.txt. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * sbbuild/licenses/cddl-gpl-2-cp.txt.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 2004-2007 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */


package org.starbeans.modules.seqformat.model;

import java.beans.PropertyChangeEvent;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;
import java.util.Vector;
import org.openide.nodes.Node;
import org.starbeans.modules.seqformat.source.Element;
import org.starbeans.modules.seqformat.source.SrcElement;
import org.starbeans.modules.seqformat.source.SrcException;

/**
 * This is the root point of the source text model. This object creates individual
 * model's elements and performs model-related tasks. All elements in the model have
 * (indirect) references to this object so they can contact it.
 * @author  svata
 * @version 
 */
abstract public class DefaultSeqModel implements SeqModel, SeqModel.Updater, Runnable {

    /** Mutex object for the whole model. The lock is obtained during model updates so
     * change detection and property changes are really atomic.
     */
    private Object          locked;
    
    private SeqModel.Env    env;
    
    private Object          writerNotify;
    
    private Thread          writingThread;
    
    private int             readerCount;
    
    // derived attributes:
    private BindingFactory  bindingFactory;
    
    private WrapperFactory  wrapperFactory;
    
    /** Event queue map for individual elements in the model. The map is keyed by
     * element instances, values maps <property_name> -> <property change event>
     */
    private EventQueue  eventQueue;
    
    /** Number of nested atomic locks.
     */
    private int         writeLocks;
    
    private int         masterLocks;
    
    /**
     * True, if all changes made in the transaction should be commited upon lock
     * release.
     */
    private boolean     transactionCommited;
    
    /**
     * True, if the transaction is constrained - changes are consulted with
     * VetoableChangeListeners and checked for consistency.
     */
    private boolean     transactionConstrained;
    
    /**
     * True, if the model is currently dispatching events.
     */
    private boolean     eventDispatch;

    /**
     * List of currently registered pre-commit listners.
     */
    Collection<CommitListener> preCommitListeners;
    
    /**
     * List of currently registered post-commit listeners.
     */
    Collection<CommitListener> postCommitListeners;
    
    /**
     * Holds top-level EventQueues in the order of dispatching. New queues are appened
     * during unlock to preserve the order of event generation.
     */
    LinkedList<EventQueue>  outputQueue;
    
    boolean              firingEvents;
    
    public DefaultSeqModel(SeqModel.Env env) {
        this.env = env;
        writerNotify = new Object();
        this.eventQueue = createEventQueue(null);
    }

    
    /** Clone the model object; the method does not clone the whole model, it rather clones
     * the management object so that objects created by the clone are compatible with
     * the original one. Usable mainly for paralel analysis of two sources and matching
     * them up.
     */
    public Object clone() {
        return null;
    }

    //=================== Implementation ElementCreator ======================//

    public SrcElementImpl createSource() {
        return createSourceFormat();
    }
    
    abstract protected SrcElementImpl createSourceFormat();


    public SeqElementImpl createSeqElement(SrcElement parent) {
        return createSeqElementFormat(parent);
    }
    
    abstract protected SeqElementImpl createSeqElementFormat(SrcElement parent);
    
    //========================================================================//
    
    // Call in ParsingSuppor.Processor.parseLockModel()
    public void addPreCommitListener(CommitListener l) {
        synchronized (this) {
            if (preCommitListeners == null)
                preCommitListeners = new LinkedList<CommitListener>();
        }
        synchronized (preCommitListeners) {
            preCommitListeners.add(l);
        }
    }
   
    // Used by ModelEventAdapter and SourceConnectionSupport
    public void addPostCommitListener(CommitListener l) {
        synchronized (this) {
            if (postCommitListeners == null)
                postCommitListeners = new LinkedList<CommitListener>();
        }
        synchronized (postCommitListeners) {
            postCommitListeners.add(l);
        }
    }
    
    public void removePreCommitListener(CommitListener l) {
        if (preCommitListeners == null)
            return;
        synchronized (preCommitListeners) {
            preCommitListeners.remove(l);
        }
    }
    
    public void removePostCommitListener(CommitListener l) {
        if (postCommitListeners == null)
            return;
        synchronized (postCommitListeners) {
            postCommitListeners.remove(l);
        }
    }
    
    final void notifyEventsDispatched(boolean dispatchOn) {
        this.eventDispatch = dispatchOn;
    }
    
    public final Object writeLock() {
        Object l = doWriteLock();
        if (masterLocks == -1) 
            eventQueue = createEventQueue(eventQueue);
        return l;
    }
    
    private EventQueue createEventQueue(EventQueue parent) {
        return new EventQueue(parent);
    }
    
    /**
     * The method obtains a write lock, but integrates all messages to the outside
     * queue instead of creating a local one.
     */
    final Object masterWriteLock() {
        Object l = doWriteLock();
        if (masterLocks > 0) {
            try {
                releaseWriteLock(l);
            } catch (SrcException ex) {
                // should NOT happen.
            }
            throw new IllegalStateException("Nested master locks!!"); // NOI18N
        }
        masterLocks = writeLocks;
        eventQueue = createEventQueue(eventQueue);
        return l;
    }
    
    public Object tryWriteLock() {
        return tryWriteLock(false);
    }
    
    private Object tryWriteLock(boolean master) {
        synchronized (writerNotify) {
            if (locked == null || writingThread == Thread.currentThread())
                return master ?
                    masterWriteLock() :
                    writeLock();
        }
        return null;
    }
    
    public final Object doWriteLock() {
        synchronized (writerNotify) {
            if (writingThread == Thread.currentThread()) {
                if (eventDispatch) {
                    throw new IllegalStateException(
                        "Modification from inside the event handler are banned"); // NOI18N
                }
                writeLocks++;
                return locked;
            }
            if (locked != null) {
                try {
                    writerNotify.wait();
                } catch (InterruptedException ex) {
                    throw new IllegalStateException("Interrupted"); // NOI18N
                }
            }
            // new lock - constrain the transaction.
            transactionConstrained = true;
            writeLocks++;
            locked = writerNotify;
            this.writingThread = Thread.currentThread();
            return locked;
        }
    }
    
    /**
     * Releases the write lock held on the model. If there are any outstanding
     * events, the CommitListener events are fired prior to releasing the lock.
     * If the operation is not aborted, write lock is released and queued 
     * PropertyChangeEvents are fired out. 
     */
    public final void releaseWriteLock(Object handle) throws SrcException {
        releaseWriteLock(handle, false);
    }
    
    final void releaseWriteLock(Object handle, boolean forkThread) throws SrcException {
        if (handle == null)
            throw new IllegalArgumentException("Invalid lock: " + handle); // NOI18N
        if (locked == null)
            throw new IllegalStateException("Model not locked."); // NOI18N
        synchronized (writerNotify) {
            if (handle != locked)
                throw new IllegalArgumentException("Invalid unlock attempt."); // NOI18N
        }
        
        EventQueue result = null;
        
        if (masterLocks == -1)
            result = mergeEventQueues();
        else if (masterLocks >= writeLocks) {
            result = mergeEventQueues();
            masterLocks = -1;
        }
        if (--writeLocks > 0) {
            return;
        }
        eventQueue = null;

        // fix change events so they contain snapshots of
        // new state in addition to the live reference.        
        result.fixupChanges();
        // can throw SrcException
        firePreCommitEvents(result);
        enqueue(result);
        synchronized (writerNotify) {
            this.writingThread = null;
            locked = null;
            writerNotify.notifyAll();
        }
        if (forkThread) {
            org.openide.util.RequestProcessor.getDefault().post(this);  //processOutputQueue();
        } else {
            processOutputQueue();   //Just proceed directly
        }
    }
    
    /**
     * This method operates under a write lock -- nobody else can modify or create the queue,
     * but somebody may be reading it.
     */
    private void enqueue(EventQueue q) {
        if (outputQueue == null)
            outputQueue = new LinkedList<EventQueue>();
        synchronized (outputQueue) {
            outputQueue.addLast(q);
        }
    }
    
    public void run() {
        processOutputQueue();
    }
    
    private void processOutputQueue() {
        if (this.outputQueue == null)
            return;
        
        EventQueue bit;
        synchronized (outputQueue) {
            if (firingEvents)
                return;
            firingEvents = true;
        }
        try {
            while (true) {
                synchronized (outputQueue) {
                    if (outputQueue.isEmpty()) {
                        return;
                    }
                    bit = outputQueue.removeFirst();
                }
                Map<ElementImpl, Collection<PropertyChangeEvent>> smap = bit.getSourceMap();
                bit.fireEvents();
                firePostCommitEvents(bit,smap);
            }
        } finally {
            synchronized (outputQueue) {
                firingEvents = false;
            }
        }
    }
    
    private void firePreCommitEvents(EventQueue q) {
        fireCommitEvents(preCommitListeners, q, null);
    }
    
    private void firePostCommitEvents(EventQueue what, Map<ElementImpl, Collection<PropertyChangeEvent>> smap) {
        fireCommitEvents(postCommitListeners, what, smap);
    }

    private void fireCommitEvents(Collection<CommitListener> origListeners, EventQueue data, Map<ElementImpl, Collection<PropertyChangeEvent>> smap) {
        if (origListeners == null || data == null || data.isEmpty())
            return;
        
        Collection<CommitListener> listeners;
        
        synchronized (origListeners) {
            listeners = new Vector<CommitListener>(origListeners);
        }
        
        Set removed = data.getRemovedElements();
        Set created = data.getCreatedElements();
        Map changed = data.getChangedElements();
        
        /*
         * Removed - for now. A cleaner spec of the veto semantic needed.
        if (isConstrained()) {
            for (Iterator it = listeners.iterator(); it.hasNext(); ) {
                CommitListener l = (CommitListener)it.next();
                l.queryCommitChanges(created, removed, changed, writeLocks);
            }
        }
         */

        // if the lock will be released after this operation, fire summary events
        // to watchers as well.
        for (Iterator<CommitListener> it = listeners.iterator(); it.hasNext(); ) {
            CommitListener l = it.next();
            l.changesCommited(created, removed, changed, smap);
        }
    }
    
    private EventQueue mergeEventQueues() {
        EventQueue parent = eventQueue.getParent();
        if (parent != null) {
            if (transactionCommited) {
                eventQueue.mergeToParent();
            } else {
                // PENDING: cancel ordinary events in the queue,
                // backfire all vetoable changes.
            }
            eventQueue = parent;
        }
        transactionCommited = false;
        return eventQueue;
    }
    
    public void commitChanges() {
        if (locked == null || this.writingThread != Thread.currentThread()) {
            throw new IllegalStateException("Sanity check: commit outside lock"); // NOI18N
        }
        if (masterLocks > 0 && masterLocks < writeLocks)
            return;
        this.transactionCommited = true;
    }
    
    public void runAtomic(Runnable r) throws SrcException {
        Object token = writeLock();
        try {
            r.run();
            commitChanges();
        } finally {
            releaseWriteLock(token);
        }
    }
    
    protected boolean isConstrained() {
        return this.transactionConstrained;
    }
    
    protected final <T extends Node.Cookie> T findElementCookie(Element el, Class<T> clazz) {
        return env.findElementCookie(el, clazz);
    }

    /**
     * Enters a special locked mode with disabled constraint and veto
     * checking.
     */
    public boolean runUpdate(Runnable r, boolean disableConstraints) 
    throws SrcException {
        Object token = tryWriteLock(true);
        if (token == null)
            return false;
        boolean saveConstraint = this.transactionConstrained;
        
        try {
            this.transactionConstrained = !disableConstraints;
            r.run();    // run processor with stage=1 ... run updatemodel
            /*
        } catch (SrcException ex) {
            throw ex;
             */
            commitChanges();
        } catch (Exception ex) {
            throw new SrcException("Unexpected implementation error"); // NOI18N
        } finally {
            this.transactionConstrained = saveConstraint;
            releaseWriteLock(token, true);
        }
        return true;
    }
    
    private void doRunAtomic(ExceptionRunnable r) throws SrcException {
        // nothing done here (for now).
        try {
            r.run();
        } catch (SrcException ex) {
            throw ex;
        } catch (Exception ex) {
            throw new SrcException("Unexpected implementation error."); // NOI18N
        }
    }
    
    public void runAtomic(ExceptionRunnable r) throws SrcException {
        Object token = writeLock();
        try {
            doRunAtomic(r);
            commitChanges();
        } finally {
            releaseWriteLock(token);
        }
    }
    
    public final void readLock() {
        synchronized (writerNotify) {
            if (writingThread != null) {
                if (writingThread == Thread.currentThread()) {
                    readerCount++;
                    return;
                }
            }
            if (locked != null) {
                try {
                    writerNotify.wait();
                } catch (InterruptedException ex) {
                    // PENDING: rethrow another (Runtime) exception.
                }
            }
            if (locked == null) {
                locked = Thread.currentThread();
            }
            readerCount++;
        }
    }

    public final void releaseReadLock() {
        synchronized (writerNotify) {
            if (--readerCount == 0) {
                if (writingThread == null) {
                    locked = null;
                    writerNotify.notifyAll();
                }
            }
        }
    }
    
    EventQueue getEventQueue() {
        return this.eventQueue;
    }
    
    public boolean isWriteLocked() {
        return this.writingThread != null;
    }
    
    public void notifyElementChanged(Element ref, Element old) {
        eventQueue.elementChanged(ref, old);
    }
    
    public void notifyElementCreated(Element ref) {
        eventQueue.elementCreated(ref);
    }
    
    public void notifyElementRemoved(Element ref) {
        eventQueue.elementRemoved(ref);
    }
    
    public void fireModelElementChange(ElementImpl bean, PropertyChangeEvent evt) {
        queuePropertyChange(bean, evt.getPropertyName(), evt);
    }
    
    public void fireModelElementChange(Element.Impl bean, PropertyChangeEvent evt) {
        queuePropertyChange((ElementImpl)bean, evt.getPropertyName(), evt);
    }
    
    // Model-private protocol

    protected BindingFactory getBindingFactory() {
        if (this.bindingFactory != null) {
            return this.bindingFactory;
        }
        return this.bindingFactory = env.getBindingFactory();
    }
    
    protected WrapperFactory getWrapperFactory() {
        if (this.wrapperFactory != null) {
            return this.wrapperFactory;
        }
        return this.wrapperFactory = env.getWrapperFactory();
    }
    
    // Implementation:
    private void queuePropertyChange(ElementImpl bean, String name, PropertyChangeEvent evt) {
        eventQueue.addPropertyChange(bean, evt);
    }
    
    protected ElementImpl getElementImpl(Element el) {
        return el.getCookie(ElementImpl.class);
    }
    
    public void updateMembers(Element target, String propertyName, Element[] els, 
        int[] orderIndices, int[] optMap) {          
        // Call from SrcInfo
        if (target instanceof SrcElement) {
            SrcElementImpl impl = (SrcElementImpl)getElementImpl(target);       
            impl.updateMembers(propertyName, els, orderIndices, optMap);
        }
    }
    
    public void updateMemberOrder(Element target, String id, 
        Element[] orderedMembers) {
        // Overridden in subclass
    }
    
    public void activate(Element target) {
        ElementImpl impl = getElementImpl(target);
        impl.notifyCreate();
    }
    
    public Binding getElementBinding(Element target) {
        ElementImpl impl = getElementImpl(target);
        return impl.getRawBinding();
    }
    
    public Element findElement(Element.Impl impl) {
        if (!(impl instanceof ElementImpl))
            return null;
        return ((ElementImpl)impl).getElement();
    }
    
    public void firePropertyChange(Element el, PropertyChangeEvent evt) {
        ElementImpl impl = getElementImpl(el);
        if (impl == null)
            return;
        impl.firePropertyChangeEvent(evt);
    }
 
    public void invalidateModel(SrcElement el) {
        SrcElementImpl impl = (SrcElementImpl)getElementImpl(el);
        Object token = writeLock();
        try {
            if (impl == null)
                return;
            impl.notifyRemove();
            commitChanges();
        } finally {
            try {
                releaseWriteLock(token);
            } catch (SrcException x) {
            }
        }
    }

}
