/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or sbbuild/licenses/cddl-gpl-2-cp.txt. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * sbbuild/licenses/cddl-gpl-2-cp.txt.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 2004-2007 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */


package org.starbeans.modules.seqformat.model;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyVetoException;
import java.beans.VetoableChangeListener;
import java.beans.VetoableChangeSupport;
import java.util.Collection;
import java.util.LinkedList;
import java.util.Vector;
import org.openide.nodes.CookieSet;
import org.openide.nodes.Node;
import org.openide.util.NbBundle;
import org.starbeans.modules.seqformat.source.Element;
import org.starbeans.modules.seqformat.source.ElementProperties;
import org.starbeans.modules.seqformat.source.Identifier;
import org.starbeans.modules.seqformat.source.SrcException;
import org.starbeans.modules.seqformat.source.SrcVetoException;

/**
 * Abstract base class that implements a single element in the source hierarchy.
 * This class contains the base services as event listening support, per-element
 * and per-model locking and various utility functions.
 *
 * @author  Svatopluk Dedic, Petr Hamernik, Jaroslav Tulach, George Waldon
 */
public abstract class ElementImpl implements 
    Element.Impl, ElementProperties, Node.Cookie, ElementEvents {

    /** The element for this class.
     */
    private transient Element               element;

    /**
     * Collection of registered PropertyChangeListeners.
     */
    private transient Collection<PropertyChangeListener> propListeners;
        
    /** Vetoable change support.
     */
    private transient VetoableChangeSupport vetoSupport;

    /**
     * Binding to the underlying source/whatever.
     */
    private transient Binding               binding;
    
    /**
     * True, if the element has been already created and was not destroyed.
     */
    private transient boolean               valid;

    /**
     * Set of cookies for this element. The set conveniently contains only cookies for
     * the implementation class hierarchy.
     */
    private transient CookieSet                       cookieSet;

    /**
     * Model that has created this element and that is used to manage the element.
     * The element delegates management functions like locking and event queueing to
     * its model.
     */
    private transient DefaultSeqModel      model;

    /**
     * Flag that is used to signal that the element is being created. Some processing
     * is reduced to minimum during this mode and event firing is supressed.
     */
    private transient boolean               inCreation;

    /** Flag to indicate that the element is marked, for example for inserting
     * new elements;
     */
    private boolean IsMarked;


    /** To be created statically in base subclass
     */
    protected static Binding NULL_BINDING = new NullBinding();
    
    /**
     * Enables debug trace messages for event generation
     */
    private static final boolean DEBUG_EVENTS = false;

    private static final long serialVersionUID = 200311201100000L;
    
    /**
     * Initializes model and sets the mode of the element to inCreation
     */
    protected ElementImpl(DefaultSeqModel model) {
        this.model = model;
        this.inCreation = true;
    }

    /**
     * Attaches the abstract layer to the element; since all properties required
     * for the Binding to operate are available, the binding is created as well.
     */
    public void attachedToElement(Element el) {
        this.element = el;
    }
    
    /**
     * Binds the element to a particular underlying Binding. The function does nothing
     * if the binding was already establised; it is impossible to rebind an element once
     * the binding was established (this constraint may be overriden by descendants)
     *
     * @param b binding to use for element i/o operations.
     */
    public void setBinding(Binding b) {
        if (this.binding != null)
            return;
        if (b instanceof Node.Cookie) {
            getCookieSet().add((Node.Cookie)b);
        }
        binding = b;
    }
    
    // Listener interface functions
    ///////////////////////////////////////////////////////////////////////////////////
    
    private String describeEvent(PropertyChangeEvent evt) {
        StringBuffer sb = new StringBuffer();
        sb.append(evt.getPropertyName());
        Object oldV = evt.getOldValue();        
        sb.append(" old = " + describeValue(evt.getOldValue())); // NOI18N
        sb.append(" new = " + describeValue(evt.getNewValue())); // NOI18N
        return sb.toString();
    }
    
    private String describeValue(Object o) {
        if (o instanceof Identifier) {
            Identifier id = (Identifier)o;
            return id.getName();
        } else if (o instanceof Identifier[]) {
            StringBuffer sb = new StringBuffer();
            sb.append("[ ");
            Identifier[] ids = (Identifier[])o;
            for (int i = 0; i < ids.length; i++) {
                if (i > 0)
                    sb.append(", ");
                sb.append(describeValue(ids[i]));
            }
            sb.append(" ]");
            return sb.toString();
        } else return "null"; // NOI18N
    }

    /** Adds property listener. 
     */
    public void addPropertyChangeListener (PropertyChangeListener l) {
        if (DEBUG_EVENTS) {
            System.err.println("[" + this + "] attaching listener " + l); // NOI18N
        }
        if (propListeners == null) {
            synchronized (this) {
                // new test under synchronized block
                if (propListeners == null) {
                    propListeners = new LinkedList<PropertyChangeListener>();
                    initializeListenerSupport();
                }
            }
        }
	synchronized (propListeners) {
	    propListeners.add(l);
	}
    }

    /**
     * Return false if the element is fully created. 
     */
    public boolean isCreated() {
        return this.inCreation;
    }
    
    /** Removes property listener */
    public void removePropertyChangeListener (PropertyChangeListener l) {
        if (DEBUG_EVENTS) {
            System.err.println("[" + this + "] removing listener " + l); // NOI18N
        }
	if (propListeners != null) {
	    synchronized (propListeners) {
		propListeners.remove(l);
	    }
	}
    }

    /** Adds property vetoable listener */
    public void addVetoableChangeListener (VetoableChangeListener l) {
        if (DEBUG_EVENTS) {
            System.err.println("[" + this + "] attaching veto listener " + l); // NOI18N
        }
        if (vetoSupport == null) {
            synchronized (this) {
                // new test under synchronized block
                if (vetoSupport == null) {
                    vetoSupport = new VetoableChangeSupport (element);
                    initializeListenerSupport();
                }
            }
        }
        vetoSupport.addVetoableChangeListener (l);
    }
    
    /** Removes property vetoable listener */
    public void removeVetoableChangeListener (VetoableChangeListener l) {
        if (DEBUG_EVENTS) {
            System.err.println("[" + this + "] removing veto listener " + l); // NOI18N
        }
        if (vetoSupport != null) {
            vetoSupport.removeVetoableChangeListener (l);
        }
    }

    /** Base method for marking the current insertion point.
     */
    public void markCurrent(boolean mark) {
        IsMarked = mark;
    }
 
    /** Test if the element is marked.
     */
   public boolean isMarked() {
       return IsMarked;
    }
    
    /** Returns the abstract wrapper for this implementation.
     */
    public final Element getElement() {
        return element;
    }
    
    /**
     * Returns the cookie set so that extension objects can plug in and
     * extend the element implementation.
     * @return r/w cookie set instance.
     */
    public final CookieSet getCookieSet() {
        if (cookieSet == null) {
            synchronized (this) {
                if (cookieSet == null) {
                    cookieSet = new CookieSet();
                    initializeCookies(cookieSet);
                }
            }
        }
        return cookieSet;
    }
    
    protected void initializeCookies(CookieSet set) {
        set.add(this);
    }
    
    /**
     * Returns true, if the element is still valid - that is present in the model.
     * @return true, if the element is still valid.
     */
    public boolean isValid() {
        return this.valid;
    }
    
    // Support functions 
    /////////////////////////////////////////////////////////////////////////////////
    
    /** Fires property change event. The event is fired as "own" property change event,
     * so it is captured in the queue and counted for the summary information.
    * @param name property name
    * @param o old value
    * @param n new value
    */
    protected final void firePropertyChange(String name, Object o, Object n) {
        fireOwnPropertyChange(new PropertyChangeEvent(getElement(), name, o, n));
    }
    
    /**
     * Fires a PropertyChangeEvent to the listeners, if there's a listener.
     */
    protected final void firePropertyChangeEvent(PropertyChangeEvent evt) {
        if (DEBUG_EVENTS) {
            StringBuffer sb = new StringBuffer();
            sb.append('[');
            sb.append(toString());
            sb.append("] Dispatching change: "); // NOI18N
            sb.append(describeEvent(evt));
            System.err.println(sb.toString());
        }
	if (propListeners == null)
	    return;
	
	Vector<PropertyChangeListener> listeners;
	synchronized (propListeners) {
	    listeners = new Vector<PropertyChangeListener>(propListeners);
	}
	for (int i = 0; i < listeners.size(); i++) {
	    PropertyChangeListener l = listeners.elementAt(i);
	    l.propertyChange(evt);
	}
    }
    
    /**
     * Fires an arbitrary property change event about intrinsic property change.
     * Events fired though this method will show up in the change summary events
     * fired after the model's lock is released. The method does nothing if the element
     * is in the creation phase.<P>
     * <B>Important note:<B> since the implementation will record the current element's
     * state by creating a copy/clone, it is <B>required</B> that the change is fired
     * before the element updates its internal state so that the clone contains the
     * old one.
     */
    protected final void fireOwnPropertyChange(PropertyChangeEvent evt) {
        if (isCreated())
            return;
        EventQueue q = getModelImpl().getEventQueue();
        q.elementChanged(this);
        addPropertyChange(evt);
    }
    
    protected Element cloneSelf() {
        throw new UnsupportedOperationException("clone unsupported on "  // NOI18N
            + getClass());
    }

    /**
     * Creates a suitable binding for the element. The binding is created with the
     * help of BindingFactory available from the element's model object; descendants 
     * are required to route the request to the most appropriate factory's method.
     */
    protected abstract Binding createBinding(Element el);

    /**
     * Adds an extrinsic property change event to the final event queue.
     * Events channeled through this method do not contribute to the summary events,
     * for example properties that contain children sub-elements should use
     * this method to avoid unnecessary element cloning.
     */
    public final void addPropertyChange(PropertyChangeEvent evt) {
        if (isCreated())
            return;
        getModelImpl().getEventQueue().addPropertyChange(this, evt);
    }
    
    /** Fires a vetoable change on property `name' from old value `o' to new value `n'.
     * @param name name of the property that is being changed.
     * @param o old value of the property; can be null.
     * @param n new value of the property; can be null.
     */
    protected final void fireVetoableChange(String name, Object o, Object n) throws PropertyVetoException {
        if (isCreated())
            return;
        if (vetoSupport != null) {
            try {
                getModelImpl().notifyEventsDispatched(true);
                vetoSupport.fireVetoableChange(name, o, n);
            } finally {
                getModelImpl().notifyEventsDispatched(false);
            }
        }
    }
    
    /**
     * Fires arbitrary pre-constructed vetoable change event.
     * @param evt vetoable event that should be fired.
     */
    public final void fireVetoableChange(PropertyChangeEvent evt) throws SrcException {
        if (isCreated())
            return;
         checkVetoablePropertyChange(evt);
    }
    
    private void doFireVetoableChange(PropertyChangeEvent evt) throws PropertyVetoException {
        if (DEBUG_EVENTS) {
            StringBuffer sb = new StringBuffer();
            sb.append('[');
            sb.append(toString());
            sb.append("] Dispatching veto: "); // NOI18N
            sb.append(describeEvent(evt));
            System.err.println(sb.toString());
        }
        if (vetoSupport != null) {
            try {
                getModelImpl().notifyEventsDispatched(true);
                vetoSupport.fireVetoableChange(evt);
            } finally {
                getModelImpl().notifyEventsDispatched(false);
            }
        }
    }
    
    /** The method tries to fire a Vetoable Change event; if (one of) listener throws 
     * PropertyVetoException, the exception is examined whether it holds a wrapped SourceException.
     * If so, that inner SourceException is rethrown, otherwise, the PropertyVetoException is
     * wrapped into a SourceException and thrown to the caller.
     */
    protected void checkVetoablePropertyChange(PropertyChangeEvent evt) throws SrcException {
        if (isCreated() || !isConstrained())
            return;
        try {
            doFireVetoableChange(evt);
        } catch (SrcVetoException ex) {
            // rethrow the original exception.
            throw ex.getNestedException();
        } catch (PropertyVetoException ex) {
            // rethrow the veto as a general SourceException.
            throw new SrcException(ex.getMessage());
        }
    }
    
    /**
     * Returns the storage binding for this element. The element must be 
     * backed up by a text file/document, but the changes themselves and the binding
     * to the underlying document are performed by other delegate object.
     * If the element is just being created, the function returns NULL_BINDING that accepts
     * all requests, but does nothing. This way the element is blocked from altering the
     * storage until it is fully created.
     *
     * @return The binding for this element, or, if the element has inCreation flag,
     * it returns the default NULL_BINDING.
     */
    public final Binding getBinding() {
        if (isCreated())
            return NULL_BINDING;
        return this.binding;
    }

    /**
     * Returns the actual binding without any shielding for elements that are not yet
     * created.
     */
    public final Binding getRawBinding() {
        return this.binding;
    }
    
    /** Invalidates the element. The element will be no longer valid and may refuse some
     * operations that would depend on other objects in the model - since it will be no
     * longer considered to be a part of a model.
     */
    protected void invalidate() {
        setValid(false);
    }
    
    private void setValid(boolean valid) {
        boolean old = this.valid;
        if (old == valid)
            return;
        this.valid = valid;
        if (old)
            // do not clone the element upon THIS property change.
            addPropertyChange(new PropertyChangeEvent(getEventSource(),
                PROP_VALID, 
                valid ? Boolean.FALSE : Boolean.TRUE,
                valid ? Boolean.TRUE : Boolean.FALSE 
            ));
    }

    /**
     * Checks whether the element is still valid. If not, throws a SourceException
     * to indicate that the calling operation cannot be completed.
     * @throw SourceException if the element is not valid/present in the model.
     */
    protected void checkValid(Object lockToken) throws SrcException {
        if (isValid()) {
            return;
        }
        releaseLock(lockToken);
        throw new SrcException(NbBundle.getMessage(ElementImpl.class,"EXC_ElementInvalid")); //NOI18N
    }

    /** Retrieves cookie supported by the Element. In general, implementation
     * classes are ALWAYS available as cookies, so they can be extracted from both ElementImpls
     * and the abstract counterparts.
     */
    public <T extends Node.Cookie> T getCookie(Class<T> desired) {
        // return this instance, if the cookie is directly supported.
        if (desired.isAssignableFrom(getClass()))
            return desired.cast(this);
        // ask the CookieSet
        T ret = getCookieSet().getCookie(desired);
        if (ret != null)
            return ret;
        return getModelImpl().findElementCookie(getElement(), desired);
    }
    
    
    /** Finds the source element - the root for the hierarchy this element are part of.
     */
    protected abstract SrcElementImpl findSource();
    
    // Package-private protocol
    ////////////////////////////////////////////////////////////////////////
   
    protected abstract boolean parentValid();
    
    public void notifyCreate() {
        inCreation = false;
        setValid(true);
        if (parentValid())
            notifyElementCreated();
    }
    
    protected void notifyElementCreated() {
        getModelImpl().getEventQueue().elementCreated(getElement());
    }
 
    /**
     * Notifies the element that it has been removed.
     */
    public void notifyRemove() {
        invalidate();
        getModelImpl().getEventQueue().elementRemoved(getElement());
    }
    
    /**
     * Checks whether the element can be removed. Fires a Vetoable property change
     * event on PROP_VALID from true to false.
     */
    protected void checkRemove() throws SrcException {
        if (isCreated() || !isConstrained())
            return;
        PropertyChangeEvent evt = new PropertyChangeEvent(getElement(),
            PROP_VALID, Boolean.TRUE, Boolean.FALSE);
        checkVetoablePropertyChange(evt);
    }
    
    /** Determines whether there's somebody interested in the property. Subclasses may
     * then optimize whether they should ever generate PropertyChangeEvent.
     */
    protected boolean hasListeners(String propName) {
	if (vetoSupport.hasListeners(propName))
	    return true;
	if (propListeners == null)
	    return false;
	synchronized (propListeners) {
	    return !propListeners.isEmpty();
	}
    }
    
//    /**
//     * Attempts to run atommically an operation. Throws SourceException if the runnable
//     * throws any kind of exception
//     * @deprecated use {@link #takeLock}/{@link #releaseLock} instead.
//     */
//    protected void runAtomic(ExceptionRunnable r) throws SrcException {
//        model.runAtomic(r);
//    }
    
    protected void initializeListenerSupport() {
    }

    /** 
     * Implementation of {@link ElementEvents} interface; returns the element to
     * be reported as the source of events.
     */
    public final Object getEventSource() {
        return getElement();
    }
    
    /**
     *  Implementation of {@link ElementEvents} interface; returns the implementation
     * object paired with the event source (this object).
     */
    public final ElementImpl getElementImpl() {
        return this;
    }
    
    protected abstract void setParent(ElementImpl parent);
    
    public void setParent(Element parent) {
        setParent(parent.getCookie(ElementImpl.class));
        setBinding(createBinding(this.element)); 
    }
    
    /**
     * Convenience method that retrieves reference to the model that had created
     * the element.
     */
    public final DefaultSeqModel getModelImpl() {
        return this.model;
    }
    
    /**
     * Takes a <B>simple</B> lock preventing other threads from modifying
     * the model. Before it returns, the method also checks whether the element
     * is still valid. If not, it releases the lock and throws a SrcException. 
     * Event resulting from an operation made under this lock will be added
     * to the event queue only if the operation is commited properly (see 
     * commit), i.e. lock released before the lock taken again. Therefore, under
     * this lock, operations are conducted in a sequential fashion.
     * 
     * @return token object that should be later used to free the lock.
     */
    protected final Object takeLock() throws SrcException {
        if (isCreated())
            return null;
        Object o = getModelImpl().writeLock();
        checkValid(o);
        return o;
    }
    
    /** Attempts to obtain model's write lock.
     * 
     * @return null if the model is already locked.
     * @throws org.starbeans.modules.seqformat.source.SrcException
     */
    protected final Object tryTakeLock()  throws SrcException {
        if (isCreated())
            return null;
        Object o = getModelImpl().tryWriteLock();
        checkValid(o);
        return o;
    }
    /**
     * Releases write lock on the model. If an invalid token object is passed,
     * the model will throw IllegalArgumentException
     * @param o token object for lock release operation.
     */
    protected final void releaseLock(Object o) throws SrcException {
        if (isCreated())
            return;
        getModelImpl().releaseWriteLock(o);
    }
    
    protected final void takeReadLock() {
        if (isCreated())
            return;
        getModelImpl().readLock();
    }
    
    protected final void releaseReadLock() {
        if (isCreated())
            return;
        getModelImpl().releaseReadLock();
    }


    /**
     * Takes a <B>master</B> lock. This lock differs from the ordinary in that it
     * does not create nested event queue, but merges all events to the current
     * event queue. The master lock is used when the model's implementation is 
     * issuing nested model operations.
     */
    protected final Object takeMasterLock() throws SrcException {
        if (isCreated())
            return null;
        Object l =  getModelImpl().masterWriteLock();
        checkValid(l);
        return l;
    }
    
    /** Clone whatever is necessary from the model to populate an element anew.
     */
    protected abstract void createFromModel(Element model) throws SrcException;
    
    /**
     * Causes changes made by the last locked operation to be confirmed and,
     * upon lock release, merged into the higher operation.
     */
    protected final void commit() {
        if (!isValid())
            return;
        getModelImpl().commitChanges();
    }
    
    /**
     * Returns true, if constraints on elements should be checked. This can
     * be disabled during some special operations (like external model updates).
     * If the method returns false, no VetoableChangeListeners should be informed.
     * @return true, if constraints are enabled, false otherwise.
     */
    protected final boolean isConstrained() {
        return getModelImpl().isConstrained();
    }

    
}
