/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or sbbuild/licenses/cddl-gpl-2-cp.txt. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * sbbuild/licenses/cddl-gpl-2-cp.txt.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 2004-2007 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */
package org.starbeans.modules.seqformat.model;

import java.beans.PropertyChangeEvent;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;
import org.starbeans.modules.seqformat.source.Element;

/**
 * Event/change queue that collects modification events made to a model in one locked
 * level of operations. After the model is unlocked, the events are merged up to the
 * parent. Upon unlocking the model entirely, the top-level queue will contain
 * both PropertyChangeEvents generated during the model locked operation and summary
 * changes that should be fired to CommitListeners.
 *
 * @author  sdedic
 * @version 
 */
class EventQueue extends Object {

    /**
     * The queue contains a stream of PropertyChangeEvents fired from locked operation(s), in the
     * order of their creation/firing.
     */
    Collection<ElementChangeEvent> eventQueue;
    /**
     * Map of elements -> event queue. Events are sorted by their creation time in the
     * queue; new elements are appended at the end.
     */
    Map<ElementImpl, Collection<PropertyChangeEvent>> sourceMap;
    /**
     * Map that collects elements that have been changes and their corresponding
     * original properties (in a form of cloned element).
     */
    Map<Element, Element> changedElements;
    /**
     * Set that collects new elements.
     */
    Set<Element> newElements;
    /**
     * Set that collects elements that have been removed.
     */
    Set<Element> removedElements;
    /**
     * The parent event queue that collects events originating in the outer lock.
     */
    EventQueue parent;
    private boolean firingEvents;

    /** Constructs an event queue, optionally linking it to a parent event queue.
     */
    public EventQueue(EventQueue parent) {
        this.parent = parent;
    }

    /**
     * Creates a notice that an element was changed. If the element is not registered
     * with the event queue, it is cloned to obtain the old version and put into the
     * element queue.
     */
    void elementChanged(Element el, Element oldVersion) {
        if (containsChanges(el)) {
            return;
        }
        if (changedElements == null) {
            changedElements = new HashMap<Element, Element>(17);
        }
        changedElements.put(el, oldVersion);
    }

    /**
     * Helper method that creates a notice that an element was changed. The passed
     * impl is asked to clone itself, if the element is not found in the changed
     * map already.
     */
    void elementChanged(ElementImpl impl) {
        Element el = impl.getElement();
        if (containsChanges(el)) {
            return;
        }
        if (changedElements == null) {
            changedElements = new HashMap<Element, Element>(17);
        }
        changedElements.put(el, impl.cloneSelf());
    }

    /**
     * Creates a notice that an element was created.
     */
    void elementCreated(Element el) {
        if (newElements == null) {
            newElements = new HashSet<Element>(17);
        }
        newElements.add(el);
    }

    /**
     * Creates a notice that the element was removed.
     */
    void elementRemoved(Element el) {
        if (newElements != null) {
            newElements.remove(el);
        }
        if (removedElements == null) {
            removedElements = new HashSet<Element>(17);
        }
        removedElements.add(el);
    }

    /**
     * Determines whether there's already a change notification for the passed element.
     * @return true, if the element was already reported as changed.
     */
    public boolean containsChanges(Element el) {
        if (newElements != null &&
                newElements.contains(el)) {
            return true;
        }
        if (changedElements != null &&
                changedElements.containsKey(el)) {
            return true;
        }
        return false;
    }

    private static class ElementChangeEvent {

        private ElementImpl impl;
        private PropertyChangeEvent evt;

        ElementChangeEvent(ElementImpl impl, PropertyChangeEvent evt) {
            this.impl = impl;
            this.evt = evt;
        }

        ElementImpl implentation() {
            return impl;
        }

        PropertyChangeEvent event() {
            return evt;
        }
    }

    /**
     * Records a property change event that occurs on a particular element. The change
     * is recorded into the element's queue.
     */
    synchronized void addPropertyChange(ElementImpl impl, PropertyChangeEvent evt) {
        ElementImpl source = impl;
        Collection<PropertyChangeEvent> queue = getQueue(source);
        queue.add(evt);
        if (eventQueue == null) {
            eventQueue = new LinkedList<ElementChangeEvent>();
        }
        eventQueue.add(new ElementChangeEvent(impl, evt));
    }

    /**
     * Master method that fires PropertyChange events from the queues collected so far.
     */
    void fireEvents() {
        synchronized (this) {
            if (firingEvents) {
                return;
            }
            firingEvents = true;
        }
        for (Collection<ElementChangeEvent> col = pollEventQueue(); col != null && !col.isEmpty(); col = pollEventQueue()) {
            for (Iterator<ElementChangeEvent> it = col.iterator(); it.hasNext();) {
                ElementChangeEvent ece = it.next();
                ElementImpl impl = ece.implentation();
                PropertyChangeEvent evt = ece.event();
                impl.firePropertyChangeEvent(evt);
            }
        }
        synchronized (this) {
            firingEvents = false;
        }
    }

    private void fireElementEvents(ElementImpl impl, Collection<PropertyChangeEvent> pceQueue) {
        for (Iterator<PropertyChangeEvent> it = pceQueue.iterator(); it.hasNext();) {
            PropertyChangeEvent evt = it.next();
            impl.firePropertyChangeEvent(evt);
        }
    }

    private synchronized Map pollEvents() {
        Map m = sourceMap;
        sourceMap = null;
        return m;
    }

    private synchronized Collection<ElementChangeEvent> pollEventQueue() {
        Collection<ElementChangeEvent> c = eventQueue;
        eventQueue = null;
        sourceMap = null;
        return c;
    }
    
    Map<ElementImpl, Collection<PropertyChangeEvent>> getSourceMap() {
        return sourceMap;
    }

    private synchronized Collection<PropertyChangeEvent> getQueue(ElementImpl source) {
        Collection<PropertyChangeEvent> c;

        if (sourceMap == null) {
            sourceMap = new HashMap<ElementImpl, Collection<PropertyChangeEvent>>(17);
            c = null;
        } else {
            c = sourceMap.get(source);
        }

        if (c != null) {
            return c;
        }
        c = new LinkedList<PropertyChangeEvent>();
        sourceMap.put(source, c);
        return c;
    }

    /**
     * Fixates change events - for each changed element, it creates a snapshot of it to
     * record its exact state at the time of unlock for the reference of listeners that
     * are contacted out of the locked exec region.
     */
    void fixupChanges() {
        if (changedElements == null) {
            return;
        }
        for (Iterator<Map.Entry<Element, Element>> it = changedElements.entrySet().iterator(); it.hasNext();) {
            Map.Entry<Element, Element> en = it.next();
            Element source = en.getKey();
            Element oldSnapshot = en.getValue();

            ElementImpl impl = source.getCookie(ElementImpl.class);
            // impl != null!
            Element newSnapshot = impl.cloneSelf();

            //en.setValue(new Object[] { oldSnapshot, newSnapshot }); //???
            en.setValue(newSnapshot);
        }
    }

    private void mergeChild(EventQueue other) {
        if (other.removedElements != null) {
            // discard any information about elements that were previously 
            // created. Discard all such elements from the removed set
            if (newElements != null) {
                Collection<Element> copy = new HashSet<Element>(other.removedElements);
                other.removedElements.removeAll(newElements);
                newElements.removeAll(copy);
            }
            if (removedElements == null) {
                removedElements = other.removedElements;
            } else {
                removedElements.addAll(other.removedElements);
            }
        }
        if (other.newElements != null) {
            if (newElements == null) {
                newElements = other.newElements;
            } else {
                newElements.addAll(other.newElements);
            }
        }
        if (other.changedElements != null) {
            if (newElements != null) {
                other.changedElements.keySet().removeAll(newElements);
            }
            if (changedElements != null) {
                other.changedElements.putAll(changedElements);
            }
            changedElements = other.changedElements;
        }
        if (other.sourceMap != null) {
            mergePropertyEventMaps(other.sourceMap);
        }
        if (other.eventQueue != null) {
            if (eventQueue == null) {
                eventQueue = other.eventQueue;
            } else {
                eventQueue.addAll(other.eventQueue);
            }
        }
    }

    /**
     * Merges the otherMap of property changes events into this one; the events are
     * merged so that they are appended at the end of the existing queue, if the
     * element's queue already exists, or the whole queue for an element is copied
     * into out map.
     */
    private synchronized void mergePropertyEventMaps(Map<ElementImpl, Collection<PropertyChangeEvent>> otherMap) {
        if (sourceMap == null) {
            sourceMap = otherMap;
            return;
        }
        for (Iterator<Map.Entry<ElementImpl, Collection<PropertyChangeEvent>>> otherIterator = otherMap.entrySet().iterator();
                otherIterator.hasNext();) {
            Map.Entry<ElementImpl, Collection<PropertyChangeEvent>> otherEntry = otherIterator.next();
            ElementImpl otherKey = otherEntry.getKey();
            Collection<PropertyChangeEvent> myQueue = sourceMap.get(otherKey);

            if (myQueue == null) {
                sourceMap.put(otherKey, otherEntry.getValue());
            } else {
                myQueue.addAll(otherEntry.getValue());
            }
        }
    }

    /**
     * Merges all information in this queue into the parent one, if there's any.
     * The operation is destructive for this instance.
     */
    void mergeToParent() {
        if (parent == null) {
            return;
        }
        parent.mergeChild(this);
    }

    final Map getChangedElements() {
        return this.changedElements;
    }

    final Set getCreatedElements() {
        return this.newElements;
    }

    final Set getRemovedElements() {
        return this.removedElements;
    }

    final EventQueue getParent() {
        return this.parent;
    }

    void clearSummary() {
        newElements = removedElements = null;
        changedElements = null;
    }

    public boolean isEmpty() {
        if (newElements != null && !newElements.isEmpty()) {
            return false;
        }
        if (changedElements != null && !changedElements.isEmpty()) {
            return false;
        }
        return removedElements == null || removedElements.isEmpty();
    }
}
