/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or sbbuild/licenses/cddl-gpl-2-cp.txt. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * sbbuild/licenses/cddl-gpl-2-cp.txt.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 2004-2007 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */


package org.starbeans.modules.seqformat.model;

import java.beans.PropertyChangeEvent;
import java.util.*;
import org.starbeans.modules.seqformat.source.*;

/**
 *
 * @author  sdedic
 * @version 
 */
abstract public class ObjectCollection extends IndexedPropertyBase implements ElementOrder {
    /**
     * Threshold for dump identity comparisons. If the avg number of compare ops *2 is
     * more than this value, faster hashing algorithm is used for discovering identities.
     * The value should balance the tradeoff between creating new objects and the comparisons.
     */
    public static final int SMALL_THRESHOLD = 500;
    
    /** Reference to the collection's container binding - manages contained objects.
     */
    protected Binding.Container       containerBinding;
    
    /** Interface used for broadcasting events
     */
    protected ElementEvents           events;
    
    /** The actual contents of the collection.
     */
    protected Element[]               elements;
    
    protected ElementCreator          creator;
    
    public ObjectCollection(Binding.Container container,
        ElementCreator creator, ElementEvents events, String propertyName) {
        super(propertyName);
        this.events = events;
        this.creator = creator;
        this.containerBinding = container;
    }
    
    public Element[] getElements() {
        if (elements == null)
            return createEmpty(0);
        return elements;
    }
    
    /** Attempts to find appropriate positions(s) for inserting the new element(s).
     * The method is expected to return array of elements that should precede the 
     * matching elements being inserted. For insertion at the beginning of the collection,
     * the method should use a special tag value, Positioner.FIRST. If the method returns
     * null, elements are appended at the end of the collection.
     */
    protected abstract Element[] findPositions(Element[] elements);
    
    /**
     * Creates a copy of the passed element using local implementation and creates an
     * empty (uninitialized) binding for it.
     */
    protected abstract ElementImpl createElement(Element parent);
    
    /** The method is called for creation of array of exactly the specific element
     * type so that typecasts in collection's clients succeed.
     */
    protected abstract Element[] createEmpty(int size);
    
    /** Copies an element over to this collection, creating a local implementation of
     * the element.
     */    
    protected Element copyElement(Element model) {
        ElementImpl impl = createElement((Element)events.getEventSource());
        try {
            impl.createFromModel(model);
        } catch (SrcException ex) {
            //This can't be an expected exception
            IllegalStateException e = new IllegalStateException("Exception thrown during cloning of element"); //NOI18N
            e.initCause(ex);
            throw e;
        }
        return impl.getElement();
    }
    
    /** Adds the specified members to the collection. The implementation uses
     * Finder to find appropriate place(s) in the collection to insert the new
     * members.
     */
    public void addMembers(Element[] els) throws SrcException {
        Element[] insertingEls = new Element[els.length];
        Element[] elements = getElements();
        
        for (int i = 0; i < els.length; i++) {
            insertingEls[i] = copyElement(els[i]);
        }

        Element[] precedings = findPositions(insertingEls);
        Element[] newElems = createEmpty(insertingEls.length + elements.length);

        int[] insertIndices = new int[newElems.length];
        
        // If an element wants to be inserted after another one, that another one
        // needs to be inserted first (if it is also new one)
        Map<Element,Object> m = new HashMap<Element,Object>(precedings.length * 4 / 3);
        for (int i = 0; i < precedings.length; i++) {
            // throw exception if prev is not a member of this class.
            Object o = m.get(precedings[i]);
            if (o != null) {
                Collection<Object> c;
                if (!(o instanceof Collection)) {
                    c = new LinkedList<Object>();
                    c.add(o);
                    m.put(precedings[i], Object.class.cast(c));
                } else {
                    c = (Collection)o;
                }
                c.add(Object.class.cast(insertingEls[i]));
            } else {
                m.put(precedings[i], Object.class.cast(insertingEls[i]));
            }
        }

        Element prevElem = Positioner.FIRST;
        int newPos = 0, oldPos = 0;
        int inserted = 0;

        for (oldPos = 0; oldPos <= elements.length; oldPos++) {
            Object dependent = m.get(prevElem);

            if (dependent != null) {
                //Util.log("inserting after global item #" + oldGlobalPos); // NOI18N
                LinkedList<Object> toInsert = new LinkedList<Object>();

                do {
                    if (dependent != null) {
                        if (dependent instanceof Collection)
                            toInsert.addAll((Collection)dependent);
                        else
                            toInsert.add(dependent);
                    }

                    Element newEl = (Element)toInsert.removeFirst();
                    insertIndices[inserted] = newPos;
                    newElems[newPos] = newEl;
                    inserted++;
                    newPos++;

                    dependent = m.get(newEl);
                } while (!toInsert.isEmpty() || dependent != null);
                // inserted everything chained after prevElem
            }
            
            if (oldPos >= elements.length)
                break;
            
            prevElem = elements[oldPos];
            newElems[newPos] = elements[oldPos];
            newPos++;
        }
        if (inserted < insertingEls.length)
            throw new IllegalArgumentException("Inconsistent insertion strategy"); // NOI18N
        
        Change chng = new Change();
        chng.insertIndices = insertIndices;
        chng.inserted = Arrays.asList(insertingEls);
        chng.phaseCount = 1;
        chng.sizeDiff = insertingEls.length;

        doSetMembers(newElems, chng);
    }
    
    /**
     * Removes exactly the members specified in toRemove. Elements are matched against
     * the set of current members based on object identity; if an element is not found in
     * the current set, the implementation may throw IllegalArgumentException.
     */
    public void removeMembers(Element[] els) throws SrcException {
        Element[] elements = getElements();
        int[] removeIndices = createIdentityMap(els);

        // sanity check:
        for (int i = 0; i < removeIndices.length; i++) {
            if (removeIndices[i] == -1)
                throw new SrcException("Element does not exist"); // NOI18N
        }

        Element[] newElems = createEmpty(elements.length - els.length);
        //Util.log("newElems = " + newElems.length); // NOI18N
        int[] offsets = new int[elements.length];
        int newPos = 0;
        int removed = 0;
        int nextRemove = removeIndices[0];
        for (int i = 0; i < elements.length; i++) {
            if (i == nextRemove) {
                offsets[i] = -i -1;
                if (++removed < removeIndices.length)
                    nextRemove = removeIndices[removed];
                else
                    nextRemove = -1;
            } else {
                offsets[i] = newPos - i;
                newElems[newPos] = elements[i];
                newPos++;
            }
        }
        Change ch = new Change();
        ch.removed = Arrays.asList(els);
        ch.removeIndices = removeIndices;
        ch.offsets = offsets;
        ch.sizeDiff = - els.length;
        ch.phaseCount = 1;
        
        // elements & positions will be updated from inside doSetMembers.
        doSetMembers(newElems, ch);
    }
    
    public void updateMembers(Element[] members, int[] map) {
        Collection newMembers = null;
        
        if (this.elements != null) {
            Collection removedMembers;
            Element[] els = getElements();

            if (els != null) {
                if (map == null)
                    map = createIdentityMap(members);

                Change chng = computeChanges2(els, members, map);
                if (chng.phaseCount == 0)
                    return;
                events.addPropertyChange(createPropertyEvent(chng, events.getEventSource(),
                    els, members));
                newMembers = chng.inserted;
                removedMembers = chng.removed;
            } else {
                newMembers = Arrays.asList(members);
                removedMembers = null;
            }
            ElementImpl parent = events.getElementImpl();
            boolean fire = parent.isValid();
            if (fire) {
                if (removedMembers != null && !removedMembers.isEmpty()) {
                    for (Iterator it = removedMembers.iterator(); it.hasNext(); ) {
                        Element el = (Element)it.next();
                        ElementImpl impl = getElementImpl(el);
                        impl.notifyRemove();
                    }
                }
                if (newMembers != null && !newMembers.isEmpty()) {
                    for (Iterator it = newMembers.iterator(); it.hasNext(); ) {
                        Element el = (Element)it.next();
                        ElementImpl impl = getElementImpl(el);
                        impl.notifyCreate();
                    }
                }
            }
        } else if (!events.getElementImpl().isCreated() && members.length > 0) {
            ElementImpl parent = events.getElementImpl();
            boolean fire = parent.isValid();
            for (int i = 0; i < members.length; i++) {
                Element el = members[i];
                ElementImpl impl = getElementImpl(el);
                impl.notifyCreate();
            }
        }
        this.elements = members;
    }
    
    public void setMembers(Element[] members) throws SrcException {
        Change chng;
        
        Element[] els = getElements();
        Element[] newEls = createEmpty(members.length);
        int[] map = createIdentityMap(members);
        
        int memberPos = 0;
        boolean copied = false;
        for (int i = 0; i < map.length; i++) {
            if (map[i] == -1) {
                newEls[i] = copyElement(members[i]);
                copied = true;
            } else {
                newEls[i] = members[i];
            }
        }
        
        chng = computeChanges2(els, newEls, map);
        
        doSetMembers(newEls, chng);
    }
    
    protected Binding getBinding(Object o) {
        return getBinding((Element)o);
    }
    
    protected Binding getBinding(Element el) {
        ElementImpl impl = getElementImpl(el);
        return impl == null ? null : impl.getBinding();
    }
    
    public ElementImpl getElementImpl(Element el) {
        return el.getCookie(ElementImpl.class);
    }
    
    public void changeMembers(Element[] members, int operation) throws SrcException {
         switch (operation) {
            case Element.Impl.ADD:
                addMembers(members);
                break;
            case Element.Impl.SET:
                setMembers(members);
                break;
            case Element.Impl.REMOVE:
                removeMembers(members);
                break;
        }
    }
    
    /**
     * The actual worker method, that rewrites the collection's content and the associated
     * Bindings.
     */
    protected void doSetMembers(Element[] newMembers, Change chg) throws SrcException {
        if (chg.phaseCount == 0)
            return;

        Element[] elements = getElements();
        boolean deferredCreation = events.getElementImpl().isCreated();

        MultiPropertyChangeEvent evt;

        if (!deferredCreation) {
            if (events.getElementImpl().isConstrained() && (chg.removed != null)) {
                for (Iterator it = chg.removed.iterator(); it.hasNext(); ) {
                    getElementImpl((Element)it.next()).checkRemove();
                }
            }
            evt = createPropertyEvent(chg,
            events.getEventSource(), getElements(), newMembers);
            fireVetoableChange(evt);
            containerBinding.changeMembers(evt);

            if (chg.inserted != null)
                for (Iterator it = chg.inserted.iterator(); it.hasNext(); ) {
                    getElementImpl((Element)it.next()).notifyCreate();
                }
            if (chg.removed != null)
                for (Iterator it = chg.removed.iterator(); it.hasNext(); ) {
                    getElementImpl((Element)it.next()).notifyRemove();
                }
            addPropertyChange(evt);
        }

        this.elements = newMembers;
    }
    
    protected void fireVetoableChange(PropertyChangeEvent evt) throws SrcException {
        events.fireVetoableChange(evt);
    }
    
    protected void addPropertyChange(PropertyChangeEvent evt) {
        events.addPropertyChange(evt);
    }
    
    protected int[] createIdentityMap(Element[] els) {
        Element[] olds = getElements();
        return super.createIdentityMap(els, olds);
    }

}
