/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or sbbuild/licenses/cddl-gpl-2-cp.txt. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * sbbuild/licenses/cddl-gpl-2-cp.txt.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 2004-2007 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */

package org.starbeans.modules.seqformat.model;

import java.beans.PropertyChangeEvent;
import org.openide.nodes.Node;
import org.starbeans.modules.seqformat.source.*;

/**
 *
 * @author  sdedic, George Waldon
 */
public interface SeqModel extends ElementCreator {
    /** Acquires a read lock on the model. There can be no changes until the
     * lock is released. The same thread that has acquired the write lock may choose
     * to acquire a read lock as well. The opposite is true as well, as long there's only
     * a single reader - the current thread.
     */
    public void readLock();
    
    /** Releases the read lock.
     */
    public void releaseReadLock();
    
    /**
     * Returns true, if the model is write-locked currently. However, this method does not
     * guarantee, that, should it return false, writeLock called subsequently will not
     * block.
     */
    public boolean isWriteLocked();

    /**
     * Adds a listener that will be notified before the commit phase is over - that is
     * before the write lock is released. Please, use this type of listener with extreme
     * caution and do not make global changes from it and avoid event firing if possible.
     */
    public void addPreCommitListener(CommitListener l);

    /**
     * Removes pre-commit listener from the model.
     */
    public void removePreCommitListener(CommitListener l);
    
    /**
     * Attaches a CommitListener to the model. Post-commit listener gets a list of
     * changes after the lock is released. The listener can make further operations on the model,
     * risking that further events may be delayed.
     */
    public void addPostCommitListener(CommitListener l);

    /**
     * Removes the post-commit listener from the model.
     */
    public void removePostCommitListener(CommitListener l);
    
    /**
     * Executes an atomic action.
     */
    public void runAtomic(Runnable r) throws SrcException;

    /**
     */
    public void commitChanges();
    
    /**
     * Acquires a write lock on the model and returns an opaque handle. There may be
     * at most *one* write lock active at any given time. Other threads asking for
     * a lock will be blocked.
     * @return handle for the write lock.
     */
    public Object writeLock();

    /**
     * Tries to acquire a write lock; if someone is holding the lock right now,
     * the method returns immediately null. You can use this method as a nonblocking
     * attempt to acquire the write lock.
     *
     * @return handle for the write lock, or null if the model is currently locked.
     */
    public Object tryWriteLock();
    
    /** Releases the write lock from the model. Requires the handle returned by
     * the prior call to writeLock(). If the handle is invalid, the method does not
     * unlock and throws IllegalArgumentException.
     * @param handle for unlocking.
     * @throws SourceException, if something goes wrong during the commit phase.
     * @throws IllegalArgumentException if the lock handle is invalid.
     */
    public void releaseWriteLock(Object handle) throws SrcException;

    /** Environmental interface that helps interaction between the model and its
     * environment. Virtually any services required by the model from outside should
     * be performed, or obtained from, this interface.
     */
    public static interface Env {
        /**
         * Callback function that will create a binding for an Element. The method can
         * return null, in which case it can be invoked later for that particular Impl,
         * or the Element will be assigned NullBinding, eventually.
         * @param impl implementation instance of the model's Element.
         * @return binding between the Element and its storage.
         */
        public BindingFactory getBindingFactory();
        
        /** Factory, that "wraps" element with the abstract layer fom the openide, optionally
         * providing some levels of indirection to an element.
         */
        public WrapperFactory getWrapperFactory();
              
        /**
         * Since the model does not care for cookies, it needs to be supported by the
         * environment if somebody asks for a cookie on the Element.
         */
        public <T extends Node.Cookie> T findElementCookie(Element el, Class<T> requestedClass);
                
    }

    public interface Updater extends SeqModel, ElementProperties {
        public void updateMembers(Element target, String propertyName, 
            Element[] els, int[] orderIndices, int[] optionalMap);

        /**
          * Updates members in a specified container of a target element. In general,
          * container types correspond with indexed property names.
          */
        public void updateMemberOrder(Element target, String containerType,
            Element[] orderedMembers);
        
        public Binding getElementBinding(Element target);
        
        /**
         * Special method that explicitly makes an element active. Inactive elements do not
         * fire property changes etc. Elements become automatically active, when inserted
         * into active container or when the container is activated.
         */
        public void activate(Element target);
        
        /**
         * Makes the whole model contents invalid.
         */
        public void invalidateModel(SrcElement el);

        /**
         * Runs the supplied runnable object in a special execution mode that can disable
         * property veto checks
         */
        public boolean runUpdate(Runnable r, boolean disableVetos) throws SrcException;

        /**
         * Fires some property change event on behalf of some element managed by
         * this model. Note that this <B>is a hack</B> method provided mostly
         * for PROP_COOKIES property change as most of the properties are
         * provided by the model's environment.
         */
        public void firePropertyChange(Element el, PropertyChangeEvent evt);
    }
}
