/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or sbbuild/licenses/cddl-gpl-2-cp.txt. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * sbbuild/licenses/cddl-gpl-2-cp.txt.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 2004-2007 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */


package org.starbeans.modules.seqformat.parser;

import java.io.IOException;
import java.io.Reader;
import java.util.*;
import javax.swing.event.ChangeListener;
import javax.swing.event.ChangeEvent;
import org.openide.filesystems.FileObject;
import org.openide.text.CloneableEditorSupport;
import org.starbeans.modules.seqformat.*;


/**
 *
 * @author  sdedic, George Waldon
 * @version
 */
public class ParseSourceRequest implements ParsableObjectRequest, ErrConsumer {
    public static final int     STATE_WAITING = 0;
    public static final int     STATE_READING = 1;
    public static final int     STATE_CANCELLED = 2;
    public static final int     STATE_ANALYSIS = 3;
    public static final int     STATE_UPDATING = 4;
    public static final int     STATE_COMPLETED = 10;

    SeqParser.Env           environment;
    CloneableEditorSupport  editSupp;
    ChangeListener          listener;
    int                     state;
    boolean                 valid;
    int                     syntaxErrors;
    ElementFactory          builder;
    private List<ParserMessage> errors = null;

    public ParseSourceRequest() {
        state = STATE_WAITING;
        valid = true;
    }

    /* Setter used by ParsingSupport.Processor */
    public void setEnvironment(SeqParser.Env env) {
        environment = env;
    }
    
    /* Public getter for the ParserEngine */
    public SeqParser.Env getEnvironment() {
        return environment;
    }

    /* Setters used by ParsingSupport.Processor */
    public void setEditorSupport(CloneableEditorSupport editor) {
        editSupp = editor;
    }

    public synchronized void addChangeListener(ChangeListener l) throws TooManyListenersException {
        if (listener != null)
            throw new TooManyListenersException();
        listener = l;
    }

    public synchronized void removeChangeListener(ChangeListener l) {
        if (listener == l)
            listener = null;
    }

    /**
     * Notifies the request that the source text has been changed. This causes
     * cancellation of the request in some cases.
     */
    public void sourceChanged() {
        if (state == STATE_READING) {
            // cancel the request only in reading state; if the text is already
            // read, the request can still be cariied out.
            valid = false;
        }
    }

    public void modelChanged() {
        if (state != STATE_WAITING && state != STATE_COMPLETED) {
            valid = false;
        }
    }

    public void setSyntaxErrors(int errors) {
        this.syntaxErrors = errors;
    }

    public int getSyntaxErrors() {
        return syntaxErrors;
    }

    public void setSemanticErrors(int errors) {
        // ignore - for now.
    }
    
    /**
     * DocumentModelBuilder actually carries out tasks associated with the Factory.
     */
    public ElementFactory getFactory() {
        if (builder == null)
            builder = createBuilder(editSupp);
        return builder;
    }
    
    protected ElementFactory createBuilder(CloneableEditorSupport supp) {
        BuilderFactory factory = ((ParserEnv)environment).findCookie(null,BuilderFactory.class);
        return factory.createBuilder();
    }

    public DocumentModelUpdater getUpdater() {
        DocumentModelUpdater result = null;
        ElementFactory efbuilder = getFactory();
        if (efbuilder instanceof DocumentModelUpdater) 
            result = (DocumentModelUpdater)efbuilder;
        return result;
    }

    public void notifyReschedule() {
        // clear old data.
        builder = null;
        enterState(STATE_WAITING);
    }

    protected void enterState(int state) {
        this.state = state;
        if (listener != null)
            listener.stateChanged(new ChangeEvent(this));
    }

    /** Get the source for the parser
     *
     * The method is implemented so that it reads the whole contents from the environment's
     * Reader and returns the resulting char array.
     */
    public char[] getSource() throws IOException { 
        // unsynchronized; the exact sequence really does not matter here too much;
        // the real problem arises that AFTER the contents is read.
        valid = true;
        enterState(STATE_READING);

        Reader r = environment.getSourceText();
        char[] buf=Util.readContents(r);
        ElementFactory efbuilder = getFactory();
        if (efbuilder instanceof DocModelBuilder) {
            ((DocModelBuilder)builder).setContent(buf, editSupp.isDocumentLoaded());
        }
        return buf;
    }

    /**
     * Returns true, IF the request was not invalidated by the ParserSupport,
     * or because of change in the document while parsing.
     */
    public boolean isValid() {
        return this.valid;
    }

    public boolean needsProcessing() {
        // always needs to process (?)
        //return this.valid;
        //GW yes! This prevents further parsing in the ParserEngine and brings infinite 
        // repost of the request; see ParsingSupport:612     addRequest(this, priority);... 
        return true;
    }

    public void notifyStart() {
    }

    public void notifyComplete() {
        // safepoint: everything was processed.
        enterState(STATE_COMPLETED);
        if (errors==null)
            errors=new ArrayList<ParserMessage>();
    }

    /** Return name+extension */
    public String getSourceName() {
        if (environment != null )
            return environment.getSourceName();
        else 
            return null;
    }
 
    /** Return file object */
    public FileObject getSourceFile() {
        if (environment != null )
            return environment.getSourceFile();
        else 
            return null;
    }
    
    public ErrConsumer getErrConsumer() {
        return this;
    }

    public void pushError(Object severity, FileObject errorFile, int line, int column, int length, String message) {
        Integer lineInt;
        
        // errors in different file
        if (errorFile != null && !errorFile.getNameExt().equals(getSourceName()))
            return;
        
        ParserMessageImpl newErr =
            new ParserMessageImpl(severity,line,column,length,message);
        if (errors == null) {
            errors = new ArrayList<ParserMessage>(50);
        }
        errors.add(newErr);
    }

    /** @return the set of errors encountered while processing this
     * request */
    public Collection<ParserMessage> getMessages() {
        return errors;
    }
}

