/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or sbbuild/licenses/cddl-gpl-2-cp.txt. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * sbbuild/licenses/cddl-gpl-2-cp.txt.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 2004-2007 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */


package org.starbeans.modules.seqformat.parser;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import org.openide.text.Annotation;
import org.openide.text.Line;
import org.openide.text.Annotatable;
import org.openide.text.Line.Set;


/**
 *
 * @author  th125165, George Waldon
 */
class ParserAnnotation extends Annotation implements PropertyChangeListener {
    
    private Object severity;
    private final String error;
    private final int line,column,length;
    private int state;
    private Line docline;
    private ParserAnnotation chained;
    
    private static final int STATE_NEW=1;
    private static final int STATE_ATTACHED = 2;
    private static final int STATE_DETACHED = 3;
    
    /** Creates a new instance of ParserAnnotation.
     *  When the length is given, the line part is counted from
     *  the column to column + length
     */
    ParserAnnotation(Object s,int l, int c, int len, String err) {
        severity=s;
        line=l;
        column=c;
        length = len;
        error = err;
        state = STATE_NEW;
    }
    
    public String getAnnotationType() {
        if(severity==ParserMessage.ERROR)
            return "org-starbeans-modules-seqformat-parser_error_annotation";  //NOI18N
        else
            return "org-starbeans-modules-seqformat-parser_warning_annotation";  //NOI18N
    }
    
    public String getShortDescription() {
        // Localize this with NbBundle:
        if (chained!=null)
            return error+"\n\n"+chained.getShortDescription();
        return error;
    }
    
    int getLine() {
        return line;
    }
    
    int getColumn() {
        return column;
    }
    
    int getLength() {
        return length;
    }
    
    String getError() {
        return error;
    }
    
    void chain(ParserAnnotation anno) {
        if (chained!=null)
            chained.chain(anno);
        else
            chained=anno;
    }
    
    private int getState() {
        return state;
    }
    
    protected void notifyAttached(final Annotatable toAnno) {
        super.notifyAttached(toAnno);
        docline.addPropertyChangeListener(this);
        state = STATE_ATTACHED;
    }
    
    protected void notifyDetached(Annotatable fromAnno) {
        super.notifyDetached(fromAnno);
        docline.removePropertyChangeListener(this);
        state=STATE_DETACHED;
    }
    
    public boolean equals(Object obj) {
        boolean eq=shallowEquals(obj);
        
        if (!eq)
            return false;
        if (chained!=null)
            return chained.equals(((ParserAnnotation)obj).chained);
        return true;
    }
           
    private boolean shallowEquals(Object obj) {
        if (obj instanceof ParserAnnotation) {
            ParserAnnotation ann=(ParserAnnotation)obj;
            
            if (this==obj)
                return true;
            if (line!=ann.getLine())
                return false;
            if (column!=ann.getColumn())
                return false;
            if (!error.equals(ann.getError()))
                return false;
            if (getState()==STATE_DETACHED || ann.getState()==STATE_DETACHED)
                return false;
            return true;
        }
        return false;
    }
    
    public void attachToLineSet(Set lines) {
        char string[];
        int start,end;
        Line.Part part;

        try {
            docline=lines.getCurrent(line-1);
        } catch (IndexOutOfBoundsException ex) {
            // the document has been changed and the line is deleted
            return;
        }
        String annTxt = docline.getText();
        if (annTxt == null) return; // document is already closed
        string = annTxt.toCharArray();
        
        start=0;
        end=string.length-1;
        while (start<=end && string[start]<=' ') {
            start++;
        }
        while (start<=end && string[end]<=' ') {
            end--;
        }
        if (start<=end)
            part=docline.createPart(start,end-start+1);
        else
            part=docline.createPart(0,string.length);
        attach(part);
    }
    
    public void propertyChange(PropertyChangeEvent ev) {
        String type = ev.getPropertyName();
        
        // User edited the line, assume error should be cleared.
        if (type == null || type == Annotatable.PROP_TEXT) {
            detach();
        }
    }
}
