/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or sbbuild/licenses/cddl-gpl-2-cp.txt. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * sbbuild/licenses/cddl-gpl-2-cp.txt.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 2004-2007 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */


package org.starbeans.modules.seqformat.parser;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.lang.ref.Reference;
import java.lang.ref.ReferenceQueue;
import java.lang.ref.WeakReference;
import java.util.*;
import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.event.ChangeListener;
import javax.swing.event.ChangeEvent;
import org.openide.nodes.Node;
import org.openide.text.CloneableEditorSupport;
import org.openide.util.RequestProcessor;
import org.openide.util.Task;
import org.openide.util.Utilities;
import org.starbeans.modules.seqformat.model.ElementImpl;
import org.starbeans.modules.seqformat.model.SeqModel;
import org.starbeans.modules.seqformat.model.CommitListener;
import org.starbeans.modules.seqformat.srcgen.DocBinding;
import org.starbeans.modules.seqformat.source.*;
import org.starbeans.modules.seqformat.SeqDataObject;
import org.starbeans.modules.seqformat.Util;

/**
 * This class serves as a default implementation of SeqParser interface and coordinates
 * one-way interaction between a SeqModel and the underlying StyledDocument. The other
 * way is implemented through various bindings attached to individual model elements.
 * The parser is responsible for maintaining `dirty' status for the model data, with respect
 * to changes made to the document, parsing when the client asks to parse and updating
 * the model contents when appropriate.
 *
 * @author  sdedic, George Waldon
 * @version 
 */
public class ParsingSupport implements SeqParser {
    private static final Logger LOGGER = Logger.getLogger(ParsingSupport.class.getName());
    
    private volatile PropertyChangeSupport  propSupport;
    
    public static final String PROP_STATUS = "status"; // NOI18N

    
    /** Initially true. Set to false when the Support is no longer valid.
     */
    private boolean valid;
    
    /**
     * True during the model update operation. When true, the refImplementation
     * cannot be cleared, otherwise the updater thread may lock up.
     */
    private boolean updating;

    /** When true indicates that the source used for the current or last parsing
     * is still unchanged.
     */
    private boolean clean;
    
    private Env parsingEnv;
    
    /** IOException saved (?) from the last parsing process. If no change is
     * reported to the underlying document, the exception is immediately
     * thrown without attempting to parse.
     */
    private SrcException     savedException;
    
    /**
     * SrcElement that is bound to the proxy. The proxy, if needed constructs the
     * real Impl. The element, once constructed, is not freed unless the ParsingSupport
     * dies as well.
     */
    private SrcElement      src;
    
    /**
     * The request object that is filed to the parsing queue. Since the request has not
     * yet touched by the RequestProcessor, it is safe to change parameters for it.
     */
    private Processor       currentRequest;
    
    /** The request that is currently running, null if none. This request is currently
     * being serviced by the RequestProcessor.
     */
    private Processor       runningRequest;

    /**
     * Language model interface.
     */
    SeqModel           model;

    /**
     * Extended interface for updating the language model.
     */
    SeqModel.Updater   updater;

    /**
     * Proxy for ScrElement's implementation, can handle some requests.
     */
    private SrcImplProxy  proxy;
    
    /**
     * A Weak/Soft reference to the real implementation.
     */
    private Reference       refImplementation;
    
    /**
     * Parsing status of the source - it is one of SrcElement.STATUS_* symbolic
     * constatnts.
     */
    private int             status;
    
    /**
     * Collection of ChangeListeners registered with this Support.
     */
    volatile Collection<ChangeListener> changeList;
    
    DocBinding              docBinding;
    
    SeqDataObject           sdo;
    
    /**
     * Creates a ParsingSupport and initializes it with an environment, binding,
     * updater interface and the model.
     */
    public ParsingSupport(Env parsingEnv, SeqDataObject sdo, DocBinding docBinding, SeqModel.Updater updater, SeqModel model) {
        this.parsingEnv = parsingEnv;
        this.sdo = sdo;
        this.docBinding = docBinding;
        this.updater = updater; // DefaultSeqModel
        this.model = model; // DefaultSeqModel
        this.valid = true;
    }
    
    public void addChangeListener(ChangeListener l) {
        if (changeList == null) {
            synchronized (this) {
                if (changeList == null) 
                    changeList = new LinkedList<ChangeListener>();
            }
        }
        synchronized (changeList) {
            changeList.add(l);
        }
    }
    
    public void removeChangeListener(ChangeListener l) {
        if (changeList == null)
            return;
        synchronized (changeList) {
            changeList.remove(l);
        }
    }

    /**
     * Adds a property change listener to this object.
     */
    public void addPropertyChangeListener(PropertyChangeListener l) {
        if (propSupport == null) {
            synchronized (this) {
                if (propSupport == null)
                    propSupport = new PropertyChangeSupport(this);
            }
        }
        propSupport.addPropertyChangeListener(l);
    }
    
    /**
     * Removes the change listener.
     */
    public void removePropertyChangeListener(PropertyChangeListener l) {
        if (propSupport == null)
            return;
        propSupport.removePropertyChangeListener(l);
    }
    
    /** Schedules a refresh/reparse of the associated document. The request is given
     * the passed priority. If there are some errors during parsing and acceptErrors
     * is false, the hierarchy is <B>not</B> updated. After the parsing task finishes,
     * the hiearchy will reflect the state of the document at the time parse() was called,
     * or a more recent state, if there were multiple parse() requests before the task
     * was processed.
     * @return Task that can be tested for completion or can be listened on
     */
    public Task parse(int priority, boolean ignoreClean, boolean acceptErrors) {
        
        LOGGER.log(Level.FINE, "Entering Parse", new Exception()); //NOI18N
        ParsableObjectRequest req = new ParseSourceRequest();
        Processor immediate;
        SrcElement.Impl i = null;
        // #20100 - order locks properly: getEditorSupport() is likely to 
        // access/lock some CookieSet, and cookies tend to call/synchronize
        // on the parser support.
        CloneableEditorSupport editSupport = docBinding.getEditorSupport();
        synchronized (this) {
            Util.log(ParsingSupport.class,"Got parse request, prio = " + priority + " ignoreClean = " + ignoreClean + " acceptErrs = " + acceptErrors); // NOI18N
            Util.log(ParsingSupport.class,"Data = " + getSourceImpl() + " clean = " + this.clean); // NOI18N
            Util.log(ParsingSupport.class,"Parsing task = " + currentRequest + "/" + runningRequest); // NOI18N 
            
            if (currentRequest != null) {
                // if THIS is the parsing thread, then -- well, it's bad. 
                // temporal resolution: take out the request IMMEDIATELY.
                if (PARSING_RP.isRequestProcessorThread()) {
                    Util.log(ParsingSupport.class,"Running in parsing thread!"); // NOI18N
                    immediate = currentRequest;
                } else {                
                    currentRequest.enableErrors(acceptErrors);
                    currentRequest.setPriority(priority);
                    Util.log(ParsingSupport.class,"Running in parsing thread!"); // NOI18N
                    return currentRequest.getClientTask();
                }
            } else {
                i = getSourceImpl();
                if (i != null && this.clean && !ignoreClean) {
                    Util.log(ParsingSupport.class,"Returning finished task"); // NOI18N
                    return new FinishedTask(i);
                }
                Processor proc = new Processor(priority, parsingEnv, editSupport,req);
                proc.enableErrors(acceptErrors);
                if (PARSING_RP != null && PARSING_RP.isRequestProcessorThread()) {
                    immediate = proc;
                } else {
                    addRequest(proc, priority);
                    return proc.getClientTask();
                }
            }
        }
        immediate.run();
        i = getSourceImpl();
        return new FinishedTask(i); 
    }  
    
    public synchronized Task getCurrentTask() {
        if (currentRequest!=null)
            return currentRequest.getClientTask();
        return new FinishedTask(null);
    }

    private static class FinishedTask extends Task {
        private Object hook;
        
        public FinishedTask(Object o) {
            super(null);
            hook = o;
        }
    }
    
    public void fireElementPropertyChange(Element source, PropertyChangeEvent evt) {
        if (source == getSource()) {
            proxy.propertyChange(evt);
        } else {
            updater.firePropertyChange(source, evt);
        }
    }
    
    /**
     * Returns the parser engine used by this Support.
     */
    public ParserEngine getParserEngine() {
        return sdo.getCookie(ParserEngine.class);
    }

    /**
     * Prepares data from the source. If some data already
     * exists, they are not refreshed, and the Task is returned 
     * in already finished state.
     * @return task object that is set to completed after the data is prepared.
     */
    public Task prepare() {
        return parse(PRIORITY_BACKGROUND, false, false);
    }
    
    /** Inquire about the validity of this parser.
     */
    public boolean isValid() {
        return valid;
    }

    /**
     * Invalidates contents of the parser support. This method should be called
     * when the original data source is discarded. The parser support is not
     * expecting to be valid ever again after this call.
     */
    public void invalidate() {
        SrcElement.Impl impl;
        synchronized (this) {
            if (!valid)
                return;
            impl = getSourceImpl();
            if (impl == null)
                return;
            valid = false;
        }
        updater.invalidateModel(getSource());
        synchronized (this) {
            if (!updating)
                refImplementation = null;
        }
        changeStatus(SrcElement.STATUS_NOT);
    }
        
    
    private void fireParsingEvent(final Collection<ParserMessage> messages) {
        final Object hook = getSourceImpl();
            
        EVENT_RP.post(new Runnable() {
                public void run() {
                    Parsing.fireEvent(sdo, hook, messages);
                }
            });
    }
    
    protected void fireStateChange() {
        if (changeList == null)
            return;
        Collection<ChangeListener> copy;
        
        synchronized (changeList) {
            if (changeList.isEmpty())
                return;
            copy = new ArrayList<ChangeListener>(changeList);
        }
        ChangeEvent e = new ChangeEvent(this);
        
        for (Iterator<ChangeListener> it = copy.iterator(); it.hasNext(); ) {
            try {
                it.next().stateChanged(e);
            } catch (RuntimeException ex) {
                LOGGER.log(Level.INFO,"",ex); //NOI18N
            }
        }
    }

    private SrcElement createSourceProxy() {
        proxy = sdo.getCookie(SrcImplProxy.class); 
        proxy.attachSupport(this);
        return parsingEnv.getLocalSrcElement(proxy);     
    }
      
    public SrcElement getSource() {
        if (this.src != null)
            return src;
        synchronized (this) {
            if (src != null)
                return src;
            return src = createSourceProxy();
        }
    }
    
    public SeqModel getModel() {
        return this.model;
    }
    
    public SrcElement.Impl getSourceImpl() {
        Object impl = refImplementation == null ? null : refImplementation.get();
        return (SrcElement.Impl)impl;
    }
    
    public SrcElement.Impl findSourceImpl() throws SrcException {
        synchronized (this) {
            SrcElement.Impl impl = getSourceImpl();
            if (impl != null)
                return impl;
            Util.log("impl = null"); // NOI18N
            if (savedException != null)
                throw savedException;
        }
        throw new SrcException("Cannot acquire source"); // NOI18N
    }
    
    /**
     * Invalidate the data.
     */
    public void sourceChanged() {
        Processor req = currentRequest;        
        clean = false;
        if (req != null)
            req.sourceChanged();
    }
    
    /**
     * Retrieves the Throwable that caused the parsing to abort. This will be
     * mainly an IOException.
     * @return Throwable that aborted the last most recent parse request.
     */
    public SrcException getErrorCause() {
        return savedException;
    }

    
    <T extends Node.Cookie> T  findCookieForSource(Class<T> type) {
        if (src == null)
            return null;
        return parsingEnv.findCookie(getSource(), type);
    }
    
    /**
     * Retrieves the parsing status of the source. The value is one of the symbolic
     * SrcElement.STATUS_* constants.
     */
    public int getStatus() {
        int s = status;
        if (s != SrcElement.STATUS_OK)
            return s;
        SrcElement.Impl impl = getSourceImpl();
        if (impl == null)
            changeStatus(SrcElement.STATUS_NOT);
        return status;
    }
        
    protected void changeStatus(int newStatus) {
        int oldStatus = status;
        status = newStatus;
        if (propSupport != null && propSupport.hasListeners(null))
            propSupport.firePropertyChange(PROP_STATUS, oldStatus, newStatus);
        fireStateChange();
    }


    //--------------------------------------------------------------------------------//
    
    /**
     * Called from within the RequestProcessor's Runnable to actually update
     * the model with gathered data. The update lock has been already acquired by the
     * caller.
     */
    // Builder = instantce of DocumentModelBuilder / SrcInfo
    private SrcElement.Impl updateSourceModel(DocumentModelUpdater builder)
        throws SrcException {


        SrcElement.Impl currentImpl = getSourceImpl();
        SrcElement.Impl finalImpl = currentImpl;
        boolean created;

        synchronized (this ) {
            if (!valid)
                throw new SrcException.IO("Source was deleted"); //NOI18N 
            updating = true;
            Util.log("updating"); //NOI18N 
            if ((created = (currentImpl == null))) {
                // Create local SrcElement.Impl in DefaultSeqModel (= updater)
                currentImpl = updater.createSource();   
            }
        }
        try {
            // this call lazy-initializes the proxy
            SrcElement s = getSource();  // Create Proxy and SourceElement if not yet done

            docBinding.enableGenerator(false);
            proxy.notifyInProgress(Thread.currentThread(), currentImpl);

            if (created) {
                currentImpl.attachedToElement(s);
                // Set the binding of SrcBinding =  SrcText.bindSource(s)
                ((ElementImpl)currentImpl).setBinding(docBinding.bindSource(s));    
            }

            Util.log("Trying to update the model"); // NOI18N
            builder.updateModel(updater, s, currentImpl);
            // finalImpl is null if currentimpl was created for the first time;
            if (finalImpl == null)
                // See DefaultSeqModel, send a notifyCreate() to all implementations
                updater.activate(s);
            Util.log("Update bindings"); // NOI18N
            builder.updateBindings(docBinding);
            finalImpl = currentImpl;
            synchronized (this) {
                updating = false;
                if (!valid) 
                    throw new SrcException.IO("Source was deleted"); // NOI18N
            }
            Util.log("register data = " + finalImpl); // NOI18N
            registerData(finalImpl);
        } finally {
            docBinding.enableGenerator(true);
            proxy.setModelDelegate(finalImpl, created); //set delegate...see srcimplproxy
        }
        return finalImpl;
    }
    
    public Task addParsingRunnable(Runnable r, int priority) {
        return PARSING_RP.post(r, priority);
    }
    

    static final Runnable EMPTY_RUNNABLE = new Runnable() {
        public void run() {}
    };
    
    /**
     * Helper class that is put to the parsing RequestProcessor to carry out 
     * a specific ParseRequest.
     */
    private class Processor extends Object implements
        Runnable, CommitListener, ParseObjectRequest {
        Processor             chained;
        int                   priority;
        RequestProcessor.Task ownTask;
        boolean               errorsOK;
        int                   stage;
        int                   resultStatus;
        ParsableObjectRequest request;
        T                     task;
        
        Processor(int priority, Env env, CloneableEditorSupport supp,ParsableObjectRequest req) {
            request=req;
            task = new T();
            request.setEnvironment(env); 
            request.setEditorSupport(supp); 
        }
        
        /**
         * If run() is directly invoked on the request processor task, it means
         * that the request processor thread itself needs to complete the task before
         * executing. We have to evaluate the task at once.
         */
        protected void directRun() {
            do {
                run();
            } while (stage >= 0);
        }
        
        public void setPriority(int prior) {
            if (this.priority > prior)
                return;
            priority = prior;
            // if the task cannot be cancelled, it is being processed right now,
            // or it was already completed.
            if (ownTask.cancel()) {
                addRequest(this, prior); 
            }
        }
        
        public void enableErrors(boolean enable) {
            errorsOK |= enable;
        }
        
        public void chainRequest(Processor other) {
            chained = other;
        }
        
        public void setProcessorTask(RequestProcessor.Task t) {
            ownTask = t;
        }
                
        public void run() {
            Util.log("processing request " + this + " stage " + stage); // NOI18N
            try {
                switch (stage++) {
                    case 0:
                        Util.log(ParsingSupport.class,"Returning finished task"); // NOI18N
                        parseLockModel();
                        break;
                    case 1:
                        updateModel();
                        break;
                }
            } catch (SrcException.IO e) {
                // the exception is expected - I/O in reading sources. Mark the source as errorenous.
                resultStatus = SrcElement.STATUS_ERROR;
            } catch (Throwable e) {
                savedException = new SrcException(e.getLocalizedMessage());
                parsingEnv.annotateThrowable(savedException, e);
                LOGGER.log(Level.INFO,"********* PARSING ERROR *************",savedException); //NOI18N
                resultStatus = SrcElement.STATUS_ERROR;
            } finally {
                stage--;
            }
            Util.log("request " + this + " stage " + (stage + 1) + " end"); // NOI18N
            if (stage > 0)
                return;
            
            if (resultStatus != SrcElement.STATUS_NULL) {
                complete();
            } else {
                // GW Something went wrong but it will 
                // most likely go wrong again the next time.
                // So we better complete here to allow reparsing
                // instead of looping.
                resultStatus = SrcElement.STATUS_NOT;
                complete();
                
                // it's a pity -- reschedule the request.
                //request.notifyReschedule();
                //Util.log("Rescheduling request"); // NOI18N
                //stage = 0;
                //addRequest(this, priority);
            }
        }
        
        private void updateModel() throws SrcException {
            model.removePreCommitListener(this);
            if (getSyntaxErrors() > 0 && !errorsOK && getSourceImpl() != null) {
                // don't update the model at all.
                resultStatus = SrcElement.STATUS_PARTIAL;
                return;
            }
            if (isValid()) {
                // we have the write lock, no-one changed the model
                // since we started.
                SrcElement.Impl i = null;
                
                // updater = instance of DocumentModelBuilder
                DocumentModelUpdater updater = request.getUpdater();
                if (updater != null) 
                    i = updateSourceModel( updater );
                if (i == null) {
                    resultStatus = SrcElement.STATUS_NOT;
                } else {
                    // SUCCESS (?)
                    resultStatus = getSyntaxErrors() > 0 ?
                        SrcElement.STATUS_PARTIAL : SrcElement.STATUS_OK;
                }
                return;
            }
        }
        
        private void parseLockModel() throws SrcException {
            model.addPreCommitListener(this);
            resultStatus = SrcElement.STATUS_NULL;
            synchronized (ParsingSupport.this) {
                runningRequest = this;
                Util.log("Running request = " + this); // NOI18N
            }
            savedException = null;
            try {
                runningRequest = this;
                process(getParserEngine());
                if (isValid()) {
                    Util.log("Request " + this + " processed. Still valid"); // NOI18N
                    stage = 1;
                    Util.log("trying to run update"); // NOI18N
                    ParsingSupport.this.updater.runUpdate(this, true);  
                }
            } catch (IOException ex) {
                savedException = new SrcException.IO(ex);
                parsingEnv.annotateThrowable(savedException, ex);
                LOGGER.log(Level.INFO,"********* PARSING ERROR *************",savedException); //NOI18N
                resultStatus = SrcElement.STATUS_ERROR;
            } catch (InternalError er) {
                // the parser sometimes issues this one -- just issue an error
                savedException = new SrcException(er.getMessage());
                LOGGER.log(Level.INFO,"********* PARSING ERROR *************",savedException); //NOI18N
                parsingEnv.annotateThrowable(savedException, er);
                resultStatus = SrcElement.STATUS_ERROR;
            } finally {
                model.removePreCommitListener(this);
            }
            runningRequest = null;
        }
        
        public void complete() {
            Collection<ParserMessage> messages;
            
            synchronized (ParsingSupport.this) {
                if (currentRequest == this)
                    currentRequest = null;
                messages = request.getMessages();
            }
            changeStatus(resultStatus);
            fireParsingEvent(messages);
            task.complete();
            if (chained != null)
                chained.complete();
            // signal -- processing is finally over.
            stage = -1;
       }
        
        public void changesCommited(Set created, Set removed, Map changed, Map<ElementImpl, Collection<PropertyChangeEvent>> sourcemap) {
            request.modelChanged();
        }

        /**
         * Causes the request to be passed immediately to the parsing engine.
         * If the request is already processed, it returns immediately.
         */
        public void process(ParserEngine eng) throws IOException {
            eng.register(this,getSourceFile());
            eng.process(this);
        }

        /**
         * Returns a task clients can wait for. This is *NOT* a task which processes the
         * request, since the request can be rescheduled several times before it is
         * completed (if it is completed).
         */
        public Task getClientTask() {
            return task;
        }

        /**
         * Notifies the request that the source text has been changed. This causes
         * cancellation of the request in some cases.
         */
        public void sourceChanged() {
            request.sourceChanged();
        }

        /* ParseObjectRequest method is delegated to request object */
        public void setSyntaxErrors(int errors) {
            request.setSyntaxErrors(errors);
        }

        public void setSemanticErrors(int errors) {
            request.setSemanticErrors(errors);
        }

        public void notifyStart() {
            request.notifyStart();
        }

        public void notifyComplete() {
            request.notifyComplete();
        }

        public boolean isValid() {
            return request.isValid();
        }

        public boolean needsProcessing() {
            return request.needsProcessing();
        }

        public int getSyntaxErrors() {
            return request.getSyntaxErrors();
        }

        public Collection<ParserMessage> getMessages() {
            return request.getMessages();
        }

        public ElementFactory getFactory() {
            return request.getFactory();
        }

        public char[] getSource() throws java.io.IOException { 
            clean = true;
            return request.getSource();
        }
        
        public SeqParser.Env getEnvironment() {
            return  request.getEnvironment();
        }

        public ErrConsumer getErrConsumer() {
            return request.getErrConsumer();
        }

        public org.openide.filesystems.FileObject  getSourceFile() {
            return request.getSourceFile();
        }
        
        public String getSourceName() {
            return request.getSourceName();
        }

        /* --------------------------------------------------------------------*/

        /** The class is NOT static intentionally - it holds a hard reference
         * to the request, keeping its data alive.
         */
        private class T extends org.openide.util.Task {
            T() {
                super(EMPTY_RUNNABLE);
            }

            // Used?
            public void run() {           
                Processor.this.directRun();
            }

            protected void complete() {
                super.notifyFinished();
            }
        }
    }
    
    /* --------------------------------------------------------------------*/
    static RequestProcessor    PARSING_RP;
    
    /** processor dedicated for parser events. */
    private static RequestProcessor EVENT_RP;
    
    private void addRequest(Processor proc, int priority) {
        synchronized (this) {
            if (PARSING_RP == null) {
                PARSING_RP = new RequestProcessor("Sequence source parsing"); // NOI18N
                EVENT_RP = new RequestProcessor("Sequence Parser Event Queue"); // NOI18N
            }
            if (currentRequest != proc) {
                if (currentRequest != null) 
                    proc.chainRequest(currentRequest);
                currentRequest = proc;
            }
            proc.setProcessorTask(PARSING_RP.post(proc, priority));
        }
    }
    
    /** Static method to test if a thread has been posted by the parsing support
     * RequestProcessor.
     * 
     * @param t
     * @return
     */
    public static boolean isSeqProcessorThread(Thread t) {
        String name = t.getName();
        if(name!=null && name.equals("Sequence source parsing"))
            return true;
        return false;
    }
    
    protected void registerData(SrcElement.Impl data) {
        Reference<SrcElement.Impl> newRef;

        synchronized (ParsingSupport.class) {
            newRef = new DataFinalizer<SrcElement.Impl>(this, data, Utilities.activeReferenceQueue());
        }
        synchronized (this) {
            refImplementation = newRef;
        }
    }
    
    protected void notifyFinalized(Reference refImpl) {
        int oldStatus;
        
        synchronized (this) {
            if (refImplementation != refImpl)
                return;
            refImplementation = null;
        }
        LOGGER.fine("Source implementation finalized"); //NOI18N
        changeStatus(SrcElement.STATUS_NOT);
    }
    
    private static class DataFinalizer<T> extends WeakReference<T> implements Runnable {
        Reference<ParsingSupport> refSupp;

        DataFinalizer(ParsingSupport supp, T data, ReferenceQueue<? super T> rqueue) {
            super(data, rqueue);
            refSupp = new WeakReference<ParsingSupport>(supp);
        }
        
        public void run() {
            Reference f;

            ParsingSupport supp = refSupp.get();
            if (supp != null)
                supp.notifyFinalized(this);
        }
    }
}
