/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or sbbuild/licenses/cddl-gpl-2-cp.txt. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * sbbuild/licenses/cddl-gpl-2-cp.txt.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 2004-2007 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */


package org.starbeans.modules.seqformat.parser;

import java.beans.PropertyChangeListener;
import java.io.Reader;
import java.io.IOException;
import javax.swing.event.ChangeListener;
import org.openide.nodes.Node;
import org.openide.util.Task;
import org.starbeans.modules.seqformat.model.SeqModel;
import org.starbeans.modules.seqformat.source.Element;
import org.starbeans.modules.seqformat.source.SrcCookie;
import org.starbeans.modules.seqformat.source.SrcElement;
import org.starbeans.modules.seqformat.source.SrcException;

/**
 *
 * @author  sdedic, George Waldon
 * @version 
 */
public interface SeqParser extends SrcCookie {
    /**
     * This priority level is recommended for background operations, that the user
     * does not need to wait for.
     */
    public static final int    PRIORITY_BACKGROUND = Thread.MIN_PRIORITY;
    /**
     * Normal priority is recommended for parse operations that needs to be completed soon,
     * but should not block the IDE. This priority is internally used in prepare() call.
     */
    public static final int    PRIORITY_NORMAL = Thread.MIN_PRIORITY + 1;
    /**
     * Recommended priority for carrying out blocking calls. This priority is used
     * by blocking calls on SrcElement.
     */
    public static final int    PRIORITY_DEMAND = Thread.MAX_PRIORITY - 1;
    /**
     * Recommended priority for hierarchy refresh operations. This priority is used
     * in the editor suppport after the user stops editing for a period of
     * time.
     */
    public static final int    PRIORITY_REFRESH = PRIORITY_NORMAL;

    /**
     * Adds a Listener on the parser.
     */
    public void addPropertyChangeListener(PropertyChangeListener l);
    
    /**
     * Removes the listener from the parser.
     */
    public void removePropertyChangeListener(PropertyChangeListener l);
    
    /**
     * Listens on parser state changes (more events than just PropertyChanges!)
     */
    public void addChangeListener(ChangeListener l);
    
    /**
     * Removes the change listener from the parser.
     */
    public void removeChangeListener(ChangeListener l);
    
    /**
     * Returns the Element representing the source.
     */
    public SrcElement getSource();
    
    /**
     * Prepares the source hierarchy; after the task is finished, there are some
     * parsed data. If the model already exists, the task returned is already finished.
     * The returned task, after it completes, also guarantees that the model will not
     * be garbage collected.
     */
    public Task prepare();
    
    /**
     * Schedules a parse task with various parameters.
     * @param force schedules a parse even if the underlying document contains no
     * unparsed changes. The method returns a task that you can wait for, or attach
     * TaskListeners to. See {@link #prepare} for more information.
     * @param acceptErrors force model update even if the parsed data indicate syntax
     * errors.
     * @param priority priority of the task.
     * @return task for synchronization purposes. The task is marked finished after the
     * model is created or refreshed.
     */
    public Task parse(int priority, boolean force, boolean acceptErrors);
    
    /**
     * Returns current running parser task, if there is no task running returns FinishedTask
     * @return task for synchronization purposes.
     */
    public Task getCurrentTask();
    
    /**
     * Returns the overall current parser's status. The returned value is one of
     * the manifested constants SourceElement.STATUS_*.
     * @return status information
     */
    public int getStatus();
    
    /** Inquire about the validity of this parser.
     */
    public boolean isValid();
    
    /**
     * Retrieves the exception that describes the last failure to parse the
     * input. This will be mostly SourceException.IO.
     */
    public SrcException  getErrorCause();
    
    /**
     * Returns the <B>real</B> implementation of the source provided by the data model.
     */
    public SrcElement.Impl   getSourceImpl();
    
    /**
     * Returns the model the parser operates on.
     */
    public SeqModel            getModel();

    /**
     * Interface that the support uses to talk to the outside world.
     */
    public interface Env {

        /**
         * Used by SeqParser to create local SrcElement in the environement.
         */
        public SrcElement getLocalSrcElement(SrcElement.Impl sei);

        /**
         * Returns the FileObject corresponding to the source, if there's one.
         * The method may return null if there is no FileObject == the source
         * is only in memory.
         * @return FileObject or <code>null</code>
         */
        public org.openide.filesystems.FileObject getSourceFile();

        /**
         * Returns the name of the source, if one was assigned. If {@link #getSourceFile}
         * returns non-null, then this method must return getSourceFile().getNameExt().
         * The returned value may used for error reporting when the passed source is
         * not file-based.
         *
         * @return name of the source or <code>null</code>
         */
        public String getSourceName();

        /** Retrieves the source text for parsing. Can throw IOException, if the
         * source cannot be retrieved.
         */
        public Reader getSourceText() throws IOException;

        /**
         * Finds appropriate cookie for the element. Dedicated to
         * SrcElement, but intentionally casted like in SeqModel.Env
         */
        public <T extends Node.Cookie> T findCookie(Element el, Class<T> requestedClass);

        
        /**
         * Annotates wrapping throwable with the nested one - whatever that means to the
         * environment.
         */
        public void annotateThrowable(Throwable wrapper, Throwable nested);
    }
}
