/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or sbbuild/licenses/cddl-gpl-2-cp.txt. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * sbbuild/licenses/cddl-gpl-2-cp.txt.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 2004-2007 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */


package org.starbeans.modules.seqformat.parser;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.VetoableChangeListener;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import javax.swing.text.BadLocationException;
import javax.swing.text.StyledDocument;
import org.openide.nodes.Node;
import org.openide.text.NbDocument;
import org.openide.util.Exceptions;
import org.openide.util.Task;
import org.starbeans.modules.seqformat.model.ElementImpl;
import org.starbeans.modules.seqformat.source.Element;
import org.starbeans.modules.seqformat.source.ElementPrinter;
import org.starbeans.modules.seqformat.source.ElementPrinterInterruptException;
import org.starbeans.modules.seqformat.source.SeqElement;
import org.starbeans.modules.seqformat.source.SrcElement;
import org.starbeans.modules.seqformat.source.SrcException;

    /**
     * The class serves as a delegating proxy that forwards all model-related
     * requests to the actual implementation in the model while it processes
     * management functions that are (for historical reasons) available on
     * SrcElement.
     */
    abstract public class SrcImplProxy implements SrcElement.Impl, Node.Cookie, PropertyChangeListener {
        
    private static final SeqElement[] NO_SEQS = new SeqElement[0];
    
    /**
     * PropertyChangeListeners attached to the SrcElement.
     */
    LinkedList<PropertyChangeListener> listeners;
    
    /**
     * The Element representing this proxy/impl. Once set, the element cannot be
     * changed.
     */
    SrcElement              srcElement;

    /**
     * The real parser core 
     */
    ParsingSupport          supp;
    
    /**
     * Thread, that should not be blocked during data extraction. Instead of
     * blocking on parser's completion, this thread should be directed at
     * temporary implementation.
     */
    Thread                  dontBlockThread;
    
    /**
     * Task to wait for. If not null, all getters and setters need to be blocked
     * on that task; data are live until that task finishes.
     */
    Task                    waitTask;
    
    /**
     * Temporary delegate object to use during data extraction. The thread that
     * creates the data structure need to access the SrcElement, possibly through
     * the OpenAPI calls.
     */
    SrcElement.Impl        tempDelegate;
    
    private static final long serialVersionUID = 200311231047000L;
    
    /**
     * Called just after the proxy is constructed to
     * conveniently pass the ParsingSupport.
     */    
    public void attachSupport(ParsingSupport parsingSupp) {
        this.supp = parsingSupp;
        supp.addPropertyChangeListener(this);
    }
    /**
     * Returns the SrcElement representing the hierarchy.
     */
    protected SrcElement getElement() {
        return this.srcElement;
    }

    
    //============== Implementation Element.Impl ===============================    
    /**
     * API callback; called when the SrcElement is constructed. This implementation
     * only records the reference
     */
    public void attachedToElement(Element el) {
        this.srcElement = (SrcElement)el;
    }
    
    /**
     * Adds a PropertyChangeListener
     */
    public synchronized void addPropertyChangeListener(PropertyChangeListener l) {
        if (listeners == null) {
            listeners = new LinkedList<PropertyChangeListener>();
        }
        boolean attach = listeners.isEmpty();
        if (l == null)
            return;
        listeners.add(l);
        SrcElement.Impl impl = supp.getSourceImpl();
        if (impl != null && attach)
            impl.addPropertyChangeListener(this);
    }
    
    public synchronized void removePropertyChangeListener(PropertyChangeListener l) {
        if (listeners == null)
            return;
        listeners.remove(l);
        // does not remove (this) listener from the SrcElement.Impl
    }
    
    public Object readResolve() {
        return null;
    }

    
    /**
     * Implementation of getCookie(). The implementation fakes cookies of all interfaces
     * on this object; if the cookie is not available, it asks the environment to
     * provide the cookie and IF the cookie descends from the implementation hierarchy
     * passes the request on to the real SrcElement's implementation, optionally
     * parsing it from the source.
     */
    public <T extends Node.Cookie> T getCookie(Class<T> type) {
        if (type.isAssignableFrom(getClass()))
            return type.cast(this);
        SrcElement.Impl impl = supp.getSourceImpl();
        if (impl != null)
            return impl.getCookie(type);
        // PENDING: ask the environment about cookies.
        if (!ElementImpl.class.isAssignableFrom(type))
            return this.supp.findCookieForSource(type);
        impl = safeFindModelDelegate();
        if (impl == null)
            return null;
        return impl.getCookie(type);
    }
    
    /**
     * No effect on SrcElements, there's only one of them in a source :-)
     */
    public void markCurrent(boolean beforeAfter) {
    }
    
    public boolean isMarked() {
        return false;
    }
    
    // The three next = implementation of Impl2
    public void addVetoableChangeListener(VetoableChangeListener l) { 
        // Not implemented!
    }
        
    public void removeVetoableChangeListener(VetoableChangeListener l) {
        // Not implemented!
    }

    public boolean isValid() {
        // Not implemented!
        return true;
    }
   
    //============== Implementation SrcElement.Impl ============================        
    public SeqElement[] getAllSeqElements() {
        SrcElement.Impl impl = safeFindModelDelegate();
        if (impl != null)
            return impl.getAllSeqElements();
        return NO_SEQS;
    }
    
    public boolean acceptSeqElement(SeqElement el) {
        SrcElement.Impl impl = safeFindModelDelegate();
        if (impl != null)
            return impl.acceptSeqElement(el);
        return false;
    }
    
    public void changeSeqElements(SeqElement[] els, int mode) throws SrcException {
        SrcElement.Impl impl = safeFindModelDelegate();
        if (impl != null)
            impl.changeSeqElements(els,mode);
    }
    
    /**
     * Retrieves the status of the parsing operation. Delegates to ParsingSupport.
     */
    public int getStatus() {
        return supp.getStatus();
    }
    
    /**
     * Implementation of prepare() call. Delegates to ParsingSupport's
     * prepare(), ignores errors and does nothing if the source is up-to-date.
     */
    public Task prepare() {
        return supp.parse(ParsingSupport.PRIORITY_BACKGROUND, false, false);
    }

    /**
     * Exclusively locks the model for the execution of the Runnable object. Note that
     * the implementation does *NOT* lock the document! Also, do *NOT* call the
     * method if you have already locked the document for writing unless you
     * _REALLY_ know what you are doing.
     */
    public void runAtomic(final Runnable run) {
        final StyledDocument doc;
        try {
            doc = supp.docBinding.getEditorSupport().openDocument();
        } catch (java.io.IOException ex) {
            return;
        }
        try {
            supp.model.runAtomic(new Runnable() {
                public void run() {
                    try {
                        //supp.docBinding.enableAtomicAsUser(true);
                        NbDocument.runAtomic(doc, run);
                    } finally {
                        //supp.docBinding.enableAtomicAsUser(false);
                    }
                }
            });
        } catch (SrcException ex) {
        }
    }
    /**
     * Creates an atomic transaction, that respects the guarded sections inside
     * the document, over the model. Again, this does *NOT* lock the document.
     */
    public void runAtomicAsUser(final Runnable run) throws SrcException {
        final StyledDocument doc;
        try {
            doc = supp.docBinding.getEditorSupport().openDocument();
        } catch (java.io.IOException ex) {
            throw new SrcException.IO(ex);
        }
        supp.model.runAtomic(new Runnable() {
            public void run() {
                try {
                    supp.docBinding.enableAtomicAsUser(true);
                    try {
                        NbDocument.runAtomicAsUser(doc, run);
                    } catch (BadLocationException ex) {
                        //
                    }
                } finally {
                    supp.docBinding.enableAtomicAsUser(false);
                }
            }
        });
    }

 
    //=========================================================================
    
    
    /**
     * Attempts to locate the Impl object provided by the "real" model. If the object
     * cannot be acquired (for example missing or unparseable source), it throws a
     * SrcException.
     */
    protected SrcElement.Impl findModelDelegate() throws SrcException {   
        /* This is a gate for the INITIALIZING thread to get through the
         * proxy. All other threads will fail miserably waiting on the
         * parsing task. */
        // dontBlockThread is the parsing thread in parsingsupport
        if (dontBlockThread == Thread.currentThread()) {
            return tempDelegate;
        }
        
        // optionally throws a SrcException, if the error cause is known.
        SrcElement.Impl candidate = supp.getSourceImpl();
        if (candidate != null)
            return candidate;

        Task t;
        t = supp.parse(ParsingSupport.PRIORITY_DEMAND, false, false); 
        t.waitFinished();
        SrcElement.Impl i = supp.findSourceImpl();
        t.isFinished();
        return i;
    }
    
    /**
     * Safe variant of the method above -- does not throw any exceptions, but rather
     * returns null, if the source element cannot be created.
     */
    protected SrcElement.Impl safeFindModelDelegate() {
        try {
            return findModelDelegate();
        } catch (SrcException ex) {
            return null;
        }
    }
    
    /**
     * PropertyChangeListener implementation that refires all property changes
     * to the listeners attached to a SrcElement.
     */
    public void propertyChange(PropertyChangeEvent evt) {
        List<PropertyChangeListener> l;
        synchronized (this) {
            if (listeners == null)
                return;
            l = new ArrayList<PropertyChangeListener>(listeners);
        }
        // remap the event so that it appears to be fired from the source element.
        evt = new PropertyChangeEvent(
            srcElement, evt.getPropertyName(), 
            evt.getOldValue(), evt.getNewValue());
        
        for (Iterator<PropertyChangeListener> it = l.iterator(); it.hasNext(); ) {
            PropertyChangeListener ll = it.next();
            ll.propertyChange(evt);
        }
    }
    
    /**
     * Called by the ParsingSupport to notify the proxy, that data extraction is
     * in progress; the call also informs about the thread that performs updates
     * and therefore should not be blocked.
     * @param t thread updating the model
     * @param tempDelegate temporary delegate that is being updated with fresh data and
     * should be delegated to during the update cycle.
     */
    void notifyInProgress(Thread t, SrcElement.Impl tempDelegate) {
        this.dontBlockThread = t;
        this.tempDelegate = tempDelegate;
    }
    
    /**
     * This method is called by the ParsingSupport when it finally constructs
     * the implementation object. Temp delegate is not needed any more and
     * non-blocking thread is obsolete, too.
     * @param delegate the final delegate to use, may be null if the update fails.
     * @param created true, if the delegate was created rather than refreshed. In that
     * case, a listener is attached to watch PropertyChangeEvents.
     */
    protected synchronized void setModelDelegate(SrcElement.Impl delegate, boolean created) {
        List ls = listeners;
        synchronized (this) {
            // clear blocking/task information
            dontBlockThread = null;
            tempDelegate = null;
            waitTask = null;
        }
        if (!created || ls == null || delegate == null)
            return;
        // add ONE property change listener:
        // The listener is *NOT* Weak intentionally; the implementation object prevents
        // the Proxy from being GCed. 
        delegate.addPropertyChangeListener(this);
    }
    
    public void print(ElementPrinter printer) throws ElementPrinterInterruptException {
    }
    
}
