/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or sbbuild/licenses/cddl-gpl-2-cp.txt. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * sbbuild/licenses/cddl-gpl-2-cp.txt.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 2004-2007 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */


package org.starbeans.modules.seqformat.source;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.beans.PropertyVetoException;
import java.beans.VetoableChangeListener;
import java.beans.VetoableChangeSupport;
import java.io.PrintWriter;
import java.io.Serializable;
import java.io.StringWriter;
import org.openide.nodes.Node;

/** Base class for representations of elements.

* @author Jaroslav Tulach, Petr Hamernik, Svatopluk Dedic, George Waldon
*/
public abstract class Element extends Object
    implements Serializable, ElementProperties, Node.Cookie {

    /** Implementation */
    protected Impl impl;

    static final long serialVersionUID =200311190938000L;
    
    /** Create a new element with the provided implementation. The implementation is
    * responsible for storing all properties of the object.
    *
    * @param impl the implementation to use
    */
    protected Element(Impl impl) {
        this.impl = impl;
        /* Allow creation of virtual Element without implementation */
        if (impl != null)
            impl.attachedToElement(this);
    }

    /** Return the parent Element to which this Element belongs 
     */
    public Element getParent() {
        // Undefined
        return null;
    }
    
    /** Add a property change listener.
    * @param l the listener to add
    * @see ElementProperties
    */
    public final void addPropertyChangeListener (PropertyChangeListener l) {
        impl.addPropertyChangeListener (l);
    }

    /** Remove a property change listener.
    * @param l the listener to remove
    * @see ElementProperties
    */
    public final void removePropertyChangeListener (PropertyChangeListener l) {
        impl.removePropertyChangeListener (l);
    }
    
    /**
     * Attaches a VetoableChange listener to the element.
     * The Listener will be notified about changes being done before the change is
     * actually done.
     * @param l instance of listener to attach.
     */
    public final void addVetoableChangeListener(VetoableChangeListener l) {
        if (impl != null)
            impl.addVetoableChangeListener(l); 
    }

    /**
     * Removes the vetoable listener from the element.
     * @param l listener to remove.
     */
    public final void removeVetoableChangeListener(VetoableChangeListener l) {
        if (impl != null)
            impl.removeVetoableChangeListener(l); 
    }

    /** Mark the current element in the context of this element.
    * The current element means the position for inserting new elements.
    * @param beforeAfter <CODE>true</CODE> means that new element is inserted before
    * the specified element, <CODE>false</CODE> means after.
    */
    public void markCurrent(boolean beforeAfter) {
        impl.markCurrent(beforeAfter);
    }

    /** Test if the element is marked.
     */
    public boolean isMarked() {
        return impl.isMarked();
    }
   
    /** Look for a cookie providing added behavior for this element.
    * The request is {@link Impl#getCookie delegated} to the current implementation.
    * Also note that <code>Element</code> implements <code>Node.Cookie</code>, and that
    * if the implementation does not provide a cookie, but the requested cookie class is
    * actually a superclass/interface of this element type, then the element itself may be
    * returned as the cookie.
    * @param type the cookie class to look for
    * @return a cookie assignable to that class, or <code>null</code> if the cookie
    *    is not supported
    */
    public <T extends Node.Cookie> T getCookie(Class<T> type) {
        T c = impl.getCookie(type);
        if ((c == null) && type.isAssignableFrom(getClass()))
            c = type.cast(this);
        return  c;
    }

    protected Object writeReplace() { 
        return impl;
    }
    
    /** Print this element (and all its subelements) into an element printer.
    * @param printer the element printer
    * @exception ElementPrinterInterruptException if the printer canceled the printing
    */
    public abstract void print(ElementPrinter printer) throws ElementPrinterInterruptException;

    /** Prints array of elements.
    * @param el the elements
    * @param printer The printer where to write
    * @return true if at least one element was printed
    * @exception ElementPrinterInterruptException if printer cancel the printing
    */
    static boolean print(Element[] el, ElementPrinter printer) throws ElementPrinterInterruptException {
        for (int i = 0; i < el.length; i++) {
            if (i > 0) {
                printer.println(""); // NOI18N
                printer.println(""); // NOI18N
            }
            el[i].print(printer);
        }
        return (el.length > 0);
    }
    
    /** Helper method that properly annotates an exception.
     *  @param message used for human-readable explanation of the exception.
     */
    protected void throwSrcException(String message) throws SrcException {
	SrcException ex = new SrcException(message);
	throw ex;    
    }
    
    protected void throwSrcException(Throwable nested) throws SrcException {
	SrcException ex = new SrcException(nested.getMessage());
	ex.initCause(nested);
	throw ex;    
    }	

    /** Get a string representation of the element.
    * @return the string
    * @see #print
    * @see DefaultElementPrinter
    */
    public String toString() {
        StringWriter sw = new StringWriter();
        PrintWriter pw = new PrintWriter(sw);
        try {
            print(new DefaultElementPrinter(pw));
        }
        catch (ElementPrinterInterruptException e) {
            // could not happen.
        }
        pw.close();
        return sw.toString();
    }

    /** Pluggable implementation of the storage of element properties.
    * @see Element#Element
    */
    public interface Impl extends Serializable {
        
        /** Add some items. */
        public static final int ADD = 1;
        /** Remove some items. */
        public static final int REMOVE = -1;
        /** Set some items, replacing the old ones. */
        public static final int SET = 0;
        
        static final long  serialVersionUID = 200311191010000L;
        
        /** Called to attach the implementation to a specific
        * element. Will be called in the element's constructor.
        * Allows implementors of this interface to store a reference to the holder class,
        * useful for implementing the property change listeners.
        *
        * @param el the element to attach to
        */
        public void attachedToElement (Element el);

        /** Add a property change listener.
        * @param l the listener to add
        */
        public void addPropertyChangeListener (PropertyChangeListener l);

        /** Remove a property change listener.
        * @param l the listener to remove
        */
        public void removePropertyChangeListener (PropertyChangeListener l);
        
        /** Implementations must be resolvable.
        * I.e., upon deserialization they must be able to recreate the
        * holder class.
        * @return an instance of the proper subclass of {@link Element}
        * @see Serializable
        */
        public Object readResolve();

        /** Get the support for a cookie, if any.
        * Changes of supported cookies are <em>not</em> fired.
        *
        * @param type the cookie class to look for
        * @return an instance assignable to that class, or <code>null</code> if the cookie
        *    is not supported
        */
        public <T extends Node.Cookie> T getCookie(Class<T> type);

        /** Mark the current element in the context of this element.
        * The current element means the position for inserting new elements.
        */
        public void markCurrent(boolean MarkIt);

        /** Test if the element is marked.
        */
        public boolean isMarked ();

        /** Adds a vetoable listener.
         * @param l instener instance to add.
         */
        public void addVetoableChangeListener(VetoableChangeListener l);
         
        /** Removes a vetoable listener.
         * @param l instener instance to remove.
         */
        public void removeVetoableChangeListener(VetoableChangeListener l);

        /**
         * Determines whether the element is still valid - if it is a part of its original model.
         * The Element should be invalidated during its removal from the model so further 
         * operations may notice that and fail in appropriate cases.
         * @return true, if the element is still valid and a part of a model.
         */
        public boolean isValid();
    } 

    //========================================================================//
    
    /** Default implementation of the Impl interface.
    * It just holds the property values.
    */
    static abstract public class Memory implements Element.Impl, Node.Cookie {
        /** the element for this implementation */
        protected Element element;

        /** Property change support */
        private PropertyChangeSupport support;
        
        /**
         * VetoableChange Support.
         */
        private VetoableChangeSupport vetoSupport;

        static final long serialVersionUID =200311191014000L;
        
        /** Attaches to element */
        public void attachedToElement (Element element) {
            this.element = element;
        }

        /** Fires property change event.
        * @param name property name
        * @param o old value
        * @param n new value
        */
        protected final void firePropertyChange (String name, Object o, Object n) {
            if (support != null) {
                support.firePropertyChange (name, o, n);
            }
        }
        
        protected final void firePropertyChange(PropertyChangeEvent evt) {
            if (support != null)
                support.firePropertyChange(evt);
        }

        /** Adds property listener */
        public synchronized void addPropertyChangeListener (PropertyChangeListener l) {
            if (support == null) {
                synchronized (this) {
                    // new test under synchronized block
                    if (support == null) {
                        support = new PropertyChangeSupport (element);
                    }
                }
            }
            support.addPropertyChangeListener (l);
        }

        /** Removes property listener */
        public void removePropertyChangeListener (PropertyChangeListener l) {
            if (support != null) {
                support.removePropertyChangeListener (l);
            }
        }

        public void addVetoableChangeListener(VetoableChangeListener l) {
            if (vetoSupport == null) {
                synchronized (this) {
                    if (vetoSupport == null)
                        vetoSupport = new VetoableChangeSupport(this);
                }
            }
            vetoSupport.addVetoableChangeListener(l);
        }

        public void removeVetoableChangeListener(VetoableChangeListener l) {
            if (vetoSupport != null)
                vetoSupport.removeVetoableChangeListener(l);
        }
        
        protected void fireVetoableChange(String name, Object o, Object v) throws PropertyVetoException {
            if (vetoSupport != null)
                vetoSupport.fireVetoableChange(name, o, v);
        }
        
        protected void fireVetoableChange(PropertyChangeEvent evt) throws PropertyVetoException {
            if (vetoSupport != null)
                vetoSupport.fireVetoableChange(evt);
        }

        /** This implementation returns always null.
        * @param type the class to look for
        * @return null.
        */
        public <T extends Node.Cookie> T getCookie(Class<T> type) {
            if (type.isAssignableFrom(getClass())) {
              return type.cast(this);
            } else {
              return null;
            }
        }

        /** Mark the current element in the context of this element.
        * The current element means the position for inserting new elements.
        * @param beforeAfter <CODE>true</CODE> means that new element is inserted before
        *        the specified element, <CODE>false</CODE> means after.
        */
        abstract public void markCurrent(boolean beforeAfter);
        
        abstract public Object readResolve();
        
        public boolean isValid() {
            return true;
        }
    }
}

