/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or sbbuild/licenses/cddl-gpl-2-cp.txt. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * sbbuild/licenses/cddl-gpl-2-cp.txt.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 2004-2007 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */


package org.starbeans.modules.seqformat.source;

import java.io.IOException;
import java.util.ResourceBundle;
import java.util.logging.Logger;
import org.openide.NotifyDescriptor;
import org.openide.DialogDisplayer;
import org.openide.loaders.DataObject;
import org.openide.util.NbBundle;
import org.openide.util.UserQuestionException;

/** This class defines utilities for editing source.
 *
 * @author Petr Hamernik, George Waldon
 */
public class SrcEditSupport {
    
    static final ResourceBundle bundle = NbBundle.getBundle(SrcEditSupport.class);
    private static final Logger LOGGER = Logger.getLogger(SrcEditSupport.class.getName());
    
    /** Invokes the runnable using SrcElement.runAtomicAsUser, if it can find
     * a source element for the given hierarchy element. If the SourceElement can't
     * be found, the runnable is executed without any protection.
     * @exception IOException If SrcException occured inside the runnable.
     */
    static public void invokeAtomicAsUser(Element element, final ExceptionalRunnable exRun) throws IOException {
        try {
            runAsUser(element, exRun);
        } catch (SrcException.IO e) {
            throw e.getReason(); //already inited!
        }
        catch (SrcException e) {
            IOException x = new IOException(e.getMessage());
            x.initCause(e);
            throw x;
        }
    }
    
    static public void runAsUser(Element ref, final ExceptionalRunnable exRun) throws SrcException {
        final SrcException sex[] = { null };
        SrcElement src = findSource(ref);
        boolean retry = false;
        do {
            sex[0] = null;
            if (src == null) {
                exRun.run();
            } else {
                src.runAtomicAsUser(new Runnable() {
                    public void run() {
                        try {
                            exRun.run();
                        } catch (SrcException e) {
                            sex[0] = e;
                        }
                    }
                });
            }
            if (sex[0] != null) {
                if (retry)
                    break;
                retry = true;
                if (sex[0] instanceof SrcException.IO) {
                    IOException iex = ((SrcException.IO)sex[0]).getReason();
                    if (iex instanceof UserQuestionException) {
                        UserQuestionException uex = (UserQuestionException)iex;
                        NotifyDescriptor.Confirmation nc = new NotifyDescriptor.Confirmation(uex.getLocalizedMessage(),
                        bundle.getString("TIT_CannotWriteFile"), NotifyDescriptor.Confirmation.YES_NO_OPTION);
                        Object o = DialogDisplayer.getDefault().notify(nc);
                        if (o == NotifyDescriptor.YES_OPTION) {
                            try {
                                uex.confirmed();
                                // HACK!! See issue #23407
                                continue;
                            } catch (IOException x) {
                                sex[0] = new SrcException.IO(x);
                                sex[0].initCause(x);
                            }
                        } else {
                            iex = new IOException(bundle.getString("ERR_CannotWriteFile")); //NOI18N
                            sex[0] = new SrcException.IO(iex);
                            sex[0].initCause(iex);
                        }
                    }
                }
                // not a UserQuestionException, unconfirmed UQE, exception from UQE.confirmed.
                break;
            }
        } while (sex[0] != null);
        
        if (sex[0] != null)
            throw sex[0];
    }
    
    /** This interface is used like runnable, but its method run
     * could throw BadLocationException.
     * @exception SourceException
     */
    public static interface ExceptionalRunnable {
        public void run() throws SrcException;
    }
    
    static public boolean isWriteable(Element element) {
        SrcElement el = findSource(element);
        if(el == null)
            return false;
        DataObject d = el.getCookie(DataObject.class);
        if (d == null) {
            return true;
        }
        return d.getPrimaryFile().canWrite();
    }
    
    /** Find the source for the specifier element.
     * @return instance of SourceElement or null, if the source can't be found.
     */
    static public SrcElement findSource(Element element) {
        if (element instanceof SeqElement && element != null) {
            return ((SeqElement) element).getSource();
        } else
            return null;
    }

}
