/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or sbbuild/licenses/cddl-gpl-2-cp.txt. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * sbbuild/licenses/cddl-gpl-2-cp.txt.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 2004-2007 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */


package org.starbeans.modules.seqformat.source;


import org.openide.util.Task;

/** Describes an entire source file.
* Note that there is no standard in-memory implementation of this element;
* every user of the class is expected to have a reasonable
* implementation according to where the source file resides.
* <p>The source element should be parsed in the background using
* {@link #prepare} before any attempts are made to access its properties
* to read or to write, or to call {@link #print};
* otherwise such accesses will block until the parse is finished.
*
* @author Petr Hamernik, Jaroslav Tulach, George Waldon
*/
public class SrcElement extends Element {
    
    /** Status indicating that no parsing are under way.
     */
    public static final int STATUS_NULL = -1;
   
    /** Status when the source element is not yet prepared.
    */
    public static final int STATUS_NOT = 0;
    
    /** Status when the source element contains unrecoverable errors.
    */
    public static final int STATUS_ERROR = 1;
    
    /** Status when the source element contains minor errors.
    */
    public static final int STATUS_PARTIAL = 2;
    
    /** Status when the source element has been parsed and is error-free.
    */
    public static final int STATUS_OK = 3;

    static final long serialVersionUID =200311191023000L;

    /** Create a new source element.
    * @param impl the pluggable implementation
    */
    public SrcElement(Impl impl) {
        super (impl);
    }

    /** @return implementation for the source
    */
    protected final Impl getSourceImpl () {
        return (Impl)impl;
    }

    /** Get the parsing status of the element.
    * This is a non-blocking operation.
    * @return one of {@link #STATUS_NOT}, {@link #STATUS_ERROR}, {@link #STATUS_PARTIAL}, or {@link #STATUS_OK}
    */
    public int getStatus () {
        return getSourceImpl ().getStatus ();
    }

    /** Begin parsing this source element.
    * This method is non-blocking; it only returns
    * a task that can be used to control the ongoing parse.
    * Initially the {@link #getStatus} should be {@link #STATUS_NOT}, and change
    * to one of the other three when parsing is complete, according to whether
    * or not errors were encountered, and their severity.
    *
    * @return a task to control the preparation of the element
    */
    public Task prepare () {
        return getSourceImpl ().prepare ();
    }


    /** Lock the underlaing document to have exclusive access to it and could make changes
    * on this SourceElement.
    *
    * @param run the action to run
    */
    public void runAtomic (Runnable run) {
        getSourceImpl().runAtomic(run);
    }

    /** Executes given runnable in "user mode" does not allowing any modifications
    * to parts of text marked as guarded. The actions should be run as "atomic" so
    * either happen all at once or none at all (if a guarded block should be modified).
    *
    * @param run the action to run
    * @exception SrcException if a modification of guarded text occured
    *   and that is why no changes to the document has been done.
    */
    public void runAtomicAsUser (Runnable run) throws SrcException {
        getSourceImpl().runAtomicAsUser(run);
    }


    /** Get all Elements recursively.
    * @return all Elements
    */
    public Element[] getAllElements () {
        return getAllSeqElements ();
    }

    /** Get all SeqElements recursively.
    * @return all SeqElements
    */
    public SeqElement[] getAllSeqElements () {
        return getSourceImpl().getAllSeqElements();
    }
    
    /* Prints the element into the element printer.
    * @param printer The element printer where to print to
    * @exception ElementPrinterInterruptException if printer cancel the printing
    */
    public void print(ElementPrinter printer) throws ElementPrinterInterruptException {
        Element.print(getAllElements(), printer);
    }

    // =============== SeqElement modification section ====================
    
    /** Test whether to accept a new SeqElement in the model. This is delegated
     * to the implementation.
     */
    public boolean acceptSeqElement(SeqElement sel) {
        return getSourceImpl().acceptSeqElement(sel);
    }
    
    /** Add a new Element. Entry point to add new sequence in the model.
     *
    * @param el the Element to add
    * @throws SrcException if impossible
    */
    public void addSeqElement (SeqElement sel) throws SrcException {
        getSourceImpl().changeSeqElements (new SeqElement[] { sel }, Impl.ADD);
    }

    /** Add some new Elements.
    * @param els the Elements to add
    * @throws SrcException if impossible
    */
    public void addSeqElements (final SeqElement[] sels) throws SrcException {
        getSourceImpl().changeSeqElements(sels, Impl.ADD);
    }


    /** Remove an Element.
    * @param el the Element to remove
    * @throws SrcException if impossible
    */
    public void removeSeqElement (SeqElement sel) throws SrcException {
        getSourceImpl().changeSeqElements(new SeqElement[] { sel }, Impl.REMOVE);
    }

    /** Remove some Elements.
    * @param els the Elements to remove
    * @throws SrcException if impossible
    */
    public void removeSeqElements (final SeqElement[] sels) throws SrcException {
        getSourceImpl().changeSeqElements(sels, Impl.REMOVE);
    }
    
    /** Pluggable behaviour for source elements.
    * @see SrcElement
    */
    public static interface Impl extends Element.Impl {
        /** Add some elements. */
        public static final int ADD = Element.Impl.ADD;
        /** Remove some elements. */
        public static final int REMOVE = Element.Impl.REMOVE;
        /** Set the element. */
        public static final int SET = Element.Impl.SET;

        static final long serialVersionUID = 200311191046000L;

        /** Get the parsing status of the element.
         * This is a non-blocking operation.
         * @return one of {@link #STATUS_NOT}, {@link #STATUS_ERROR}, {@link #STATUS_PARTIAL}, or {@link #STATUS_OK}
         */
        public int getStatus ();

        /** Begin parsing this source element.
         * This method is non-blocking; it only returns
         * a task that can be used to control the ongoing parse.
         * Initially the {@link #getStatus} should be {@link #STATUS_NOT}, and change
         * to one of the other three when parsing is complete, according to whether
         * or not errors were encountered, and their severity.
         *
         * @return a task to control the preparation of the element
         */
        public Task prepare ();

        /** Lock the underlaing document to have exclusive access to it and could make changes
        * on this SourceElement.
        *
        * @param run the action to run
        */
        public void runAtomic (Runnable run);

        /** Executes given runnable in "user mode" does not allowing any modifications
        * to parts of text marked as guarded. The actions should be run as "atomic" so
        * either happen all at once or none at all (if a guarded block should be modified).
        *
        * @param run the action to run
        * @exception SrcException if a modification of guarded text occured
        *   and that is why no changes to the document has been done.
        */
        public void runAtomicAsUser (Runnable run) throws SrcException;

        /** Get all Sequence Elements recursively.
        * @return all SeqElement
        */
        public SeqElement[] getAllSeqElements ();
        
        /** Change the set of SeqElement.
         * @param elems the Elements to change
         * @param action one of {@link #ADD}, {@link #REMOVE}, or {@link #SET}
         * @exception SrcException if the action cannot be handled
         */
        public void changeSeqElements(SeqElement[] els, int mode) throws SrcException;

        /** Test whether to accept a new SeqElement to this source.
         */
        public boolean acceptSeqElement(SeqElement sel);
    }
}
