/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or sbbuild/licenses/cddl-gpl-2-cp.txt. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * sbbuild/licenses/cddl-gpl-2-cp.txt.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 2004-2007 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */


package org.starbeans.modules.seqformat.srcgen;

import java.io.FilterWriter;
import java.io.IOException;
import java.io.StringWriter;
import java.io.Writer;
import javax.swing.text.BadLocationException;
import javax.swing.text.Position;
import javax.swing.text.StyledDocument;
import org.openide.text.CloneableEditorSupport;
import org.openide.text.NbDocument;
import org.openide.text.PositionBounds;
import org.openide.text.PositionRef;
import org.starbeans.modules.seqformat.source.SrcException;

/** 
* @author Petr Hamernik, George Waldon
*/
class CodeGenerator {  
    
    /**
     * Obtains a document based on pos' CloneableEditorSupport. If the document cannot be
     * loaded, the method wraps the IOException into SourceException.
     * @param poss positions to extract CloneableEditorSupport from
     * @throw SourceException in case that the document cannot be opened.
     */
    static StyledDocument getDocument(ElementBinding b) throws SrcException {
        return getDocument(b.wholeBounds.getBegin());
    }
    
    static StyledDocument getDocument(PositionRef pos) throws SrcException {
        try {
            CloneableEditorSupport supp = pos.getCloneableEditorSupport();
            return supp.openDocument();
        } catch (IOException ex) {
            SrcText.rethrowException(ex);
        }
        return null;
    }
    
    static final void fillTextBounds(PositionBounds range, String text) 
        throws BadLocationException, java.io.IOException {
        range.setText(text);
    }
    
    static final String readTextBounds(PositionBounds range) throws SrcException {
        try {
            return range.getText();
        } catch (Exception ex) {
            SrcText.rethrowException(ex);
        }
        return null;
    }
   

    static PositionBounds createNewLineBounds(PositionRef where, int boundsType)
    throws SrcException {
        boolean empty = boundsType == TextBinding.BOUNDS_ALL;
        return createNewLineBounds(where, boundsType, empty, empty);
    }
    
    static PositionBounds createNewLineBounds(PositionRef where, int boundsType,
        boolean emptyBefore, boolean emptyAfter) 
        throws SrcException {
        CloneableEditorSupport support = where.getCloneableEditorSupport();
        StyledDocument doc = getDocument(where);

        try {
            int beginText = where.getOffset();
            int lineIndex = NbDocument.findLineNumber(doc, beginText);
            javax.swing.text.Element rootElement = NbDocument.findLineRootElement(doc);
            javax.swing.text.Element line = rootElement.getElement(lineIndex);
            int lineBegin = line.getStartOffset();
            int lineEnd = line.getEndOffset();
            int myLineIndex = lineIndex;

            // if we are on an completely empty line, skip to the next one.
            // Also, if we are at the end of the line (or only whitespaces follow) and the next one
            // is empty, skip it to maintain text readability.
            String trimmedText;
            int textOffset;
            int insertionOffset;
            String lineText = doc.getText(lineBegin, lineEnd - lineBegin);

            trimmedText = lineText.trim();
            textOffset = lineText.lastIndexOf(trimmedText);
            insertionOffset = beginText - lineBegin;
            String formatted;
            PositionBounds bnds;
            int newBlockOffset;
            
            if (trimmedText.length() == 0 || textOffset >= insertionOffset) {
                // check: if the previous line is not empty,
                if (emptyBefore && lineBegin > 0) {
                    javax.swing.text.Element prevLine = rootElement.getElement(myLineIndex - 1);
                    String prevContents =
                        doc.getText(prevLine.getStartOffset(),
                            prevLine.getEndOffset() - prevLine.getStartOffset()
                        ).trim();
                    if (!"".equals(prevContents)) {
                        // insert a newline before the text:
                        // this line's begin is after the prev line's end
                        doc.insertString(lineBegin, "\n", null); // NOI18N
                        line = rootElement.getElement(++myLineIndex);
                        lineBegin = line.getStartOffset();
                    }
                }
                // Nonwhitespace after the insertion point, all whitespace before
                // insert separating newline at the beginning of this line:
                doc.insertString(lineBegin, formatText(doc, lineBegin, "\n"), null);// NOI18N
                formatted = formatText(doc, lineBegin, "\n"); // NOI18N
                formatted = formatted.substring(formatted.indexOf('\n') + 1);
                doc.insertString(lineBegin, formatted, null);
                newBlockOffset = lineBegin + formatted.length();
            } else if (textOffset + trimmedText.length() <= insertionOffset) {
                // Nonwhitespace before the insertion point, only whitespace after
                // retain only indentation
                boolean includingNewline = emptyBefore;
                if (doc.getLength() >= lineEnd) {
                    if (emptyBefore) {
                        javax.swing.text.Element nextLine = rootElement.getElement(myLineIndex + 1);
                        String nextContents = doc.getText(nextLine.getStartOffset(),
                            nextLine.getEndOffset() - nextLine.getStartOffset()).trim();
                        if ("".equals(nextContents)) {
                            // adjust positions after the next (empty) line:
                            lineEnd = nextLine.getEndOffset();
                            lineBegin = nextLine.getStartOffset();
                            includingNewline = false;
                        }
                        myLineIndex++;
                    }
                }
		if (doc.getLength() < lineEnd) {
		    // special case
                    if (emptyBefore && includingNewline) {
                        formatted = formatText(doc, lineEnd-1, "\n\n"); // NOI18N
                        myLineIndex += 2;
                    } else {
                        formatted = formatText(doc, lineEnd-1, "\n"); // NOI18N
                        myLineIndex++;
                    }
		    doc.insertString(lineEnd-1, formatted, null);
		    newBlockOffset = lineEnd-1 + formatted.length();
		    myLineIndex += emptyBefore ? 2 : 1;
		} else {
        	    formatted = formatText(doc, lineEnd, "\n"); // NOI18N
            	    if (!includingNewline)
                	formatted = formatted.substring(formatted.indexOf('\n') + 1);
                    doc.insertString(lineEnd, formatted, null);
	            newBlockOffset = lineEnd + formatted.length();
    		    myLineIndex++;
            	    doc.insertString(newBlockOffset, formatText(
                	doc, newBlockOffset, "\n"), null // NOI18N
                    );
		}
            } else {
                // Whitespace before and after.
                if (emptyBefore) {
                    formatted = formatText(doc, beginText, "\n\n");  // NOI18N
                    myLineIndex++;
                } else {
                    formatted = formatText(doc, beginText, "\n");  // NOI18N
                }
                    
                doc.insertString(beginText, formatted, null);
                // remember THIS offset; the text will be inserted there.
                newBlockOffset = beginText + formatted.length();
                myLineIndex++;
                doc.insertString(newBlockOffset, formatText(doc, newBlockOffset, "\n"), null); // NOI18N
            }
            
            // IF creating a whole new element AND the following line (at newBlockOffset + 1)
            // is not empty, add a additional newline
            if (emptyAfter) {
                if (doc.getLength() > newBlockOffset + 1) {
                    line = rootElement.getElement(myLineIndex + 1);
                    lineBegin = line.getStartOffset();
                    lineEnd = line.getEndOffset();
                    lineText = doc.getText(lineBegin, lineEnd - lineBegin);
                    trimmedText = lineText.trim();
                    if (!"".equals(trimmedText)) {
                        // add a additional newline
                        doc.insertString(newBlockOffset, "\n", null); // NOI18N
                    }
                } else {
                    // try to maintain a newline at the end of the file
                    // which is recommended anyway.
                    doc.insertString(newBlockOffset, "\n", null); // NOI18N
                }
            }
            Position.Bias startBias, endBias;
            switch (boundsType) {
                case TextBinding.BOUNDS_BODY:
                    startBias = Position.Bias.Forward;
                    endBias = Position.Bias.Backward;
                    break;
                default:
                    startBias = Position.Bias.Backward;
                    endBias = Position.Bias.Forward;
                    break;
            }
            PositionRef posBegin = support.createPositionRef(newBlockOffset, startBias);
            PositionRef posEnd = support.createPositionRef(newBlockOffset, endBias); 
            PositionBounds blockBounds = new PositionBounds(posBegin, posEnd);
            return blockBounds;
        } catch (Exception e) {
            SrcText.rethrowException(e);
        }
        return null;
    }

    /**
     * Removes the lines of the passed bounds, possibly collapsing empty previous and/or 
     * following line. The method assumes that the caller has locked out the document
     * (that implies that the document was already loaded by the caller)
     */
    static void clearBounds(PositionBounds bounds, boolean collapseLines) throws BadLocationException {
        StyledDocument doc = bounds.getBegin().getCloneableEditorSupport().getDocument();

        int p1 = bounds.getBegin().getOffset();
        int p2 = bounds.getEnd().getOffset();
        int p3 = 0;
        doc.remove(p1, p2 - p1);

        // remove empty space where was the element placed
        int lineIndex = NbDocument.findLineNumber(doc, p1);
        javax.swing.text.Element lineRoot;
        lineRoot = NbDocument.findLineRootElement(doc);
        
        javax.swing.text.Element line = lineRoot.getElement(lineIndex);
        p1 = line.getStartOffset();
        p2 = line.getEndOffset();
        try {
           if (doc.getText(p1, p2 - p1).trim().length() != 0)
               return;
           doc.remove(p1, p2 - p1);
        } catch (BadLocationException e) {
           return;
        }
        // p1-p2 clears the element together with the
        if (collapseLines) {
           line = lineRoot.getElement(lineIndex);
           try {
               if ("".equals(
               doc.getText(p1 = line.getStartOffset(),
               (p2 = line.getEndOffset()) - line.getStartOffset()).trim())) {
                   doc.remove(p1, p2 - p1);
               } else if (lineIndex > 0) {
                   // try the prev line:
                   line = lineRoot.getElement(lineIndex - 1);
                   if ("".equals(
                   doc.getText(p1 = line.getStartOffset(),
                   (p2 = line.getEndOffset()) - line.getStartOffset()).trim())) {
                       doc.remove(p1, p2 - p1);
                   }
               }
           } catch (BadLocationException e) {
           }
        }
    }
    
    private static String formatText(StyledDocument doc, int pos,
        String text) throws IOException {
             StringWriter stringWriter = new StringWriter();
             Writer writer = new FilterWriter(stringWriter) {};
             writer.write(text);
             writer.close();
             return stringWriter.toString();
    }
}
