/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or sbbuild/licenses/cddl-gpl-2-cp.txt. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * sbbuild/licenses/cddl-gpl-2-cp.txt.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 2004-2007 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */


package org.starbeans.modules.seqformat.srcgen;

import java.beans.BeanInfo;
import java.beans.IntrospectionException;
import java.beans.Introspector;
import java.beans.PropertyDescriptor;
import java.io.FilterWriter;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.io.Writer;
import java.util.HashMap;
import java.util.Map;
import javax.swing.text.BadLocationException;
import javax.swing.text.Document;
import javax.swing.text.Position;
import javax.swing.text.StyledDocument;
import org.openide.text.CloneableEditorSupport;
import org.openide.text.PositionBounds;
import org.openide.text.PositionRef;
import org.starbeans.modules.seqformat.source.Element;
import org.starbeans.modules.seqformat.source.ElementPrinter;
import org.starbeans.modules.seqformat.source.ElementPrinterInterruptException;
import org.starbeans.modules.seqformat.source.ElementProperties;
import org.starbeans.modules.seqformat.source.SrcException;

/**
 *
 * @author  svata, George Waldon
 * @version 
 */
abstract public class ElementBinding implements TextBinding, ElementProperties {

    /** Bounds of the element as whole.
     */
    protected PositionBounds      wholeBounds;
    
    /** Bounds of the sequence header.
     */
    protected PositionBounds      headerBounds;
    
    /** Bounds for the sequence body.
     */
    protected PositionBounds      bodyBounds;

    private Element     el;
    
    /** SrcText text.
     */
    protected SrcText             source;
    
    protected ContainerImpl       containerRef;

    /**
     * Cached Methods used for setting property values.
     */
    static Map<String,java.lang.reflect.Method> setterCache;
    
    /**
     * System-dependent line separator, or null if the system sep is not 
     * defined or is single LF char.
     */
    private static final String lineSeparator;
    
    /**
     * Length of the line separator string.
     */
    private static final int lineSeparatorLength;
    
    /**
     * Linefeed string, used internally to separate lines.
     */
    private static final String LINEFEED = "\n"; // NOI18N
    
    static {
        String sep = System.getProperty("line.separator"); // NOI18N
        if (sep == null || LINEFEED.equals(sep)) {
            lineSeparator = null;
            lineSeparatorLength = -1;
        } else {
            lineSeparator = sep;
            lineSeparatorLength = sep.length();
        }
    }
    
    /** Creates new ElementBinding */
    public ElementBinding(org.starbeans.modules.seqformat.source.Element el, SrcText s) {
        this.el = el;
        this.source = s;
        s.registerBinding(el, this);
    }

    public TextBinding.Container getContainer(String type) {
        return containerRef;
    }

    public Element getElement() {
        return this.el;
    }

    /**
     * Returns the range occupied by the element.
     * @param if true, the returned range will not include the header.
     */
    public PositionBounds getElementRange(boolean defRange) {
        if (defRange)
            return bodyBounds;
        else
            return wholeBounds; 
    }

    public PositionBounds getHeaderRange() {
        return headerBounds;
    }
    
    /**
     * Clones the target element.
     */
    protected abstract Element cloneElement();
 
    /** Give access to source.createPos to subclasses */
    protected  PositionRef createPos(int offset, Position.Bias bias) {
        return source.createPos(offset,bias);
    }
    
    /** Actually creates the element in the SrcText text. This is called by the container
     * binding's insertAfter implementation to actually insert the element. The default
     * implementation will print the element through the ElementPrinter and compute
     * positions of the element's parts.
     */
    public void create(PositionBounds bounds) throws SrcException {
        wholeBounds = bounds;  //bounds = where the creation occurs!
        regenerateWhole(this.el, true); 
    }
    
    /**
     * Prepares the binding for insertion of its sibling. This implementation does nothing,
     * but bindings that are bound together in a textual order may perform some tasks
     * in this methods.
     */
    public PositionRef prepareInsert(ElementBinding tbi, boolean after)
    throws SrcException {
        return after ? wholeBounds.getEnd() : wholeBounds.getBegin();
    }
    
    /** Initializes the new binding with a position right after this element's own pos,
     * optionally creating an empty line for the element, if needed.
     * The implementation assumes that somebody else has already locked the document for
     * writing, so that all offset operations are thread-safe.
     */
    public void insertAfter(ElementBinding b) throws SrcException {
        b.createAt(wholeBounds.getEnd());
    }
    
    /**
     * Generates the associated element representation and associates the binding
     * with a particular container. The representation is inserted between other
     * bindings, "previous" and "next". Container bounds are passed to aid in case
     * of insertion at the container's boundary. emptyBefore and emptyAfter advises
     * whether the container wants the representation to separate itself from the
     * previous, or the next element.
     */
    public void create(ContainerImpl container, 
        ElementBinding previous, ElementBinding next,
        PositionBounds containerBounds, boolean emptyBefore, boolean emptyAfter)
        throws SrcException {
            
        PositionRef beginPos, top;
        if (previous == null) {
            beginPos = containerBounds.getBegin();
        } else {
            beginPos = previous.prepareInsert(this, true);
        }
        
        if (next == null) {
            top = containerBounds.getEnd();
        } else {
            top = next.prepareInsert(this, false);
        }
        PositionBounds gap = new PositionBounds(
            beginPos, top);
        PositionRef pos = source.findFreePosition(gap);
        if (pos == null)
            throw new SrcException("No room for element"); // NOI18N
        createAt(pos, emptyBefore, emptyAfter);
        containerRef = container;
    }
    
    private void createAt(PositionRef r, boolean sepBefore, boolean sepAfter)
    throws SrcException {
        PositionBounds newBounds = CodeGenerator.createNewLineBounds(r,
            BOUNDS_ALL, sepBefore, sepAfter);
        create(newBounds);
    }
    
    private void createAt(PositionRef r) throws SrcException {
        PositionBounds newBounds = CodeGenerator.createNewLineBounds(r,
            BOUNDS_ALL);
        create(newBounds);
    }

    public void applyPropertyChange(Object bean, String property, Object value)
        throws SrcException {
        java.lang.reflect.Method setter;
        
        if (setterCache == null) {
            setterCache = new HashMap<String,java.lang.reflect.Method>(29);
            setter = null;
        } else {
            setter = setterCache.get(property);
        }
        if (setter == null) {
            try {
                BeanInfo binfo = Introspector.getBeanInfo(bean.getClass());
                PropertyDescriptor[] desc = binfo.getPropertyDescriptors();
                for (int i = 0; i < desc.length; i++) {
                    if (desc[i].getName().equals(property)) {
                        setter = desc[i].getWriteMethod();
                    }
                }
            } catch (IntrospectionException ex) {
            }
            if (setter == null) {
                throw new SrcException("Cannot find setter for " + property); // NOI18N
            }
            setterCache.put(property, setter);
        }
        try {
            setter.invoke(bean, new Object[] { value });
            return;
        } catch (IllegalAccessException ex) {
        } catch (java.lang.reflect.InvocationTargetException ex) {
        }
    }


    public ElementBinding findBindingAt(int off) {
        if (wholeBounds.getBegin().getOffset() <= off && 
            wholeBounds.getEnd().getOffset() > off)
            return this;
        else
            return null;
    }
    
    /**
     * Removes the element from the SourceText.
     * the storage
     */
    protected void remove(boolean collapseBefore, boolean collapseAfter) throws SrcException, IllegalStateException {
        try {
            CodeGenerator.clearBounds(wholeBounds, collapseBefore);
        } catch (Exception ex) {
            SrcText.rethrowException(el,ex);
        }
    }
    
    protected void moveTo(ElementBinding before, ElementBinding after) throws SrcException {
        PositionBounds oldWholeBounds = wholeBounds;
        containerRef.insertChild(this, before, after);
        try {
            CodeGenerator.clearBounds(oldWholeBounds, false);
        } catch (BadLocationException ex) {
            SrcText.rethrowException(getElement(),ex);
        }
    }
    
    //==================== CODE GENERATION SUPPORT =============================
    
    /**
     * The method retrieves a document for the element. If the document is not already
     * opened, the method opens the document and returns the instance. The method may
     * throw SrcException.IO, that contains the reason of failure reported by 
     * FileSystem/EditSupport layer.
     */
    protected StyledDocument findDocument() throws SrcException {
        return source.getDocument();
    }
    
    /**
     * Helper function: finds and initializes appropriate format writer for the document.
     * Override if necessary.
     *
     * @param doc the document
     * @param offset offset where should the text appear.
     * @param writer writer for the format output
     * @return formating writer.
     */
    protected Writer findFormatWriter(Document doc, int offset, Writer writer) {
        return new FilterWriter(writer) {};
    }

    protected void regenerateWhole(Element el,
        boolean updatePositions) throws SrcException {
        Document doc = findDocument();
        source.runAtomic(this.el, new PartialGenerator(doc, el, this, updatePositions));
    }

    /**
     * Replaces text for element's header with new contents taken from the parameter.
     * The implementation changes the document inside atomic lock.
     */
    protected void regenerateHeader(Element el) throws SrcException {
        Document doc = findDocument();
        source.runAtomic(this.el, new PartialGenerator(doc, el, this, headerBounds, 
            ElementPrinter.HEADER_BEGIN, ElementPrinter.HEADER_END));
    }
    
    protected void regenerateBody(Element el) throws SrcException {
        Document doc = findDocument();
        source.runAtomic(this.el, new PartialGenerator(doc, el, this, bodyBounds,
            ElementPrinter.BODY_BEGIN, ElementPrinter.BODY_END));
    }
    
    public boolean canInsertAfter() {
        return true;
    }
    
    protected class RemovingRunnable implements ExceptionRunnable {
        public void run() throws Exception {
            CodeGenerator.clearBounds(wholeBounds, true);
        }
    }

    public void updateBounds(int kind, PositionBounds bounds) {
        switch (kind) {
            case BOUNDS_ALL:
                wholeBounds = bounds;
                break;
            case BOUNDS_HEADER:
                headerBounds = bounds;
                break;
            case BOUNDS_BODY:
                bodyBounds = bounds;
                break;
        }
    }

    
    //==========================================================================

    /** 
     * Helper class used for regeneration
     */
    protected class PartialGenerator implements ExceptionRunnable {
        PositionBounds posBounds;
        int begin;
        int end;
        Document doc;
        Element el;
        ElementBinding eb;
        boolean update;
        
        PartialGenerator(Document doc, Element el, ElementBinding eb, boolean updateBounds) {
            update = updateBounds;
            this.eb = eb;
            this.el = el;
            this.doc = doc;
            begin = ElementPrinter.ELEMENT_BEGIN;
            end = ElementPrinter.ELEMENT_END;
            posBounds = wholeBounds;
        }
        
        PartialGenerator(Document doc, Element el, ElementBinding eb, PositionBounds pos, int begin, int end) {
            this.posBounds = pos;
            this.begin = begin;
            this.end = end;
            this.doc = doc;
            this.el = el;
            this.eb = eb;
        }
               
        
        public void run() throws Exception {
            PositionRef start = posBounds.getBegin();
            StyledDocument doc = findDocument();
            StringWriter wr = new StringWriter();
            Writer formatWriter = eb.findFormatWriter(doc, start.getOffset(), wr);
            //ElementPrinterImpl prinetr = getPrinter(indentWriter, el, begin, end);
            ElementPrinterImpl printer = update ?
                new WholeElementPrinter(formatWriter, wr) :
                new ElementPrinterImpl(formatWriter, el, begin, end);
            try {
                el.print(printer);
            }
            catch (ElementPrinterInterruptException e) {
            }
            try {
                formatWriter.close();
            } catch (java.io.IOException ex) {
                throw new InternalError(ex.getMessage());
            }
            
            CodeGenerator.fillTextBounds(posBounds, wr.toString());
            if (update) {
                ((WholeElementPrinter)printer).finish();
            }
        }
    }
    
    PositionRef getEndPosition() {
        StyledDocument doc = source.getEditorSupport().getDocument();
        return source.createPos(doc.getLength(), Position.Bias.Backward);
    }
    
    public PositionBounds findContainerBounds(TextBinding.Container cont) {
        throw new UnsupportedOperationException(this.toString());
    }

    //==========================================================================

    class WholeElementPrinter extends ElementPrinterImpl {
        StringWriter stringWriter;
        CloneableEditorSupport editor;

        int[] positions;

        WholeElementPrinter(Writer writer, StringWriter stringWriter) {
            super(writer);
            this.stringWriter = stringWriter;
            this.editor = source.getEditorSupport();
            this.positions = new int[6];    // As many as Bound types.
            java.util.Arrays.fill(positions, -1);
        }

        public void markNotify(Element el, int what) {
            if (lastText!=null && lastText.endsWith(" ")) {
                String writerText=stringWriter.getBuffer().toString();
                int len=lastText.length();
                
                if (len>0 && !writerText.endsWith(" ")) {
                    char ch=0;  // initialized only to keep compiler happy
                    int i;
                   
                    /* Remove space trailing at end of line. 
                     * Do not remove last new line and do not remove
                     * spaces if only spaces in the last line.*/
                    for (i=len-1;i>=0;i--) {
                        ch=lastText.charAt(i);
                        if (ch!=' ')
                            break;
                    }
                    if (ch!='\n') {
                        i++;
                        stringWriter.write(lastText,i,len-i);
                    }
                }
            }
            /* Compute position */
            positions[what] = stringWriter.getBuffer().length();
            if (what==ELEMENT_END && positions[BODY_END]!=-1)
                positions[BODY_END]=positions[ELEMENT_END];
        }

        /* Fix Bounds */
        void finish() {
            int offset = wholeBounds.getBegin().getOffset();
        
            if (positions[ELEMENT_BEGIN] != -1 && positions[ELEMENT_END] != -1) {
                wholeBounds = createStableBounds(positions[ELEMENT_BEGIN] + offset,
                                             positions[ELEMENT_END] + offset);
            }
            //System.err.println("wholeBounds = " + wholeBounds); // NOI18N
            if (positions[HEADER_BEGIN] != -1 && positions[HEADER_END] != -1) {
                headerBounds = createStableBounds(positions[HEADER_BEGIN] + offset,
                                                positions[HEADER_END] + offset);
            } 
            //System.err.println("headerBounds = " + headerBounds); // NOI18N
            if (positions[BODY_BEGIN] != -1 && positions[BODY_END] != -1) {
                bodyBounds = createStableBounds(positions[BODY_BEGIN] + offset,
                                              positions[BODY_END] + offset);
            }
            //System.err.println("bodyBounds = " + bodyBounds); // NOI18N
        }
	
	private PositionBounds createStableBounds(int begin, int end) {
            if ((begin == -1) || (end == -1))
                return null;

            PositionRef posBegin = editor.createPositionRef(begin, Position.Bias.Backward);
            PositionRef posEnd = editor.createPositionRef(end, Position.Bias.Forward);
            return new PositionBounds(posBegin, posEnd);
	}
    }
    
    public String toString() {
        return "Binding for " + getElement(); // NOI18N
    }

    //==========================================================================
    
    static String convertNewlines(String input) {
        if (lineSeparator == null)
            return input;

        int firstIndex;
        
        //System.err.println("original length = " + input.length()); // NOI18N
	/*
        for (int i = 0; i < input.length(); i++) {
            System.err.print(((int)input.charAt(i)) + ", ");
        }
	*/
        //System.err.println("");
        
        firstIndex = input.indexOf(lineSeparator);
        if (firstIndex == -1)
            return input;
        //System.err.println("Converting newlines..."); // NOI18N
        // replace from the beginning to the firstIndex
        StringBuffer result = new StringBuffer();
        char[] contents = input.toCharArray();
        if (firstIndex > 0)
            result.append(contents, 0, firstIndex);
        result.append(LINEFEED);

        firstIndex += lineSeparatorLength;
        int lastPos = firstIndex;
        while (firstIndex < contents.length) {
            firstIndex = input.indexOf(lineSeparator, firstIndex);
            if (firstIndex == -1) {
                // put there the rest of the string.
                result.append(contents, lastPos, contents.length - lastPos);
                return result.toString();
            }
            // put the portion not containing the separator.
            result.append(contents, lastPos, firstIndex - lastPos);
            result.append(LINEFEED);
            firstIndex += lineSeparatorLength;
            lastPos = firstIndex;
        }
        // execution comes here only if the line.sep is the last thing
        // in the string.
        result.append(contents, lastPos, firstIndex - lastPos);
        
        //System.err.println("result length = " + result.length()); // NOI18N
        
        return result.toString();
    }

    // ====================================================================

    static class ElementPrinterImpl implements ElementPrinter {
        PrintWriter writer;
        ElementPrinter filter;
        String lastText=null;

        Element printedElement;
        int beginMark;
        int endMark;
        int status;

        ElementPrinterImpl(Writer writer) {
            this(writer, null, 0, 0);
            status = 1;
        }

        ElementPrinterImpl(Writer writer, Element printedElement, int beginMark, int endMark) {
            if( writer instanceof ElementPrinter)
                this.filter = (ElementPrinter) writer;
            else
                throw new Error("Writer not instance of ELementPrinter");
            this.writer = new PrintWriter(writer);
            this.printedElement = printedElement;
            this.beginMark = beginMark;
            this.endMark = endMark;
            status = 0;
        }

        public boolean isBegin(Element element, int what) {
            return (printedElement == null) ||
                   ((element == printedElement) && (what == beginMark));
        }

        public boolean isEnd(Element element, int what) {
            return (printedElement == element) && (what == endMark);
        }

        public void markNotify(Element element, int what) {
        }

        public String getString() {
            return writer.toString();
        }

        /** Prints the given text.
        * @param text The text to write
        */
        public void print(String text) throws ElementPrinterInterruptException {
            switch (status) {
            case 0:
                lastText=null;
                return;
            case 1:
                text = convertNewlines(text);
                lastText=text;
                writer.print(text);
                break;
            case 2:
                throw new ElementPrinterInterruptException();
            }
        }
        
        /** Prints the line. New-line character '\n' should be added.
        * @param text The line to write
        */
        public void println(String text) throws ElementPrinterInterruptException {
            print(text);
            print("\n"); // NOI18N
        }

        /** Add the mark to the list, if the printer is currently caching
        * (status == 1) or this mark is the begin.
        * @param element The element to mark
        * @param what The kind of event
        */
        public void mark(Element element, int what) throws ElementPrinterInterruptException {
            filter.mark(element,what);
            switch (status) {
            case 0:
                if (isBegin(element, what)) {
                    markNotify(element, what);
                    status = 1;
                }
                break;
            case 1:
                writer.flush();
                markNotify(element, what);
                if (isEnd(element, what)) {
                    status = 2;
                    writer.close();
                    throw new ElementPrinterInterruptException();
                }
                break;
            case 2:
                throw new ElementPrinterInterruptException();
            }
            
        }
    }
}
