/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or sbbuild/licenses/cddl-gpl-2-cp.txt. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * sbbuild/licenses/cddl-gpl-2-cp.txt.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 2004-2007 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */


package org.starbeans.modules.seqformat.srcgen;

import java.io.IOException;
import java.lang.ref.Reference;
import java.lang.ref.WeakReference;
import java.util.Map;
import java.util.WeakHashMap;
import javax.swing.text.BadLocationException;
import javax.swing.text.Position;
import javax.swing.text.StyledDocument;
import org.openide.text.CloneableEditorSupport;
import org.openide.text.NbDocument;
import org.openide.text.PositionBounds;
import org.openide.text.PositionRef;
import org.starbeans.modules.seqformat.model.Binding;
import org.starbeans.modules.seqformat.source.Element;
import org.starbeans.modules.seqformat.source.SeqElement;
import org.starbeans.modules.seqformat.source.SrcElement;
import org.starbeans.modules.seqformat.source.SrcException;

/**
 * Source is not, in fact, a Binding, but provides context to the editing operations.
 * It should manage locking, interactions with the outside environment and protection
 * management.<P>
 *
 * @author  sdedic, George Waldon
 * @version 0.1
 */
public class SrcText implements DocBinding {
    /** True, if runAtomics should be, in fact, runAtomicAsUser.
     */
    int                     isAtomicAsUser;

    /** Support for editing documents.
     */
    CloneableEditorSupport  editSupport;
    
    /**
     * Undo manager to register undo events with.
     */
    DocBinding.Env                     environment;
    
    /**
     * Yes/no flag whether the whole generator is enabled. Parser updater, for example,
     * disables the code generator while it updates the source model, so the discovered
     * changes are not reflected back to the text.
     */
    boolean                 generatorEnabled;
    
    /**
     * An object used to lock tree of javax.swing.text.Elements or Bindings during
     * updates & reads.
     */
    Object                  treeLock;
    
    /**
     * Binding of the SrcElement, held by a WeakReference. Since the binding
     * is held exclusively by SrcElement.Impl implementation, it may be freed along
     * with the implementation (or according to the implementation memory policy).
     */
    Reference<Binding.Source> refSourceRoot;

    Map<Element,WeakReference<ElementBinding>> bindingMap;
    
    
    public SrcText(DocBinding.Env srcEnvironment) {
        this.environment = srcEnvironment;
        treeLock = new Object();
    }
    
    protected SrcBinding getRoot() {
        synchronized (this) {
            if (refSourceRoot == null)
                return null;
            Object r = refSourceRoot.get();
            if (r == null) 
                refSourceRoot = null;
            return (SrcBinding)r;
        }            
    }
    
    public CloneableEditorSupport getEditorSupport() {
        if (editSupport == null)
            editSupport = environment.findEditorSupport();
        return editSupport;
    }
    
    protected synchronized void registerBinding(Element el, ElementBinding b) {
        if (bindingMap == null) {
            bindingMap = new WeakHashMap<Element,WeakReference<ElementBinding>>(37);
        }
        bindingMap.put(el, new WeakReference<ElementBinding>(b));
    }

    /**
     * Enters or leaves user-mode of document editing. If the document contains
     * some user-protected text, all changes to the text will be disallowed in
     * the user mode.
     */
    public void enableAtomicAsUser(boolean enable) {
        if (enable)
            isAtomicAsUser++;
        else
            isAtomicAsUser--;
    }
    
    public void enableGenerator(boolean enable) {
        generatorEnabled = enable;
    }
    
    public boolean isGeneratorEnabled() {
        return generatorEnabled;
    }
       
    /**
     * Retrieves a document for the I/O operations. The method can throw
     * IOException wrapped into a SourceException, if the operation fails.
     */
    public StyledDocument getDocument() throws SrcException {
        try {
            return getEditorSupport().openDocument();
        } catch (IOException ex) {
            rethrowException(ex);
        }
        return null;
    }
    
    public Element findElement(int position) {
        SrcBinding sb = getRoot();
        if (sb == null)
            return null;
        synchronized (getTreeLock()) {
            ElementBinding b = sb.findBindingAt(position);
            if (b == null)
                return null;
            return b.getElement();
        }
    }
    
    public boolean isAtomicAsUser() {
        return this.isAtomicAsUser > 0;
    }
    
    public void runAtomic(final Runnable r) throws SrcException {
        runAtomic(null, r);
    }
    
    private void runAtomic(Element el, final Runnable r) throws SrcException {
       final Exception[] exc = new Exception[1];
       try {
           if (isAtomicAsUser()) {
               NbDocument.runAtomicAsUser(getDocument(), r);
           } else {
               NbDocument.runAtomic(getDocument(), r);
           }
       } catch (RuntimeException ex) {
           throw ex;
       } catch (Exception ex) {
           exc[0] = ex;
       }
       rethrowException(el, exc[0]);
    }
    
    public void updateBindings(Runnable r) {
        synchronized (getTreeLock()) {
            r.run();
        }
    }
    
    protected Object getTreeLock() {
        return treeLock;
    }

    protected static void rethrowException(Exception ex) throws SrcException {
        rethrowException(null, ex);
    }
    
    protected static void rethrowException(Element el, Exception ex) throws SrcException {
        if (ex == null)
            return;
        SrcException x;
        if (ex instanceof BadLocationException) {
           x = new SrcException.Protection(el);
        } else if (ex instanceof IOException) {
           x =  new SrcException.IO((IOException)ex);
        } else if (ex instanceof SrcException) {
           x = (SrcException)ex;
        } else {
           x = new SrcException();
        }
        x.initCause(ex);
        throw x;
    }
    
    public void runAtomic(final Element el, final ExceptionRunnable r) throws SrcException {
        final Exception[] exc = new Exception[1];
        
        Runnable r2 = new Runnable() {
            public void run() {
                try {
                    r.run();
                } catch (Exception ex) {
                    exc[0] = ex;
                    throw new IllegalStateException();
                }
            }
       };
       StyledDocument doc = getDocument();
       if (isAtomicAsUser()) {
           try {
                NbDocument.runAtomicAsUser(doc, r2);
           } catch (BadLocationException ex) {
               exc[0] = ex;
           } catch (IllegalStateException ex) {
           }
       } else {
           try {
                NbDocument.runAtomic(doc, r2);
           } catch (IllegalStateException ex) {
               // swallow - rethrow the runnable's exception instead.
           }
       }
       rethrowException(el, exc[0]);
    }
    
    protected PositionRef createPos(int offset, Position.Bias bias) {
        return getEditorSupport().createPositionRef(offset, bias);
    }
    
    protected synchronized ElementBinding findBinding(Element e) {
        if (bindingMap == null)
            return null;
        Reference r = (Reference)bindingMap.get(e);
        if (r == null)
            return null;
        return (ElementBinding)r.get();
    }
    
    public boolean canWriteInside(PositionBounds bounds) {
        return environment.findFreePosition(bounds) != null;
    }
    
    protected PositionRef findFreePosition(PositionBounds bounds) {
        return environment.findFreePosition(bounds);
    }

    //====================== Implementation of BindingFactory ================//
  
    public Binding.Source bindSource(SrcElement impl) {
        Binding.Source sb = environment.createSrcBinding(impl,this);
        this.refSourceRoot = new WeakReference<Binding.Source>(sb);
        return sb;
    }
    
    public Binding.Seq bindSeq(SeqElement impl) {
        return environment.createSeqBinding(impl, this);
    }
    
}
