/*                Bistro development code
* 
* This code may be freely distributed and modified under the
* terms of the GNU Lesser General Public Licence.  This should
* be distributed with the code.  If you do not have a copy,
* see:
*
*     http://www.gnu.org/copyleft/lesser.html
* 
* Copyright for this code is held jointly by the individual
* authors.  These should be listed in @author doc comments.
* 
*/
package org.starbeans.modules.bistro.onlineservices.ncbi;

import java.net.MalformedURLException;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Logger;
import org.starbeans.modules.bistro.data.DataThingAdapter;
import org.starbeans.modules.bistro.io.GetStreamTransmitter;
import org.starbeans.modules.bistro.io.StreamProcessor;
import org.starbeans.modules.bistro.io.StreamTransmitter;
import org.starbeans.modules.bistro.io.TransmitterException;
import org.starbeans.modules.bistro.services.Worker;
import org.starbeans.modules.bistro.task.TaskExecutionException;


/**
 * This class provides basic common functionality needed by all NCBIWorkers.
 * 
 * @author mfortner (Taverna)
 */
abstract class AbstractNCBIWorker implements Worker {

    private static Logger logger = Logger.getLogger(AbstractNCBIWorker.class.getName());
    public static final String NEWLINE = System.getProperty("line.separator");

    protected StreamProcessor streamProcessor = null;
    
    /**
     * @param inputMap
     *   A map of values to be used by the worker class.
     */
    public abstract Map execute(Map inputMap) throws TaskExecutionException;

    /**
     * This method returns an array of names of input values. These values are
     * used as keys in the inputMap.
     */
    public String[] inputNames() {
        return new String[]{"term", "maxRecords", "outputFile", "xslt", "ext"};
    }

    /**
     * This method returns an array of input mimetypes.
     */
    public String[] inputTypes() {
        return new String[]{"'text/plain'", "'text/plain'", "'text/plain'", "'text/plain'", "'text/plain'"};
    }

    /**
     * This method returns an array of output value names. These values are used
     * as keys in the outputMap.
     */
    public String[] outputNames() {
        return new String[]{"resultsXml"};
    }

    public String[] outputTypes() {
        return new String[]{"'text/xml'"};
    }

    protected String fetchData(String url) {
        return null;
    }
    protected String rettype;
    protected String term;
    protected String queryKey;
    protected String displayOption;
    protected String originalDb;
    protected int maxRecords = 1;
    protected String startTag;
    protected String endTag;
    protected String url = "http://www.ncbi.nlm.nih.gov/entrez/query.fcgi";
    protected String cmd;
    protected String cmdCurrent;
    protected Map inputMap = new HashMap();
    protected Map transmitterMap = new HashMap();
    protected Map outputMap = new HashMap();

    /**
     * This method is responsible for transmitting a map of values, and
     * processing the results
     * 
     * @param transmitterMap
     * @param streamProcessor
     * @throws MalformedURLException
     * @throws TransmitterException
     */
    protected Map transmit(Map transmitterMap) throws MalformedURLException, TransmitterException {
        StreamTransmitter transmitter = new GetStreamTransmitter();
        transmitter.setURL(this.url);
        DataThingAdapter adapter = new DataThingAdapter(transmitterMap);
        DataThingAdapter inAdapter = new DataThingAdapter(inputMap);
        String outputFile = inAdapter.getString("outputFile");
        String ext = inAdapter.getString("ext");
        logger.fine("ext: " + ext);

        logger.fine("======== outputFile: " + outputFile);
        String xslt = adapter.getString("xslt");
        logger.fine("xslt: " + xslt);
        if (xslt == null || xslt.equals("")) {
            streamProcessor = new NCBIXMLStreamProcessor(outputMap, this.startTag, this.endTag, outputFile, ext);

        } else {
            logger.fine("Using XSLT Stream Processor");
            if (outputFile == null) {
                throw new TransmitterException("The outputFile attribute was null");
            }
            streamProcessor = new NCBIXSLTStreamProcessor(outputMap, this.startTag, this.endTag, xslt, outputFile);
        }

        return transmitter.transmit(transmitterMap, streamProcessor);
    }

    public void setStreamProcessor(StreamProcessor streamProc) {
        this.streamProcessor = streamProc;
    }
}
